package seu.agent.at;


import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumSet;

import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import seu.LaunchAgents;
import seu.agent.SEUAgent;
import seu.agent.at.AmbulanceTeamsTools.partionTools.Cluster;
import seu.agent.at.AmbulanceTeamsTools.partionTools.Clustering;
import seu.common.State;
import seu.model.AgentConstants;


public abstract class AbstractAmbulanceTeamAgent extends SEUAgent<AmbulanceTeam> {
	protected ArrayList<Cluster> clusters ;
	protected int assignedClusterIndex ;
	public State currentTask ;
	public State lastTask;
	
	public void rescue(EntityID target) throws ActionCommandException {
		sendRescue(timeStep, target);
		throw new ActionCommandException(StandardMessageURN.AK_RESCUE);
	}
	
	public void rescue(Human target) throws ActionCommandException {
		rescue(target.getID());
	}

	public void load(EntityID target) throws ActionCommandException {
		sendLoad(timeStep, target);
		throw new ActionCommandException(StandardMessageURN.AK_LOAD);
	}

	public void load(Human target) throws ActionCommandException {
		load(target.getID());
	}

	public void unload() throws ActionCommandException {
		sendUnload(timeStep);
		throw new ActionCommandException(StandardMessageURN.AK_UNLOAD);
	}
	
	public Human someoneOnBoard(Human at) {
		for (StandardEntity next : world.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
			if (next == null)
				continue ;
			Human civilian = (Human)next;
			if (civilian.isPositionDefined() && civilian.getPosition().equals(at.getID())) {
				return civilian;
			}
		}
		return null;
	}
	

	@Override
	protected void initialize() {
		super.initialize();
		getClusters();
		world.getRemainCluster().init(clusters);
	}
	
	protected void getClusters(){
		Collection<StandardEntity> allAT = world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM);
		if (LaunchAgents.SHOULD_PRECOMPUTE)
		{
			String filename = "precompute/Clusters";
			File file = new File(filename);
			if (file.exists()){
				System.out.println("AT partiton file");
				clusters = Clustering.readFromFile(filename, world);
			}
			else {
				try {
					file.createNewFile();
				} catch (IOException e) {
					e.printStackTrace();
				}
				System.out.println("AT partiton file not find");
				clusters = Clustering.KMeansPlusPlus(allAT.size(),
						world.getEntitiesOfType(AgentConstants.BUILDINGS), world);
				Clustering.writeToFile(clusters, filename);
			}
		}
		else{
			clusters = Clustering.KMeansPlusPlus(allAT.size(),
					world.getEntitiesOfType(AgentConstants.BUILDINGS), world);
		}
		
		Clustering.assignAgentsToClusters(allAT, clusters, world);
		assignedClusterIndex = Clustering.getClusterIndexAgentBelong(getID(), clusters);
	//	System.out.println("My position : "+me().getPosition(world));
	//	System.out.println("My cluster index is : " + assignedClusterIndex);
	}

	@Override
	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.AMBULANCE_TEAM);
	}
}
