package seu.agent.at.AmbulanceTeamsTools.rescueTools;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Random;
import java.util.Set;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
//import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.agent.at.AmbulanceTeamsTools.partionTools.Cluster;
import seu.agent.at.AmbulanceTeamsTools.partionTools.Clustering;
import seu.agent.at.AmbulanceTeamsTools.partionTools.DistanceSortor;
import seu.agent.at.AmbulanceTeamsTools.partionTools.GeneralSortor;
import seu.agent.at.AmbulanceTeamsTools.rescueTools.EntityFreezer.Freeze_Tab;
import seu.agent.at.buriedHuman.BuriedHumanDamageInfo;
import seu.agent.at.buriedHuman.BuriedHumanInfoProcesser;
import seu.agent.at.buriedHuman.DamageType;
import seu.geom.Distance;
import seu.model.SEUWorldModel;
import seu.model.object.SEUBuilding;
import seu.util.EntityIdComparator;
import sun.management.resources.agent;

public class RescueTargetSelector {
	private SEUWorldModel world;
	protected int assignedClusterIndex ;
	private Area currentClusterCenter;
	private int currentClusterIndex;
	private final ArrayList<Cluster> clusters;
	protected HashMap<Integer , Integer> numberOfBuriedHumanInCluster;
	private Collection<EntityID> allDiedHuman;
	private Collection<EntityID> allRescuedHuman;
	private Set<EntityID> needlessHumanForRescue;
	private Set<EntityID> currentHumanForRescue;
	private Set<EntityID> humanForEasyRescue;
	private Set<EntityID> allHumanForRescue;
	private ArrayList<EntityID> allAgentForRescue;
	private Set<EntityID> removedHuman;
	private EntityID targetID;
	private Collection<StandardEntity> allRefuges;
	private BuriedHumanInfoProcesser BHI_processer;
	private EntityFreezer freezer;
	private boolean isAssignedClusterNearlyClear;
	private boolean manyVictim;
	private final int VELOCITY = 15700;
	private final int NUMBER_OF_AT_FOR_AGENT = 4;
	private final int DEATH_PERIOD = 170;
	private final int TIME_PRECISION = 5;
	private Set<EntityID> badlyHurtHuman;//niu
	private Set<EntityID> crowdedHuman;//niu   building with many civilians may be more important
	public EntityID forbiddenHuman;//niu
	
	public RescueTargetSelector(SEUWorldModel world , ArrayList<Cluster> clusters , 
			int assignedClusterIndex , BuriedHumanInfoProcesser BHI_processer , EntityFreezer freezer){
		this.world = world;
		this.currentClusterCenter = clusters.get(currentClusterIndex).getCenterEntity();
		this.assignedClusterIndex = assignedClusterIndex;
		this.currentClusterIndex = assignedClusterIndex;
		this.clusters = clusters;
		this.BHI_processer = BHI_processer;
		this.freezer = freezer;
		targetID = null;
		currentHumanForRescue = new HashSet<EntityID>();
		humanForEasyRescue = new HashSet<EntityID>();
		allHumanForRescue = new HashSet<EntityID>();
		needlessHumanForRescue = world.getBuriedHumans().getNeedlessHuman();
		allDiedHuman = world.getBuriedHumans().getAllDiedHuman();
		allRescuedHuman = world.getBuriedHumans().getAllRescuedHuman();
		allRefuges = world.getEntitiesOfType(StandardEntityURN.REFUGE);
		removedHuman = new HashSet<EntityID>();
		allAgentForRescue = new ArrayList<EntityID>();
		badlyHurtHuman = new HashSet<EntityID>();
		crowdedHuman = new HashSet<EntityID>();
		forbiddenHuman=null;
	}
	
	public void update(ChangeSet changed , Set<EntityID> needlessBuildingForRescue , boolean isAssignedClusterNearlyClear){
		this.isAssignedClusterNearlyClear = isAssignedClusterNearlyClear;
		findLoadingAT(changed);
		updateAllHumanForRescue(needlessBuildingForRescue);
		if (allHumanForRescue.isEmpty())
			return ;
//		getInDangerHuman();
		checkIfManyVictim();
		updateHumanForEasyRescue();
		updateHumanForRescue();
	}
	
	public void findLoadingAT(ChangeSet changed){
		for (EntityID entityID : changed.getChangedEntities()){
			StandardEntity se = world.getEntity(entityID);
			if (!se.getURN().equals(StandardEntityURN.AMBULANCE_TEAM.toString()))
				continue ;
			Human at = (Human)se;
			Human civilian = someoneOnBoard(at);
			if (civilian != null && !freezer.containStuff(Freeze_Tab.LOADINGAT , entityID)){
				freezer.addNewfreezingStuff(Freeze_Tab.LOADINGAT , entityID , (int)timeOnPathToRefuge(civilian.getID()));
			}
		}
	}
	
	public void updateAllHumanForRescue(Set<EntityID> needlessBuildingForRescue){
		allHumanForRescue.addAll(BHI_processer.getAllBuriedHuman());
		for (Iterator<EntityID> iterator = allHumanForRescue.iterator() ; iterator.hasNext() ;){
			EntityID humanID = iterator.next();
			Human human = (Human) world.getEntity(humanID);
			if (allDiedHuman.contains(humanID)){
				iterator.remove();
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#1#" + world.getEntity(humanID));
				continue;
			}
			if (allRescuedHuman.contains(humanID) ){
				if (targetID == null){
					iterator.remove();
//					System.out.println(world.me + "   Time : " + world.getTime() + "  remove#2#" + world.getEntity(humanID));
					continue;
				}
				else if (!targetID.equals(humanID)){
					iterator.remove();
//					System.out.println(world.me + "   Time : " + world.getTime() + "  remove#3#" + world.getEntity(humanID));
					continue;
				}
			}
			if (needlessHumanForRescue.contains(humanID)){
				iterator.remove();
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#4#" + world.getEntity(humanID));
				continue;
			}
			if (removedHuman.contains(humanID)){
				iterator.remove();
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#5#" + world.getEntity(humanID));
				continue;
			}
			if (!human.isPositionDefined() || human.getPosition() == null){
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#6#" + world.getEntity(humanID));
				iterator.remove();
				if (targetID != null && targetID.equals(humanID))
					targetID = null;
				continue;
			}
			EntityID posID = human.getPosition();
			if (!(human.getPosition(world) instanceof Building)){
				iterator.remove();
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#7#" + world.getEntity(humanID));
				continue;
			}
			if (human.getPosition(world).getURN().equals(StandardEntityURN.REFUGE.toString())){
				iterator.remove();
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#8#" + world.getEntity(humanID));
				continue ; 
			}
			if (needlessBuildingForRescue.contains(posID)){
				iterator.remove();
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#9#" + world.getEntity(humanID));
				continue;
			}
			Set<EntityID> warmBuildings = freezer.getEntitiesWithTab(Freeze_Tab.WARM_BUILDING);
			if (warmBuildings != null && warmBuildings.contains(posID)){
				iterator.remove();
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#10#" + world.getEntity(humanID));
				continue;
			}
			if (lifeTime(humanID) <= 3){
				iterator.remove();
//				System.out.println(world.me + "  Time : " + world.getTime() + "  " + human + "  damage Type : " + 
//				BHI_processer.getBuriedHumanInfo(humanID).getDamageTypeIndex());
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#11#" + world.getEntity(humanID));
				continue ;
			}
			if (lifeTime(humanID) <= timeOnPathToRefuge(humanID)){
				iterator.remove();
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#12#" + world.getEntity(humanID));
				continue ;
			}
		}
	}
	
	
	public void updateHumanForEasyRescue(){
		humanForEasyRescue.addAll(allHumanForRescue);
		for (Iterator<EntityID> iterator = humanForEasyRescue.iterator() ; iterator.hasNext() ;){
			EntityID humanID = iterator.next();
			if (!canBeRescued(humanID)){
				iterator.remove();
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#14#" + world.getEntity(humanID));
				continue;
			}
			if (shouldRemoveIfEnough(humanID)){
//				System.out.println(world.me + "   Time : " + world.getTime() + "  remove#15#" + world.getEntity(humanID));
				iterator.remove();
				continue ;
			}
		}
	}
	
	public void updateHumanForRescue(){
		if (humanForEasyRescue.isEmpty()){
			if (isAssignedClusterNearlyClear || manyVictim || world.getTime() > DEATH_PERIOD)
				humanForEasyRescue.addAll(allHumanForRescue);
			else 
				return ;
		}
		updateNumberOfBuriedHumanInCluster();
		updateCurrentHumanForRescue();
		if (currentHumanForRescue.isEmpty()){
			if (isAssignedClusterNearlyClear || manyVictim || world.getTime() > DEATH_PERIOD){
				updateRescueSpace();
				updateCurrentHumanForRescue();
			}
			else 
				return ;
		}
	}
	
	public void checkIfManyVictim(){
		int numberOfVictim = allHumanForRescue.size();
		int numberOfAT = world.getAmbulanceTeamIdList().size();
		if (numberOfVictim > 0.57 * numberOfAT)
			manyVictim = true;
		else 
			manyVictim = false;
	}
	

	public void updateCurrentHumanForRescue(){//TODO for agent
		allAgentForRescue.clear();
		if (currentClusterIndex == -1){
//			System.out.println(world.me + "   Time : " + world.getTime() + "  currentClusterIndex == -1");
			return ;
		}
		
		for (EntityID humanID : humanForEasyRescue){
			Human human = (Human)world.getEntity(humanID);
			Building pos = (Building)human.getPosition(world);
			if (!human.getURN().equals(StandardEntityURN.CIVILIAN.toString())){
				allAgentForRescue.add(humanID);
				continue;
			}
			if (targetID != null && humanForEasyRescue.contains(targetID)){
				currentHumanForRescue.add(targetID);
				continue;
			}
			int distanceFromCenterToTarget = world.getDistance(pos , currentClusterCenter);
			int distanceFromMeToTarget = world.getDistance(pos , world.me);
			if (distanceFromCenterToTarget <= 1.0 * Clustering.MAX_RADIUS || 
					((badlyHurtHuman.contains(human)) && distanceFromCenterToTarget < Clustering.MAX_RADIUS*1.5) //niu
					){
				currentHumanForRescue.add(humanID);
				continue ;
			}
			if (distanceFromCenterToTarget <= 2.5 * Clustering.MAX_RADIUS &&  distanceFromMeToTarget <= 1.5 * Clustering.MAX_RADIUS){
				currentHumanForRescue.add(humanID);
				continue ;
			}
			if (distanceFromMeToTarget <= 1.0 * Clustering.MAX_RADIUS
					|| ((badlyHurtHuman.contains(human)) && distanceFromCenterToTarget < Clustering.MAX_RADIUS*1.5) //niu
					){
				currentHumanForRescue.add(humanID);
				continue ;
			}
		}
		if (!allAgentForRescue.isEmpty())
			assignAgentToAT();
		
		if(forbiddenHuman!=null){
//	System.out.println(world.me + "  Time : " + world.getTime() + "   remove " +forbiddenHuman );
		currentHumanForRescue.remove(forbiddenHuman);
		}
		Set<EntityID> freezingHuman = freezer.getEntitiesWithTab(Freeze_Tab.HUMAN);
		if (freezingHuman == null)
			return ;
		currentHumanForRescue.removeAll(freezingHuman);
		if (currentHumanForRescue.isEmpty()){
			for (EntityID humanID : freezingHuman){
				if (humanForEasyRescue.contains(humanID))
					currentHumanForRescue.add(humanID);
			}
			freezer.removeEntityWithTab(Freeze_Tab.HUMAN);
		}
	}
	
	
	protected void assignAgentToAT(){
		Collections.sort(allAgentForRescue , new EntityIdComparator());
		HashSet<EntityID> availableAT = new HashSet<EntityID>();
		ArrayList<DistanceSortor> distancesFromAgentToAT = new ArrayList<DistanceSortor>();
		for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM)){
			Human at = (Human)se;
			if (!allAgentForRescue.contains(at.getID()) && at.isPositionDefined())
				availableAT.add(at.getID());
		}
//		System.out.println(world.me + "  Time : " + world.getTime() + " availableAT : " + availableAT);
		for (EntityID agentID : allAgentForRescue){
			Human agent = (Human)world.getEntity(agentID);
			int distance;
			distancesFromAgentToAT.clear();
			for (EntityID atID : availableAT){
				Human at = (Human)world.getEntity(atID);
				distance = world.getDistance(agent.getPosition() , at.getPosition());
				distancesFromAgentToAT.add(new DistanceSortor(atID , distance));
			}
			Collections.sort(distancesFromAgentToAT , new GeneralSortor());
			for (int i = 0 ; i < Math.min(NUMBER_OF_AT_FOR_AGENT , distancesFromAgentToAT.size()); i++){
				EntityID atID = distancesFromAgentToAT.get(i).getEntityID();
				if (world.me.getID().equals(atID)){
//					System.out.println(world.me + "   Time : " + world.getTime() + " I shoule rescue the agent " + agent);
					currentHumanForRescue.add(agentID);
					return ;
				}
			}
		}
	}
	
	public void updateNumberOfBuriedHumanInCluster(){
		numberOfBuriedHumanInCluster = new HashMap<Integer , Integer>();
		for (EntityID humanID : humanForEasyRescue){
			int index = Clustering.getCloestClusterForHuman(humanID, clusters, world);
			if (!numberOfBuriedHumanInCluster.containsKey(index))
				numberOfBuriedHumanInCluster.put(index , 0);
			numberOfBuriedHumanInCluster.put(index , numberOfBuriedHumanInCluster.get(index).intValue() + 1);
		}
	}
	
	private void updateRescueSpace(){
		currentClusterIndex = getNewRescueCluster();
		if (currentClusterIndex == -1)
			return ;
		currentClusterCenter = clusters.get(currentClusterIndex).getCenterEntity();
	}
	
	private int getNewRescueCluster(){
		int newClusterIndex ;
		ArrayList<Cluster> rescueClusters = new ArrayList<Cluster>();
		for (Integer index : numberOfBuriedHumanInCluster.keySet())
			rescueClusters.add(clusters.get(index));
		int index = Clustering.getCloestClusterForHuman(world.me.getID() , rescueClusters , world);
		if (index == -1)
			return -1;
		newClusterIndex = clusters.indexOf(rescueClusters.get(index));
		return newClusterIndex;
	}
	
	
	protected int getNearAT(Human human){
		float theNumberOfATNearby = 0;
		Set<EntityID> loadingATs = freezer.getEntitiesWithTab(Freeze_Tab.LOADINGAT);
		label : for (StandardEntity at : world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM)){
			Human humanAT = (Human)at;
			if (humanAT.getID().equals(world.me.getID())){
				theNumberOfATNearby += 1;
				continue ;
			}
			if (!humanAT.isPositionDefined())
				continue ;
			if (humanAT.getPosition().equals(world.me.getPosition())){
				theNumberOfATNearby += 1;
				continue ;
			}
		   	else if (world.getDistance(humanAT.getPosition() , human.getPosition()) < Clustering.MAX_RADIUS
		   			||((badlyHurtHuman.contains(human)) && world.getDistance(humanAT.getPosition() , human.getPosition()) < Clustering.MAX_RADIUS*1.5) //niu
		   			){
				if (humanAT.isHPDefined() && humanAT.getHP() == 0)
					continue ;
				if (loadingATs != null && loadingATs.contains(at.getID()))
					continue ;
				if (allHumanForRescue.contains(humanAT.getID()))
					continue ;
				if (human.getPosition().equals(humanAT.getPosition())){
					theNumberOfATNearby += 1;
					continue ;
				}
				for (EntityID humanID : allHumanForRescue){
					Human humanBeingRescued = (Human)world.getEntity(humanID);
					if (humanBeingRescued.getPosition().equals(humanAT.getPosition())){
						theNumberOfATNearby += 0.3;
						continue label;	
					}
				}
				theNumberOfATNearby += 0.7 ;
			}
		}
		return (int)theNumberOfATNearby;
	}
	
	protected boolean canBeRescued(EntityID humanID){
		Human human = (Human)world.getEntity(humanID);
		int minAT = getMinATToRescue(humanID);
		int theNumberOfATNearby = getNearAT(human);
//		System.out.println(world.me + "  Time : " + world.getTime() + "  for human : " + human + 
//				"  minAT : " + minAT + "  theNumberOfATNearby : " + theNumberOfATNearby);
		if (theNumberOfATNearby >= minAT)
			return true;
		else 
			return false;
	}
	
	protected boolean shouldRemoveIfEnough(EntityID humanID){
		Human human = (Human)world.getEntity(humanID);
		HashSet<EntityID> ATAroundVictim = new HashSet<EntityID>();
		HashSet<EntityID> areaAroundVictim = new HashSet<EntityID>();
		int minATToRescue;
		if (human.getURN().equals(StandardEntityURN.CIVILIAN.toString()))
			minATToRescue = getMinATToRescue(humanID);
		else 
			minATToRescue = this.NUMBER_OF_AT_FOR_AGENT;
		areaAroundVictim.add(human.getPosition());
		for (EntityID negID : world.getSeuBuilding(human.getPosition()).getObservableAreas()){
			areaAroundVictim.add(negID);
			Area neg = (Area)world.getEntity(negID);
			if (neg instanceof Building)
				areaAroundVictim.addAll(world.getSeuBuilding(negID).getObservableAreas());
			else
				areaAroundVictim.addAll(world.getSeuRoad(negID).getObservableAreas());
		}
		for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM)){
			Human at = (Human)se;
			if (allDiedHuman.contains(at))
				continue ;
			if (at.isBuriednessDefined() && at.getBuriedness() > 0)
				continue ;
			if (at.isPositionDefined() && areaAroundVictim.contains(at.getPosition()))
				ATAroundVictim.add(at.getID());
		}
		if (ATAroundVictim.size() == minATToRescue && !ATAroundVictim.contains(world.me.getID())){
//			System.out.println(world.me + "   Time : " + world.getTime() + "  true #1#  " + human);
			return true;
		}
//		System.out.println(world.me + "  Time : " + world.getTime() + "  For human : " + human + "  ATAroundVictim.size = " + ATAroundVictim.size() +
//				"  minATToRescue : " + minATToRescue);
		if (ATAroundVictim.size() > minATToRescue){
			if (!ATAroundVictim.contains(world.me.getID())){
//				System.out.println(world.me + "   Time : " + world.getTime() + "  true #2#  " + human);
				return true;
			}
			if (targetID != null && targetID.equals(humanID) && human.getPosition().equals(world.me.getPosition()))
				return false;
//			ArrayList<ElementWithDistance> distanceFromVictimToAT = new ArrayList<ElementWithDistance>();
//			for (EntityID atID : ATAroundVictim){
//				Human at = (Human)world.getEntity(atID);
//				int distance = world.getDistance(at.getPosition(), human.getPosition());
//				distanceFromVictimToAT.add(new ElementWithDistance(atID , distance));
//			}
//			Collections.sort(distanceFromVictimToAT , new GeneralSorter());
//			do {
//				EntityID removeID = distanceFromVictimToAT.remove(distanceFromVictimToAT.size() - 1).getEntityID();
//				if (removeID.equals(world.me.getID())){
////					System.out.println(world.me + "   Time : " + world.getTime() + "  true #3#  " + human);
//					return true;
//				}
//			}while (distanceFromVictimToAT.size() > minATToRescue);
			return false;
		}
		else 
			return false;
	}
	
	public boolean isHumanForRescueEmpty(){
		if (currentHumanForRescue.isEmpty())
			updateHumanForRescue();
		return currentHumanForRescue.isEmpty();
	}
	
	public boolean isNeedToRescue(EntityID humanID){
		return currentHumanForRescue.contains(humanID);
	}
	
	public void clearTargetHuman(){
		removedHuman.add(targetID);
		allHumanForRescue.remove(targetID);
		humanForEasyRescue.remove(targetID);
		currentHumanForRescue.remove(targetID);
		targetID = null;
	}
	
	public EntityID getTargetID(){
		return targetID;
	}
	
	
	protected int getMinATToRescue(EntityID humanID){
		Human human = (Human)world.getEntity(humanID);
		int lifeTime = lifeTime(humanID);
		int timeOnPath = timeOnPath(humanID);
		if (lifeTime <= timeOnPath) 
			return Integer.MAX_VALUE;
		int minATInDouble = (int)Math.ceil(human.getBuriedness()*1.0/(lifeTime - timeOnPath)) ;
//		System.out.println(world.me + "  Time : " + world.getTime() + "  " + human + "   minATInDouble : " + minATInDouble);
//		if (inDangerHuman.contains(humanID))
//			minATInDouble *= 2;
		return minATInDouble;
	}
	
	
	public int lifeTime(EntityID humanID){
		BuriedHumanDamageInfo BHI = BHI_processer.getBuriedHumanInfo(humanID);
		int damageTypeIndex = BHI.getDamageTypeIndex();
		DamageType damageType = BHI_processer.getDamageType(damageTypeIndex);
		if(damageType.getDeathTime()<170){
			badlyHurtHuman.add(humanID);
		}
		int lifeTime = damageType.getDeathTime() - world.getTime();
		return lifeTime ;
	}
	

	public  int rescueTime(EntityID humanID){
		Human human = (Human)world.getEntity(humanID);
		int cost;
		int numberOfAT ;
		int atInBuilding = 0;
		int minNeededAT ;
		for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM)){
			Human at = (Human)se;
			if (at.isPositionDefined() && at.getPosition().equals(world.me.getPosition()) && at.getPosition(world) instanceof Building)
				atInBuilding ++;
		}
		minNeededAT = getMinATToRescue(humanID);
		numberOfAT = Math.max(minNeededAT, atInBuilding);
		cost = (int)Math.ceil(human.getBuriedness() * 1.0f / numberOfAT) + timeOnPath(humanID);
		return (int)cost;
	}


	public int timeOnPath(EntityID humanID)
	{
		return timeOnPathToTarget(humanID) + timeOnPathToRefuge(humanID);
	}
	
	
	public int timeOnPathToTarget(EntityID humanID){
		Human human = (Human)world.getEntity(humanID);
		for (EntityID atID : world.getAmbulanceTeamIdList()){
			Human at = (Human)world.getEntity(atID);
			if(at.isPositionDefined() && at.getPosition().equals(human.getPosition()))
				return 0;
		}
		double distanceToHuman = 0.0;
		int time ; 
		if (!world.me.getPosition().equals(human.getPosition()) && !human.getPosition().equals(world.me.getID()))
			distanceToHuman = Distance.getDistance(world.me.getLocation(world),human.getLocation(world));
		time = (int)Math.ceil(distanceToHuman / VELOCITY); 
		return time;
	}
	

	public int timeOnPathToRefuge(EntityID humanID){
		Human human = (Human)world.getEntity(humanID);
		int distanceToRefuge = 0;
		int time;
		if (human.getURN().equals(StandardEntityURN.CIVILIAN.toString())){
//			if (BHI_processer.getBuriedHumanInfo(humanID).isDamageSure() && 
//					BHI_processer.getBuriedHumanInfo(humanID).getDamageType().equals(DamageTypeEnum.none_slight))
//				return 3;
			if (human.getPosition(world) instanceof Area)
				distanceToRefuge = getMinDistanceToRefuge(human.getID());
			else 
				distanceToRefuge = getMinDistanceToRefuge(world.me.getID());
		}
		time = (int)Math.ceil(distanceToRefuge * 1.0f / VELOCITY);
		return time + 2; //LOAD and UNLOAD each cost one time step
	}
	
	
	public  int freeTime(EntityID humanID)
	{
		int freeTime = lifeTime(humanID) - rescueTime(humanID);
		return freeTime;
	}
	
	
	public Human getTargetHuman(){
		for (EntityID humanID : currentHumanForRescue){
			Human human = (Human)world.getEntity(humanID);
			if (human.getPosition().equals(world.me.getPosition())){
				targetID = humanID;
				return human;
			}
		}
		
		
		ArrayList<EntityID> buriedHuman = new ArrayList<EntityID>(currentHumanForRescue);
		if(!world.isCommunicationLess()  && !world.getAllRefuges().isEmpty() && world.getTime()>Math.min(80, world.getConfig().timestep / 3)
				&& world.getBurningBuildings().isEmpty()){//niu
			//System.out.println(world.me + "  Time : " + world.getTime() + "   try to rescue badly buried first " );
			for(int index = 1 ; index < buriedHuman.size() ; index++){
				Human nextHuman =  (Human)world.getEntity(buriedHuman.get(index));
				if(!badlyHurtHuman.isEmpty() && !badlyHurtHuman.contains(nextHuman)){
					buriedHuman.remove(index);
				}
			}
			if(buriedHuman.isEmpty()){
				buriedHuman = new ArrayList<EntityID>(currentHumanForRescue);
			}else{
				//System.out.println(world.me + "  Time : " + world.getTime() + "   badly buried is not empty" );
			}
		
		}
		
		for(int index = 1 ; index < buriedHuman.size() ; index++){//niu 7.15
			Human nextHuman =  (Human)world.getEntity(buriedHuman.get(index));
			if (nextHuman.isPositionDefined()){		
				StandardEntity position=world.getEntity(nextHuman.getPosition());
				if(position!=null && position instanceof Building
						&&world.getSeuBuilding(position)!=null &&
						!world.getSeuBuilding(nextHuman.getPosition()).getCivilians().isEmpty()){//niu 
					
					for(Human h:world.getSeuBuilding(nextHuman.getPosition()).getCivilians()){//niu 
						if(world.getBuriedHumans().getAllDiedHuman().contains(h.getID()) ||
								world.getBuriedHumans().getAllRescuedHuman().contains(h.getID()) ){
							world.getSeuBuilding(nextHuman.getPosition()).getCivilians().remove(h);
						}
						
					}
					
					if(isBuildingWithManyCiv(world.getSeuBuilding(nextHuman.getPosition()))){
						crowdedHuman.add(buriedHuman.get(index));
					//	System.out.println(world.me + "  Time : " + world.getTime() + "   crowed human  position " + nextHuman.getPosition());
					}
			}
				}
		}
		if(!crowdedHuman.isEmpty()){
			//System.out.println(world.me + "  Time : " + world.getTime() + "   crowed human is not empty" );
			buriedHuman.clear();
			buriedHuman.addAll(crowdedHuman);	
		}
//		if(forbiddenHuman!=null){
//			System.out.println(world.me + "  Time : " + world.getTime() + "   remove " +forbiddenHuman );
//		buriedHuman.remove(forbiddenHuman);
//		}
		
		BuriedHumanSorter sorter = new BuriedHumanSorter();
		Collections.sort(buriedHuman, sorter);
		EntityID rescueID = buriedHuman.get(0);
		if (sorter.getLabel(rescueID) > 1)
			return returnTargetHuman(rescueID) ;
		ArrayList<EntityID> easyRescueHuman = new ArrayList<EntityID>();
		easyRescueHuman.add(rescueID);
		int shortestRescueTime = rescueTime(buriedHuman.get(0));
		for(int index = 1 ; index < buriedHuman.size() ; index++){
			Human nextHuman =  (Human)world.getEntity(buriedHuman.get(index));
			if (round(freeTime(nextHuman.getID()) , TIME_PRECISION) > shortestRescueTime){
				return returnTargetHuman(rescueID) ;
			}
			int rescueTime = round(rescueTime(nextHuman.getID()) , TIME_PRECISION);
			if(rescueTime < shortestRescueTime){
				shortestRescueTime = rescueTime;
				rescueID = nextHuman.getID();
				easyRescueHuman.clear();
				easyRescueHuman.add(rescueID);
			}
			else if (rescueTime == shortestRescueTime)
				easyRescueHuman.add(nextHuman.getID());
				
		}
		//if (easyRescueHuman.size() > 1)
			//System.out.println(world.me + "  Time : " + world.getTime() + "   easyRescueHuman.size : " + easyRescueHuman.size());
		Random random = new Random(world.me.getID().getValue());
		int rescueIndex = random.nextInt(easyRescueHuman.size());
		rescueID = easyRescueHuman.get(rescueIndex);
		return returnTargetHuman(rescueID);
	}
	
	public Human returnTargetHuman(EntityID rescueID){
		Human targetHuman = (Human)world.getEntity(rescueID);
		if (targetID != rescueID && currentHumanForRescue.contains(targetID)){
			freezer.addNewfreezingStuff(Freeze_Tab.HUMAN , targetID , timeOnPathToTarget(rescueID) + 3);
//			System.out.println(world.me + "   Time : " + world.getTime() + "  add " + world.getEntity(targetID) + " to freezer in getTargetHuman");
		}
		targetID = rescueID;
		return targetHuman;
	}
	
	
	public void setTargetHuman(EntityID humanID){
		targetID = humanID;
	}
	
	public int getMinDistanceToRefuge(EntityID humanID){
		if (world.getEntitiesOfType(StandardEntityURN.REFUGE).isEmpty())
			return 3;
		Human human = (Human)world.getEntity(humanID);
		int minDistance = Integer.MAX_VALUE;
		for (StandardEntity refuge : allRefuges){
			int distance = world.getDistance((Area)refuge , (Area)human.getPosition(world));
			if (distance < minDistance)
				minDistance = distance;
		}
		return minDistance;
	}
	
	
	public Civilian someoneOnBoard(Human at) {
		for (StandardEntity next : world.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
			Civilian civilian = (Civilian) next;
			if (civilian.isPositionDefined() && civilian.getPosition().equals(at.getID())) {
				return civilian;
			}
		}
		return null;
	}
	
	public void afterAct(){

		allHumanForRescue.clear();
		humanForEasyRescue.clear();
		currentHumanForRescue.clear();
		allAgentForRescue.clear();
		badlyHurtHuman.clear();
		crowdedHuman.clear();
		if (!isAssignedClusterNearlyClear)
			currentClusterIndex = assignedClusterIndex;
	}
	
	
	private class BuriedHumanSorter implements Comparator<EntityID>{

		@Override
		public int compare(EntityID humanID1, EntityID humanID2) {
			final int label1 = getLabel(humanID1);
			final int label2 = getLabel(humanID2);
			final int freeTime1 = round(freeTime(humanID1) , TIME_PRECISION);
			final int freeTime2 = round(freeTime(humanID1) , TIME_PRECISION); 
			final int rescueTime1 = round(rescueTime(humanID1) , TIME_PRECISION);
			final int rescueTime2 = round(rescueTime(humanID2) , TIME_PRECISION);
			
			if(label1 > label2)
				return -1;
			else if (label1 < label2)
				return 1;
			else if (freeTime1 < freeTime2)
				return -1;
			else if (freeTime1 > freeTime2)
				return 1;
			else if (rescueTime1 < rescueTime2)
				return -1;
			else if (rescueTime1 > rescueTime2)
				return 1;
			return 0;
		}
		
		
		public int getLabel(EntityID humanID){
			Human human = (Human)world.getEntity(humanID);
//			if (inDangerHuman.contains(humanID))
//				return 5;
//			else 
				if(human.getURN().equals(StandardEntityURN.AMBULANCE_TEAM.toString()))
				 return 4;
			else if(human.getURN().equals(StandardEntityURN.FIRE_BRIGADE.toString()))
				 return 3;
			else if(human.getURN().equals(StandardEntityURN.POLICE_FORCE.toString()))
				 return 2;
			else 
				 return 1;
		}
	}
	private int round(int value , int precision){
		int remainder = value % precision;
		value -= remainder;
		if (remainder >= precision * 1.0f / 2)
			value += precision;
		return value;
	}
	
	private boolean isBuildingWithManyCiv(SEUBuilding sb){
		if(world.isMapSmall()){
			if(sb.getCivilians().size()>=2){
				return true;
			}else{
				return false;
			}
		}else{
			if(sb.getCivilians().size()>=4){
				return true;
			}else{
				return false;
			}
		}
	}
}
