package seu.agent.at.Tasks;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
import seu.agent.SEUAgent.ActionCommandException;
import seu.agent.at.AmbulanceTeamAgent;
import seu.agent.at.AmbulanceTeamsTools.partionTools.Cluster;
import seu.agent.at.AmbulanceTeamsTools.rescueTools.BuriednessComparator;
import seu.agent.at.AmbulanceTeamsTools.rescueTools.SearchSpaceManager;
import seu.agent.at.AmbulanceTeamsTools.rescueTools.EntityFreezer.Freeze_Tab;
import seu.common.State;
import seu.model.SEUWorldModel;
import seu.util.EntityIdComparator;


public class NewSaveHumanTask {
	
private SEUWorldModel world;
private  AmbulanceTeamAgent agent;

	public NewSaveHumanTask( AmbulanceTeamAgent agent){
		this.world = agent.getModel();
		this.agent= agent;
		
	}
	
	public void updateCurrentTask(){
//		if(agent.getID().getValue()==2096227595 || agent.getID().getValue()==1201509245
//				){
//			System.out.println(agent.getID() +"      "+ world.getTime() +"  last task " + agent.lastTask);
//		}
		if (checkForSTUCK()) return ;
		if (checkForBURIED()) return ;
		if (checkForDYING()) return ;
		switch (agent.lastTask){
		case LOADING : 
			updateForLOADING();
			break;
		case RESCUING :
			updateForRESCUING();
			break;
		case STUCK:
			updateForSTUCK();
			break;
		case BURIED:
			updateForBURIED();
			break;
		case MOVING_TO_HUMAN :
			updateForMOVING_TO_HUMAN();
			break;
		case SEARCHING :
		case NO_TASK : 
		case RESCUE_OVER :
			getNewTask();
			break;
		default :
			break;
		}
	}
	
	public void executeCurrentTask() throws ActionCommandException{
//		if(agent.getID().getValue()==2096227595 || agent.getID().getValue()==1201509245 ||
//				agent.getID().getValue()==1442853710 || agent.getID().getValue()==785192614||
//				agent.getID().getValue()==2145789159
//				){
//			System.out.println(agent.getID() +"      "+ world.getTime() +"  current task " + agent.currentTask);
//		}
		switch (agent.currentTask){
		case LOADING :
			executeForLOADING();
			break;
		case RESCUING :
			executeForRESCUING();
			break;
		case SEARCHING :
			executeForSEARCHING();
			break;
		case MOVING_TO_HUMAN :
			executeForMOVING_TO_HUMAN();
			break;
		case STUCK :
			executeForSTUCK();
			break;
		case BURIED :
			executeForBURIED();
			break;
		case DYING :
			executeForDYING();
			break;
		case RESCUE_OVER :
			executeForRESCUE_OVER();
			break;
		default : break;
		}
	}
	

	protected void executeForLOADING() throws ActionCommandException{
//		System.out.println(me() + " time : " + time + "  executeForLOADING");
		if (agent.lastTask != State.LOADING){
			agent.load(agent.targetHuman);
			return ;
		}
		if (!world.getEntitiesOfType(StandardEntityURN.REFUGE).isEmpty()){
			//System.out.println("world.getEntitiesOfType(StandardEntityURN.REFUGE)"+ world.getEntitiesOfType(StandardEntityURN.REFUGE).size());
			if (agent.location() instanceof Refuge){
	//			System.out.println(targetHuman);
				agent.targetHuman = null;
				agent.unload();
				return ;
			}
		}
		else{
			EntityID areaID = world.me.getPosition();
			StandardEntity se = (Area)world.getEntity(areaID);
			if (se.getURN().equals(StandardEntityURN.ROAD.toString())){
				agent.targetHuman = null;
				agent.unload();
				return ;
			}
		}
		if (agent.targetHuman.getHP() == 0){
			agent.targetHuman = null;
			agent.unload();
			return ;
		}
		if (agent.targetSelector.lifeTime(agent.targetHuman.getID()) <= 2 && 
				!world.me.getPosition(world).getURN().equals(StandardEntityURN.REFUGE.toString())){
//			System.out.println(world.me + "   Time : " + world.getTime() + "   unload#1#");
			agent.targetHuman = null;
			agent.unload();
			return ;
		}
		if (agent.targetSelector.lifeTime(agent.targetHuman.getID()) <= agent.targetSelector.timeOnPathToRefuge(agent.targetHuman.getID())){
//			System.out.println(world.me + "   Time : " + world.getTime() + "   unload#2#");
			agent.targetHuman = null;
			agent.unload();
			return;
		}
//		if (BHI_processer.getBuriedHumanInfo(targetHuman.getID()).isDamageSure() &&
//				BHI_processer.getBuriedHumanInfo(targetHuman.getID()).getDamageType().equals(DamageTypeEnum.none_slight) &&
//						world.me.getPosition(world).getURN().equals(StandardEntityURN.ROAD.toString())){
//			System.out.println(world.me + "   Time : " + world.getTime() + "   unload#3#");
//			targetHuman = null;
//			unload();
//			return;
//		}
//		System.out.println("load target : " + targetHuman);
		
		if (!world.getEntitiesOfType(StandardEntityURN.REFUGE).isEmpty()){
			List<EntityID> path = agent.router.getMultiDest(agent.location(), 
					world.getEntitiesOfType(StandardEntityURN.REFUGE), 
					agent.router.getAtCostFunction(agent.searchSpaceManager.getMinStaticCost()) , 
					new Point(agent.me().getX(), agent.me().getY()));
			if (path.size() > 1)
				agent.nextPassArea = path.get(1);
			agent.sendMove(path);
			return;
		}
		else {
			Set<Road> roads = world.getEntrance().getEntrance((Building)world.me.getPosition(world));
			List<EntityID> path = agent.router.getMultiDest(agent.location(), 
					roads, 
					agent.router.getAtCostFunction(agent.searchSpaceManager.getMinStaticCost()) , 
					new Point(agent.me().getX(), agent.me().getY()));
			if (path.size() > 1)
				agent.nextPassArea = path.get(1);
			agent.sendMove(path);
		}
//		move(refuges , router.getAtCostFunction(searchSpaceManager.getMinStaticCost()));
	}
	
	protected void executeForRESCUING() throws ActionCommandException{
//		System.out.println(me() + " time : " + time + "  rescue target : " + targetHuman);
		world.getBuriedHumans().addRescuedHuman(agent.targetHuman.getID());
		world.getBuriedHumans().addRescuingHuman(agent.targetHuman.getID());
		agent.rescue(agent.targetHuman);
	}
	
	protected void executeForSEARCHING() throws ActionCommandException{
//		System.out.println(me() + " time : " + time + "  executeForSEARCHING");
		
		List<EntityID> path =agent.router.getMultiDest(agent.location(), 
				agent.searchSpaceManager.getSearchingDestination(), 
				agent.router.getAtCostFunction(agent.searchSpaceManager.getMinStaticCost()) , 
				new Point(agent.me().getX(),agent. me().getY()));
		if (path.size() > 1)
			agent.nextPassArea = path.get(1);
		agent.sendMove(path);
		
//		move(searchSpaceManager.getSearchingDestination() , router.getAtCostFunction(searchSpaceManager.getMinStaticCost()));
	}
	
	private void executeForMOVING_TO_HUMAN() throws ActionCommandException{
//		
		StandardEntity position = agent.targetHuman.getPosition(world);
		List<EntityID> path = agent.router.getAStar(agent.me(), 
				(Area) position, 
				agent.router.getAtCostFunction(agent.searchSpaceManager.getMinStaticCost()));
		if (path.size() >1)//niu =?
			agent.nextPassArea = path.get(1);
//		if(agent.getID().getValue()==2096227595 || agent.getID().getValue()==1201509245||
//				agent.getID().getValue()==1442853710||agent.getID().getValue()==785192614 ||agent.getID().getValue()==2145789159
//				){
		//	System.out.println(agent.getID() +"      "+ world.getTime() + " agent position "  +agent.me().getPosition() +" target "+  agent.targetHuman +" target position " + position +" path   "+ path);
//		}
		agent.sendMove(path , agent.targetHuman.getLocation(world));
		
//		move(targetHuman , router.getAtCostFunction(searchSpaceManager.getMinStaticCost()));
	}
	
	private void executeForSTUCK() throws ActionCommandException{
//		System.out.println(me() + " time : " + time + "  executeForSTUCK");
//		randomWalk();
		agent.sendRest();
	}
	
	private void executeForBURIED() throws ActionCommandException{
//		System.out.println(me() + " time : " + time + "  executeForBURIED");
		agent.sendRest();
	}	
	
	private void executeForDYING() throws SEUAgent.ActionCommandException {
		if (agent.me().getPosition(world).getURN().equals(StandardEntityURN.REFUGE.toString()))
			agent.sendRest();
		else {
			List<EntityID> path = agent.router.getMultiDest(agent.location(), 
					world.getEntitiesOfType(StandardEntityURN.REFUGE), 
					agent.router.getAtCostFunction(agent.searchSpaceManager.getMinStaticCost()) , 
					new Point(agent.me().getX(), agent.me().getY()));
			if (path.size() > 1)
				agent.nextPassArea = path.get(1);
			agent.sendMove(path);
		}
	}
	
	private void executeForRESCUE_OVER() throws SEUAgent.ActionCommandException{
		agent.lookupSearchBuildings();
//		if (!world.getEntitiesOfType(StandardEntityURN.REFUGE).isEmpty()){
//			if (agent.me().getPosition(world).getURN().equals(StandardEntityURN.REFUGE.toString()))
//				agent.sendRest();
//			else {
//				List<EntityID> path = agent.router.getMultiDest(agent.location(), 
//						world.getEntitiesOfType(StandardEntityURN.REFUGE), 
//						agent.router.getAtCostFunction(agent.searchSpaceManager.getMinStaticCost()) , 
//						new Point(agent.me().getX(), agent.me().getY()));
//				if (path.size() > 1)
//					agent.nextPassArea = path.get(1);
//				agent.sendMove(path);
//			}
//		}
//		else{
//			if (agent.me().getPosition(world).getURN().equals(StandardEntityURN.ROAD.toString()))
//				agent.sendRest();
//			else{
//				Set<Road> roads = world.getEntrance().getEntrance((Building)world.me.getPosition(world));
//				List<EntityID> path = agent.router.getMultiDest(agent.location(), 
//						roads, 
//						agent.router.getAtCostFunction(agent.searchSpaceManager.getMinStaticCost()) , 
//						new Point(agent.me().getX(), agent.me().getY()));
//				if (path.size() > 1)
//					agent.nextPassArea = path.get(1);
//				agent.sendMove(path);
//			}
//		}
	}
	
	
	protected void updateForLOADING(){
//		System.out.println(me() + " time : " + time + "  updateForLOADING");
		Human human = agent.someoneOnBoard(world.me);
		if (human != null){
			agent.currentTask = State.LOADING;
			agent.targetHuman = human;
			return;
		}
		else getNewTask();
	}
	
	protected void updateForRESCUING(){
//		System.out.println(me() + " time : " + time + "  updateForRESCUING");
		ArrayList<EntityID> victimInBuilding = new ArrayList<EntityID>();
		ArrayList<EntityID> ATInBuilding = new ArrayList<EntityID>();
		for (EntityID id : agent.changeSet.getChangedEntities()){
			StandardEntity entity = world.getEntity(id);
			if (!(entity instanceof Human))
				continue;
			Human human = (Human)entity;
			if (!human.getPosition().equals(agent.me().getPosition()))
				continue;
			if (human.getURN().equals(StandardEntityURN.AMBULANCE_TEAM.toString()) && human.getHP() > 0 &&
					human.getBuriedness() == 0 && agent.someoneOnBoard(human) == null)
				ATInBuilding.add(id);
			if (human.getURN().equals(StandardEntityURN.CIVILIAN.toString()) && agent.BHI_processer.isBuriedHuman(id) ||
					human.getBuriedness() > 0 && agent.targetSelector.isNeedToRescue(id))
				victimInBuilding.add(id);
		}
		Collections.sort(victimInBuilding , new BuriednessComparator(world));
		Collections.sort(ATInBuilding, new EntityIdComparator());
		
		while (!victimInBuilding.isEmpty()){
			Human victim = (Human)world.getEntity(victimInBuilding.get(0));
			if (victim.getURN().equals(StandardEntityURN.CIVILIAN.toString()) && victim.getBuriedness() == 0){
				if (ATInBuilding.get(0).equals(agent.me().getID())){//id最小的at装载
					agent.targetHuman = victim;
					agent.currentTask = State.LOADING;
					if (agent.targetHuman.getID().equals(agent.targetSelector.getTargetID()))
						agent.targetSelector.clearTargetHuman();
					return;
				}
				else if (agent.targetHuman.getID().equals(victim.getID())){//其他at继续救人
					agent.targetHuman = null;
					agent.targetSelector.clearTargetHuman();
					getNewTask();
					return ;
				}
				EntityID loadingAT = ATInBuilding.remove(0);
				if (!agent.freezer.containStuff(Freeze_Tab.LOADINGAT , loadingAT))
					agent.freezer.addNewfreezingStuff(Freeze_Tab.LOADINGAT , loadingAT , agent.targetSelector.timeOnPathToRefuge(victim.getID()));
			}
			else if (victim.getBuriedness() < ATInBuilding.size()){
				int removeSize = victim.getBuriedness(); // The number of AT to be removed
				for (int i = 0 ; i<removeSize ; i++){
					if (ATInBuilding.get(0).equals(agent.me().getID())){
						agent.targetHuman = victim;
						agent.targetSelector.setTargetHuman(victim.getID());
						return ;
					}
					EntityID loadingAT = ATInBuilding.remove(0);
					if (victim.getURN().equals(StandardEntityURN.CIVILIAN.toString()) && !agent.freezer.containStuff(Freeze_Tab.LOADINGAT , loadingAT))
						agent.freezer.addNewfreezingStuff(Freeze_Tab.LOADINGAT , loadingAT , agent.targetSelector.timeOnPathToRefuge(victim.getID()));
				}
				if (victim.getID().equals(agent.targetHuman.getID())){
					agent.targetHuman = null;
					agent.targetSelector.clearTargetHuman();
					getNewTask();
					return ;
				}
				victimInBuilding.remove(0);
			}
			else break;
		}
//		if (!CVInBuilding.contains(targetHuman.getID())){
//			targetHuman = null;
//		}
		EntityID humanID = world.getBuriedHumans().getNewRescuingHumanFromCommunication();
		Human human = (Human)world.getEntity(humanID);
		if (human != null && victimInBuilding.contains(humanID) && 
				agent.targetSelector.isNeedToRescue(humanID)){
			agent.targetHuman = (Human)world.getEntity(humanID);
			agent.targetSelector.setTargetHuman(humanID);
//			System.out.println(world.me + " Time : " + world.getTime() + " getRescuingHuman : " + humanID +"  in updateForRESCUING");
		}
		if (!agent.targetSelector.isNeedToRescue(agent.targetHuman.getID())){
			agent.targetHuman = null;
			agent.targetSelector.clearTargetHuman();
//			System.out.println("getNewTask for : !BHI_processer.isNeedToRescue(targetHuman.getID())");
			getNewTask();
			return ;
		}
		if (agent.changeSet.getChangedEntities().contains(agent.targetHuman.getPosition()) && 
				!agent.changeSet.getChangedEntities().contains(agent.targetHuman.getID())){
			agent.targetHuman = null;
//			System.out.println("getNewTask for : target human not exists in the building");
			agent.targetSelector.clearTargetHuman();
			getNewTask();
			return ;
		}
		if (world.me.getPosition().equals(agent.targetHuman.getPosition()) && !victimInBuilding.contains(agent.targetHuman.getID())){
			agent.targetHuman = null;
			agent.targetSelector.clearTargetHuman();
			getNewTask();
			return ;
		}
 	}

	protected void updateForSEARCHING(){
//		System.out.println(me() + " time : " + time + "  updateForSEARCHING");
		return ;
	}
	
	
	protected void updateForMOVING_TO_HUMAN(){
//		System.out.println(me() + " time : " + time + "  updateForMOVING_TO_HUMAN");
		if (!agent.targetSelector.isNeedToRescue(agent.targetHuman.getID())){
//			System.out.println(me()+ "  Time : " + time + "========1========");
			
		//	System.out.println(agent.getID() +"      "+ world.getTime() +" new target111 " );
			getNewTask();
			return;
		}
		if (!agent.targetHuman.getPosition().equals(agent.me().getPosition())){
			
			if(agent.targetHuman instanceof AmbulanceTeam){//niu 2015.7.16 解决救AT卡住不动的问题，主要原因时任务选择进入死循环。还需下一步完善
				
					if(world.getEntity(agent.targetHuman.getPosition()) instanceof Building){
						Building b =(Building)world.getEntity(agent.targetHuman.getPosition());
						if((agent.location() instanceof Road && world.getEntrance(b).contains((Road)agent.location()))
								){
							if(!world.isVisible(agent.targetHuman.getID())){
							//	System.out.println(agent.getID() +"      "+ world.getTime() +" at entrance and at not visible " );
								agent.targetSelector.forbiddenHuman=agent.targetHuman.getID();
							}
						}
				}
				
			}
			
			if (!agent.targetSelector.isHumanForRescueEmpty() 
					){
				agent.targetHuman = agent.targetSelector.getTargetHuman();
			//	System.out.println(agent.getID() +"      "+ world.getTime() +" new target222 "+agent.targetHuman );

			}
			else {
				getNewTask();
				return ;
			}
		}
	
		if (agent.targetHuman.getPosition().equals(agent.me().getPosition())){
			if(agent.targetHuman.getID().getValue()==551087011
					){
			//	System.out.println(agent.getID() +"      "+ world.getTime() +" in at position " );
			}
			if (agent.changeSet.getChangedEntities().contains(agent.targetHuman.getID())){
				agent.currentTask = State.RESCUING;
				EntityID humanID = world.getBuriedHumans().getNewRescuingHumanFromCommunication();
				Human human = (Human)world.getEntity(humanID);
				if (human != null && human.isPositionDefined() && human.getPosition().equals(world.me.getPosition()) && 
						agent.targetSelector.isNeedToRescue(humanID)){
					agent.targetHuman = (Human)world.getEntity(humanID);
					agent.targetSelector.setTargetHuman(humanID);
//					System.out.println(world.me + " Time : " + world.getTime() + " getRescuingHuman : " + humanID + "  in updateForMOVING_TO_HUMAN");
				}
				updateForRESCUING();
				return ;
			}
			else {
				
				agent.targetHuman = null;
				agent.targetSelector.clearTargetHuman();
//				System.out.println(me()+ "  Time : " + time + "========3========");
				getNewTask();
				return ;
			}
		}
	}
	
	protected void updateForSTUCK(){
		Human human = agent.someoneOnBoard(world.me);
		if (human != null){
			agent.currentTask = State.LOADING;
			agent.targetHuman = human;
			return;
		}
		getNewTask();
		return ;
	}
	
	protected void updateForBURIED(){
		Human human = agent.someoneOnBoard(world.me);
		if (human != null){
			agent.currentTask = State.LOADING;
			agent.targetHuman = human;
			return;
		}
		getNewTask();
		return ;
	}
	
	protected void getNewTask(){
		if (!agent.targetSelector.isHumanForRescueEmpty()){
			agent.currentTask = State.MOVING_TO_HUMAN;
			agent.targetHuman = agent.targetSelector.getTargetHuman();
		//	System.out.println(agent.getID() +"      "+ world.getTime() +" new target "+agent.targetHuman );

			updateForMOVING_TO_HUMAN();
			return ;
		}
		if (!agent.searchSpaceManager.isBuildingForSearchEmpty()){
			agent.currentTask = State.SEARCHING;
			updateForSEARCHING();
			return ;
		}
		agent.currentTask = State.RESCUE_OVER;
		return ;
	}
	
	protected boolean checkForSTUCK(){
//		System.out.println(me() + " time : " + time + "  checkForSTUCK");
		if (agent.isStucked(world.me)){
			
			//	System.out.println(agent.getID() +"      "+ world.getTime() +"  is stuck " );
			
			agent.currentTask = State.STUCK;
			return true;
		}
		else return false;
	}
	
	protected boolean checkForBURIED(){
//		System.out.println(me() + " time : " + time + "  checkForBURIED");
		if (agent.me().getBuriedness() > 0){
//			if(agent.getID().getValue()==2096227595 || agent.getID().getValue()==1201509245||
//					agent.getID().getValue()==	1773026775||agent.getID().getValue()== 496685447){
//				System.out.println(agent.getID() +"      "+ world.getTime() +"  is buried    " + agent.me().getBuriedness() );
//			}
			agent.currentTask = State.BURIED;
			return true;
		}
		else 
			return false;
	}
	
	protected boolean checkForDYING(){
		if (world.me.getDamage() == 0)
			return false;
		if (agent.someoneOnBoard(agent.me()) != null)
			return false;
		checkBurnInjury();
		int deathTime;
		if(agent.getBurnInjury){
			double hp = world.me.getHP();
			double damage = world.me.getDamage();
			deathTime = 0;
			double k = 0.00025;
			double l = 0.03;
			double m = 0.1;
			while (hp > 0){
				damage += k * damage * damage + l + m;
				hp = hp - damage;
				deathTime++;
			}
		}
		else
			deathTime = (int)Math.ceil(agent.me().getHP() / (agent.me().getDamage() * 1.5));
		int timeToRefuge = agent.targetSelector.timeOnPathToRefuge(agent.me().getID());
		if (deathTime <= timeToRefuge + 2) {
			agent.currentTask = State.DYING;
			return true;
		}
		else return false;
	}
	
	protected void checkBurnInjury(){
		Area area = (Area)world.me.getPosition(world);
		if (area instanceof Building){
			Building building = (Building)area;
			if (building.isOnFire())
				agent.getBurnInjury = true;
		}
	}
	

}
