package seu.agent.fb;

import java.awt.geom.Ellipse2D;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumSet;
import java.util.List;

import seu.LaunchAgents;
import seu.agent.SEUAgent;
import seu.agent.at.AmbulanceTeamsTools.partionTools.Cluster;
import seu.agent.at.AmbulanceTeamsTools.partionTools.Clustering;
import seu.model.AgentConstants;
import seu.model.SEUWorldModel;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;

public abstract class AbstractFireBrigadeAgent extends SEUAgent<FireBrigade> {
	/** The clusters of map generate by Kmeans-Plus-Plus */
	protected ArrayList<Cluster> clusters ;
	
	/** The index of the cluster which the agent assigned to at the beginning*/
	protected int assignedClusterIndex ;
	
	
/*-------------------------------------------------------------------------------------------------*/
	public EntityID extinguishTarget = null;
	public int waterPower = 0;
	
	public void extinguish(EntityID target, int water) throws  ActionCommandException {
		this.extinguishTarget = target;
		this.waterPower = water;
		sendExtinguish(timeStep, target, water);
		throw new ActionCommandException(StandardMessageURN.AK_EXTINGUISH);
	}

	public void extinguish(Building target, int water) throws  ActionCommandException {
		extinguish(target.getID(), water);
	}

	public void extinguish(EntityID target) throws  ActionCommandException {
		extinguish(target, world.getConfig().maxPower);
	}

	public void extinguish(Building target) throws  ActionCommandException {
		extinguish(target.getID());
	}
/*-------------------------------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------------------------------*/
	public boolean isExtinguishable(FireBrigade fb, Building building) {
		Pair<Integer, Integer> position = fb.getLocation(world);
		final int r = world.getConfig().extinguishableDistance;
		java.awt.geom.Area range = new java.awt.geom.Area(new Ellipse2D.Double(
				position.first() - r, position.second() - r,
				r * 2, r * 2));
		range.intersect(new java.awt.geom.Area(building.getShape()));
		return !range.isEmpty();
	}
	
	public boolean isExtinguishable(Building building) {
		return isExtinguishable(me(), building);
	}
	
	@Override
	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.FIRE_BRIGADE);
	}
	@Override
	protected void initialize() {
		super.initialize();
		//getClusters();
		
	}
	
	protected void getClusters(){
		Collection<StandardEntity> allFB = world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE);
		if (LaunchAgents.SHOULD_PRECOMPUTE)
		{
			String filename = "precompute/fbClusters";
			File file = new File(filename);
			if (file.exists()){
				System.out.println("FB partition file exists!");
				clusters = Clustering.readFromFile(filename, world);
			}
			else {
				try {
					file.createNewFile();
				} catch (IOException e) {
					e.printStackTrace();
				}
				System.out.println("FB partition file not exists!");
				clusters = Clustering.KMeansPlusPlus(allFB.size(),
						world.getEntitiesOfType(AgentConstants.BUILDINGS), world);
				Clustering.writeToFile(clusters, filename);
			}
		}
		else{
			clusters = Clustering.KMeansPlusPlus(allFB.size(),
					world.getEntitiesOfType(AgentConstants.BUILDINGS), world);
		}
		
		Clustering.assignAgentsToClusters(allFB, clusters, world);
		assignedClusterIndex = Clustering.getClusterIndexAgentBelong(getID(), clusters);
		
	}
}
