package seu.agent.fb;

import java.awt.Point;
import java.awt.Shape;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import javolution.util.FastSet;
import rescuecore2.config.Config;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.GasStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.LaunchAgents;
import seu.agent.SEUAgent;
import seu.agent.at.AmbulanceTeamsTools.partionTools.Cluster;
import seu.agent.at.AmbulanceTeamsTools.partionTools.Clustering;
import seu.agent.fb.FireBrigadestools.ProcessFile;
import seu.agent.fb.FireBrigadestools.simulators.fireSimulator;

// import seu.common.ProcessAdvantageRatio;
import seu.common.TimeOutException;
import seu.common.clustering.ClusterManager;
import seu.common.clustering.FireCluster;
import seu.common.clustering.FireClusterManager;
import seu.common.clustering.NormalBuildingCluster;
import seu.common.clustering.NormalClusterManager;
import seu.model.SEUWorldModel;
import seu.model.AgentConstants;
import seu.model.object.SEUBuilding;
import seu.standard.simplePartition.GroupingType;

/**
 * A world model for FB only.
 * 
 * @author appreciation-seu
 *
 */
public class FireBrigadeWorld extends SEUWorldModel{

/*-------------------------------------------------------------------------------------------------*/
	private ClusterManager<FireCluster> fireClusterManager;
	//private ClusterManager<NormalBuildingCluster> normalClusterManager;//niu
	private boolean mapUncontrollable=false;//niu
	public boolean finishRetreat=false;
	 public List<Cluster> clusters;
	private fireSimulator simulator;
	private float rayRate = 0.0025f;
	private  Set<SEUBuilding> largestBuildings;
	
	private Set<SEUBuilding> estimatedBurningBuildings = new FastSet<SEUBuilding>();
	
	public Set<SEUBuilding> getEstimatedBurningBuildings() {
		return this.estimatedBurningBuildings;
	}
	
	public fireSimulator getSimulator() {
		return this.simulator;
	}
	
	public ClusterManager<FireCluster> getFireClusterManager() {
		return this.fireClusterManager;
	}
	
//	public ClusterManager<NormalBuildingCluster> getNormalBuildingCluster() {
//		return this.normalClusterManager;
//	}
/*-------------------------------------------------------------------------------------------------*/
	
	public FireBrigadeWorld() {
		super();
	}
	
/*-------------------------------------------------------------------------------------------------*/
	@Override
	public void initialize(SEUAgent<? extends StandardEntity> selfAgent, Config conf, GroupingType type) {
		super.initialize(selfAgent, conf, type);
		largestBuildings= new FastSet<SEUBuilding>();
		if (LaunchAgents.SHOULD_PRECOMPUTE) {
			initConnectionValues();
		} else {
			
			for (SEUBuilding next : getSeuBuildings()) {
				if (!next.isInflammable())
					continue;
				next.initWallValue(this);
				
			}
		}
		
		this.simulator = new fireSimulator(this);
		this.fireClusterManager = new FireClusterManager(this);
		//this.normalClusterManager = new NormalClusterManager(this);
	}
	
	private void initConnectionValues() {
		String fileName = "precompute/connect.ray";
		
		try {
			readConnectedValues(fileName);
		} catch (Exception e) {
			try {
				writeConnectedValues(fileName);
			} catch (Exception e2) {
				e2.printStackTrace();
			}
		}
	}
	
	private void writeConnectedValues(String fileName) throws IOException {
		File f = new File(fileName);
		f.createNewFile();
		BufferedWriter bw = new BufferedWriter(new FileWriter(f));
		bw.write(rayRate + "\n");

		for (SEUBuilding seuBuilding : getSeuBuildings()) {
			seuBuilding.initWallValue(this);

			bw.write(seuBuilding.getSelfBuilding().getX() + "\n");
			bw.write(seuBuilding.getSelfBuilding().getY() + "\n");
			bw.write(seuBuilding.getConnectedBuildings().size() + "\n");

			for (int c = 0; c < seuBuilding.getConnectedBuildings().size(); c++) {
				SEUBuilding building = seuBuilding.getConnectedBuildings().get(c);
				Float val = seuBuilding.getConnectedValues().get(c);
				bw.write(building.getSelfBuilding().getX() + "\n");
				bw.write(building.getSelfBuilding().getY() + "\n");
				bw.write(val + "\n");
			}
		}
		bw.close();
	}

	private void readConnectedValues(String fileName) throws IOException {
		File f = new File(fileName);
		BufferedReader br = new BufferedReader(new FileReader(f));
		Float.parseFloat(br.readLine());
		String nl;
		while (null != (nl = br.readLine())) {
			int x = Integer.parseInt(nl);
			int y = Integer.parseInt(br.readLine());
			int quantity = Integer.parseInt(br.readLine());
			List<SEUBuilding> bl = new ArrayList<SEUBuilding>();
			List<EntityID> bIDs = new ArrayList<EntityID>();
			List<Float> weight = new ArrayList<Float>();
			for (int c = 0; c < quantity; c++) {
				int ox = Integer.parseInt(br.readLine());
				int oy = Integer.parseInt(br.readLine());
				Building building = getBuildingInPoint(ox, oy);
				if (building == null) {
					System.err.println("building not found: " + ox + "," + oy);
					br.readLine();
				} else {
					bl.add(getSeuBuilding(building.getID()));
					bIDs.add(building.getID());
					weight.add(Float.parseFloat(br.readLine()));
				}

			}
			Building b = getBuildingInPoint(x, y);
			getSeuBuilding(b.getID()).setConnectedBuildins(bl);
			getSeuBuilding(b.getID()).setConnectedValues(weight);
		}
		br.close();
	}
/*-------------------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------*/
	@Override
	public void update(Human me, ChangeSet changed) throws TimeOutException{
		super.update(me, changed);
		this.updateFireCluster(me, changed);
		
		// niu 5.22 2015
		if(!finishRetreat && !isMapUncontrollable() && !isCommunicationLess() 
				){//&& getTime()>100
			double mapArea = getMapDimension().getHeight() * getMapDimension().getWidth() / 1000000d;
			double fireBoundingBoxArea=0;
			int contrllbalecount=0;
			for(FireCluster fireCluster:getFireClusterManager().getClusters()){
				fireBoundingBoxArea=fireBoundingBoxArea+fireCluster.getBoundingBoxArea();
				if(!fireCluster.getBorderEntities().isEmpty() && fireCluster.isControllable()){
					contrllbalecount++;
					break;
				}
			}
			if((getFireClusterManager().getClusters().size()>1 && contrllbalecount==0)
					|| fireBoundingBoxArea/mapArea>0.7){//
			
				setMapUncontrollable(true);
				//normalClusterManager.updateClusters();
				if(clusters==null){
					//System.out.println(me + "  Time : " +getTime() + "  KMeansPlusPlus " );
				  clusters = Clustering.KMeansPlusPlus(9,getEntitiesOfType(StandardEntityURN.BUILDING), this);
				}
			}
		}
		
		
	}
	
	private void updateFireCluster(Human me, ChangeSet changed) throws TimeOutException {
		this.simulator.update();
		fireClusterManager.updateClusters();
	}
/*-------------------------------------------------------------------------------------------------*/
	public Set<EntityID> getAreaInShape(Shape shape) {
		Set<EntityID> result = new FastSet<EntityID>();
		for (StandardEntity next : this.getEntitiesOfType(AgentConstants.BUILDINGS)) {
			Area area = (Area) next;
			if (!(area.isXDefined() && area.isYDefined()))
				continue;
			Point p = new Point(area.getX(), area.getY());
			if (shape.contains(p))
				result.add(area.getID());
		}
		return result;
	}
	
	public List<EntityID> getFreeFireBrigades() {
		List<EntityID> freeFireBrigade = new ArrayList<EntityID>();
		List<EntityID> atRefuge = new ArrayList<EntityID>();
		FireBrigade fireBrigade;
		
		freeFireBrigade.addAll(getFireBrigadeIdList());
		freeFireBrigade.removeAll(getStuckedAgents());
		freeFireBrigade.removeAll(getBuriedHumans().getTotalBuriedHuman());
		for (EntityID next : freeFireBrigade) {
			fireBrigade = getEntity(next, FireBrigade.class);
			if (!fireBrigade.isPositionDefined() || getEntity(fireBrigade.getPosition()) instanceof Refuge) {
				atRefuge.add(next);
			}
		}
		freeFireBrigade.removeAll(atRefuge);
		return freeFireBrigade;
	}
	
	public boolean isMapUncontrollable() {
		return mapUncontrollable;
	}
	public void setMapUncontrollable(boolean b) {
		mapUncontrollable=b;
	}
	
	public Cluster findLargestCluster() {
		if (clusters == null || clusters.isEmpty())
			return null;
		Cluster resultCluster = null;
		double MAXAREA = Double.MIN_VALUE;
		 double REFUGE_COEFFICIENT = 0.01;
		 double GAS_STATION_COEFFICIENT = 0.4;
//		Set<NormalBuildingCluster> surroundByOtherCluster = new HashSet<NormalBuildingCluster>();
//		Set<NormalBuildingCluster> dyingOrNotExpanableCluster = new HashSet<>();
		for (Cluster cluster : clusters) {
			//double AREA = cluster.getBoundingBoxArea();
			double value = 0;
	        int refugeNo = 0;
	        int gasStationNo = 0;
	        for (StandardEntity next : getEntitiesOfType(StandardEntityURN.BUILDING)) {
	             
	            	 
	            	 Building building= (Building)next;	
	            	 SEUBuilding sb=getSeuBuilding(next);
	            	 if(cluster.getCluster().contains(building.getID())){
	            	 if (building instanceof Refuge ) {
//		                    value += 1000;
		                    refugeNo++;
		                    
		                } else if (building instanceof GasStation && sb.getEstimatedFieryness() == 0) {
//		                    value += 1000;
		                    gasStationNo++;
		                }
	            	 }
	            
	              if (sb.getEstimatedFieryness() == 0  && 
		            		cluster.getCluster().contains(building.getID()) ) {		      
		                    value += building.getTotalArea();       
		            }
	        }

	            value += (refugeNo * REFUGE_COEFFICIENT * value);
	            value += (gasStationNo * GAS_STATION_COEFFICIENT * value);
	       
	        if (MAXAREA < value) {
	        	MAXAREA = value;
	        	resultCluster = cluster;
	        }
				
		}
		
		
		return resultCluster;
	}
}