package seu.agent.fb.FireBrigadestools.simulators;

import java.util.List;
import java.util.Map;
import javolution.util.FastMap;

import org.uncommons.maths.number.NumberGenerator;
import org.uncommons.maths.random.GaussianGenerator;

import seu.agent.fb.FireBrigadeWorld;
import seu.model.object.SEUBuilding;

public class fireSimulator {
    
    public static float GAMMA = 0.2f;						// resq-fire.gamma
    public static float WATER_COEFFICIENT = 20.0f;			// resq-fire.water_thermal_capacity
    
    private static float BURN_RATE_AVERAGE =  0.15f;		// resq-fire.burn-rate-average
    private static float BURN_RATE_VARIANCE =  0.02f;		// resq-fire.burn-rate-variance
    private static int RANDOM_SEED = 23;					// resq-fire.randomseed
    
    private NumberGenerator<Double> burnRate = 
    		new GaussianGenerator(BURN_RATE_AVERAGE, BURN_RATE_VARIANCE, new java.util.Random(RANDOM_SEED));

    private FireBrigadeWorld world;
    
    public fireSimulator(FireBrigadeWorld world) {
        this.world = world;
    }

    /**
     * this method update building fuel and energy.
     * and get new fieriness and temperature like main fireSimulator.
     */
    public void update() {
    	/*str = "time = " + world.getTime() + ", agent = " + world.getControlledEntity();*/
        burn();
        cool(true);
        if (world.isNoRadio()) {
        	exchangeBuildingNoRadio();
        } else {
        	exchangeBuilding();
        }
        
        cool(false);
        
        for (SEUBuilding b : world.getSeuBuildings()) {
        	b.setVisible(false);
        }
    }

    private void burn() {
        double burnRate = this.burnRate.nextValue();
        for (SEUBuilding b : world.getSeuBuildings()) {
        	if (b.isVisible())
        		continue;
            if (b.getEstimatedTemperature() >= b.getIgnitionPoint() && b.isInflammable() && b.getFuel() > 0) {
                
            	float consumed = b.getConsume(burnRate);
                
            	if (consumed > b.getFuel()) {
                    consumed = b.getFuel();
                }
                
                b.setEnergy(b.getEnergy() + consumed, "burn");
                b.setFuel(b.getFuel() - consumed);
               
                b.setPrevBurned(consumed);
            } else {
                b.setPrevBurned(0f);
            }
        }
    }

    private void exchangeBuildingNoRadio() {
    	for (SEUBuilding b : world.getSeuBuildings()) {
            exchangeWithAir(b);
        }
    	for (SEUBuilding b : world.getSeuBuildings()) {
    		double radiationEnergy = b.getRadiationEnergy();
    		radiationEnergy /= 2.311738394;
    		b.setEnergy(b.getEnergy() - radiationEnergy, "exchangeBuildingNoRadio");
    	}
    }
    
    private void exchangeBuilding() {
        for (SEUBuilding b : world.getSeuBuildings()) {
            exchangeWithAir(b);
        }
        Map<SEUBuilding, Double> radiation = new FastMap<SEUBuilding, Double>();
        for (SEUBuilding b : world.getSeuBuildings()) {
            double radEn = b.getRadiationEnergy();
            
            radiation.put(b, radEn);
        }
        for (SEUBuilding b : world.getSeuBuildings()) {
            double radiationEnergy = radiation.get(b);
            if (world.isCommunicationLess() || world.isCommunicationMedium() || world.isCommunicationLow()) {
                radiationEnergy /= 2.311738394;
            }
            List<SEUBuilding> bs = b.getConnectedBuildings();
            List<Float> vs = b.getConnectedValues();

            for (int c = 0; c < vs.size(); c++) {
            	SEUBuilding temp = bs.get(c);
                double oldEnergy = temp.getEnergy();
                double connectionValue = vs.get(c);
                double a = radiationEnergy * connectionValue;
                double sum = oldEnergy + a;
                bs.get(c).setEnergy(sum, "exchangeBuilding-recever");
            }
            
            b.setEnergy(b.getEnergy() - radiationEnergy, "exchangeBuilding-emitting");
        }
    }

    private void exchangeWithAir(SEUBuilding b) {

        double oldTemperature = b.getEstimatedTemperature();
        double oldEnergy = b.getEnergy();

        if (oldTemperature > 100) {
            b.setEnergy(oldEnergy - (oldEnergy * 0.042), "exchangeWithAir");
        }
    }

    private void cool(boolean first) {
        for (SEUBuilding building : world.getSeuBuildings()) {
        	if (building.isVisible()) {
        		building.setWaterQuantity(0);
        		continue;
        	}
            waterCooling(building);
        }
    }

    private void waterCooling(SEUBuilding b) {
        double lWATER_COEFFICIENT = (b.getEstimatedFieryness() > 0 && b.getEstimatedFieryness() < 4) ? 
        		WATER_COEFFICIENT: WATER_COEFFICIENT * GAMMA;
        
        if (b.getWaterQuantity() > 0) {
            double dE = b.getEstimatedTemperature() * b.getCapacity();
            
            if (dE <= 0) {
                return;
            }
            double effect = b.getWaterQuantity() * lWATER_COEFFICIENT;
            int consumed = b.getWaterQuantity();
            
            if (effect > dE) {
                double pc = 1 - ((effect - dE) / effect);
                effect *= pc;
                consumed *= pc;
            }
            b.setWaterQuantity(b.getWaterQuantity() - consumed);
            b.setEnergy(b.getEnergy() - effect, "waterCooling");
        }
    }
}
