package seu.agent.fb.extinguishtasks;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import com.sun.org.apache.regexp.internal.recompile;

import javolution.util.FastMap;
import javolution.util.FastSet;
import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.GasStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent.ActionCommandException;
import seu.agent.fb.FireBrigadeAgent;
import seu.agent.fb.FireBrigadeWorld;
import seu.agent.fb.FireBrigadestools.FbUtilities;
import seu.agent.fb.FireBrigadestools.fbTarget.FBTarget;
//import seu.agent.fb.FireBrigadestools.targetSelector.FireBrigadeTarget;
import seu.common.TimeOutException;
import seu.common.clustering.FireCluster;
import seu.model.AgentConstants;
import seu.model.object.SEUBuilding;
import seu.model.object.SEURoad;
import seu.model.route.pov.CostFunction;
import seu.util.Util;

public class ExtinguishBehavior {

	private int waterPower;
	private SEUBuilding target;
	private FireBrigadeWorld world;
	
	private FireBrigadeAgent underlyingAgent;
	private FireBrigade controlledEntity;
	private EntityID agentId;
	
	
	protected Map<Integer, SEUBuilding> extinguishTargetAndTimeMap = new FastMap<Integer, SEUBuilding>();
	
	public ExtinguishBehavior(FireBrigadeWorld world) {
		this.world = world;
		this.underlyingAgent = (FireBrigadeAgent) world.getAgent();
		this.controlledEntity = (FireBrigade)world.getControlledEntity();
		this.agentId = world.getAgent().getID();
	}
	
	
	public void extinguish(FireBrigadeWorld model, FBTarget fbTarget) throws ActionCommandException, TimeOutException {
		
		if (fbTarget == null) {

			return;
		}
		
		if (fbTarget.getSeuBuilding() == null) {

			searchDyingCluster((FireCluster)fbTarget.getCluster());
			return;
		}
			
		this.target = fbTarget.getSeuBuilding();
		
		EntityID targetId = fbTarget.getSeuBuilding().getId();
		
		int distanceToTarget = model.getDistance(this.agentId, targetId);
		int locationDistanceToTarget = model.getDistance(controlledEntity.getPosition(), targetId);
		int maxExtinguishDistance;
		
		if (target.getSelfBuilding() instanceof GasStation) {
			maxExtinguishDistance = model.getConfig().viewDistance;
		} else {
			maxExtinguishDistance = model.getConfig().extinguishableDistance;
		}
		
		if (distanceToTarget > maxExtinguishDistance && locationDistanceToTarget > maxExtinguishDistance
				) {
			StandardEntity locationToGo = this.chooseBestLocationToExtinguish(model, fbTarget, controlledEntity);
			if (locationToGo != null ) {
				List<EntityID> path = underlyingAgent.getPath(underlyingAgent.location(), 
						(Area)locationToGo, world.getRouter().getNormalCostFunction());
				if (path != null) {
					if (path.size() > 1) {

						this.underlyingAgent.moveOnPlan(path);
						// System.out.println("In time: " + world.getTime() + " Agent: " + underlyingAgent + " move to tar " );
					} else {
						currentLocationExtinguish(1);
					}
				}
				
			} else {
				currentLocationExtinguish(2);
			}
			
		} else {
			//要注意不要让fb走进房屋
			 
			underlyingAgent.preWater();
			
			Set<EntityID> forbiddenLocations = this.getForbiddenLocation(model, fbTarget, agentId);
			if (!forbiddenLocations.isEmpty() && forbiddenLocations.contains( this.underlyingAgent.location().getID())
					) {//&& ((FireCluster)fbTarget.getCluster()).isControllable()
				
				StandardEntity newLocation = this.chooseBestLocationToExtinguish1(model, fbTarget, controlledEntity);
				if (newLocation != null) {

					this.underlyingAgent.newMove((Area)newLocation, world.getRouter().getNormalCostFunction());
				}
			}
			
//			if(underlyingAgent.firsttimeextinguish){
//				underlyingAgent.firsttimeextinguish=false;
//			System.out.println("in time: " + world.getTime() + ", agent: "
//					+ world.getControlledEntity()
//					+ " finish wait, and has waited for  " );
//			}
			
			
			SEUBuilding building=fbTarget.getSeuBuilding();
			if(world.isVisible(building.getId())){//信道差时，容易估算错
				if(building.getSelfBuilding().isFierynessDefined() && building.getSelfBuilding().isTemperatureDefined() &&
						(building.getSelfBuilding().getFieryness()==0 || building.getSelfBuilding().getFieryness()>=4)
						&& building.getSelfBuilding().getTemperature()<25){
//					System.out.println("in time: " + world.getTime() + ", agent: "
//							+ world.getControlledEntity()
//							+ " visial not to extinguish  " +  building.getSelfBuilding().getID() +" temp "+ building.getSelfBuilding().getTemperature());
//					//currentLocationExtinguish(3);
					return;
				}
			}
			if (this.controlledEntity.getWater() != 0 
					) {//niu && building.getEstimatedFieryness() >= 0 && building.getEstimatedFieryness() <= 4
				//&&building.getEstimatedTemperature()>25
				this.underlyingAgent.target = fbTarget.getSeuBuilding();
				underlyingAgent.targettime=world.getTime();
				
				this.waterPower = FbUtilities.waterNeededToExtinguish( fbTarget.getSeuBuilding());
				
				if(waterPower>((FireBrigade) world.getControlledEntity()).getWater() || waterPower> world.getConfig().maxPower){
				this.waterPower = FbUtilities.calculateWaterPower(world, fbTarget.getSeuBuilding());
				}
				fbTarget.getSeuBuilding().setWaterQuantity(fbTarget.getSeuBuilding().getWaterQuantity() + waterPower);
				
				extinguishTargetAndTimeMap.put(world.getTime(), fbTarget.getSeuBuilding());
				//System.out.println(world.getTime()+"   "+underlyingAgent.getID()+  "   extinguish" + fbTarget.getSeuBuilding().getId());

				this.underlyingAgent.extinguish(fbTarget.getSeuBuilding().getId(), waterPower);
			} else {

			}
		}
		underlyingAgent.isThinkTimeOver("extinguishTarget");
	}
	
	private void searchDyingCluster(FireCluster cluster) throws ActionCommandException, TimeOutException {

		cluster.removeFromAllEntities(lookedBuildings());
		Set<EntityID> dyingClusterEntities = cluster.getAllEntities();

		if (dyingClusterEntities.isEmpty())
			return;

		Set<StandardEntity> dest = new FastSet<StandardEntity>();
		for (EntityID id : dyingClusterEntities) {
			dest.add(world.getEntity(id));
		}

		Point selfL = new Point(controlledEntity.getX(), controlledEntity.getY());
		CostFunction costFun = underlyingAgent.getRouter().getNormalCostFunction();
		Area loca = (Area) world.getSelfPosition();

		List<EntityID> path = underlyingAgent.getRouter().getMultiDest(loca, dest, costFun, selfL);
		EntityID destId = path.get(path.size() - 1);
		double cost = underlyingAgent.getRouter().getRouteCost();
		double dis = world.getDistance(loca.getID(), destId);

		if (cost > 10 * dis) {
			cluster.removeFromAllEntities(unreachables(world.getEntity(destId, Area.class)));
			
		} else {
			
			
			
			this.underlyingAgent.sendMove(path);
		}
	}
	
	private List<EntityID> lookedBuildings() {
		List<EntityID> result = new ArrayList<EntityID>();

		Human agent;
		for (StandardEntity next : world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE)) {
			agent = (Human) next;

			if (agent.getID().getValue() == underlyingAgent.getID().getValue()) {
				for (EntityID visible : underlyingAgent.getChanged()) {
					StandardEntity v_entity = world.getEntity(visible);
					if (v_entity instanceof Building) {
						result.add(visible);
					}
				}
			}
			
			if (agent.isPositionDefined())
				result.add(agent.getPosition());
		}

		return result;
	}
	
	private Set<EntityID> unreachables(Area area) {
		Set<EntityID> result = new FastSet<EntityID>();
		
		if (area instanceof Building) {
			for (Road entrance : world.getEntrance().getEntrance((Building)area)) {
				for (Building next : world.getEntrance().getBuilding(entrance)) {
					result.add(next.getID());
				}
			}
		} else {
			List<Building> bu_s = world.getEntrance().getBuilding((Road)area);
			if (bu_s != null) {
				for (Building next : bu_s) {
					result.add(next.getID());
				}
			}
		}
		
		return result;
	}
	
	/** 
	 * This method handle the case when fb was stucked. I write this method here for the simplicity
	 * of communication handling.
	 */
	public void extinguishNearbyWhenStuck(SEUBuilding targetBuilding) throws ActionCommandException, TimeOutException {
		underlyingAgent.isThinkTimeOver("extinguishNearbyWhenStuck");
		if (targetBuilding == null)
			return;
		this.target = targetBuilding;
		this.waterPower = FbUtilities.waterNeededToExtinguish( target);
		if(waterPower>((FireBrigade) world.getControlledEntity()).getWater()){
		this.waterPower = FbUtilities.calculateWaterPower(world, target);
		}
		targetBuilding.setWaterQuantity(targetBuilding.getWaterQuantity() + waterPower);
		extinguishTargetAndTimeMap.put(world.getTime(), targetBuilding);
		
		
		//System.out.println(world.getTime()+"   "+underlyingAgent+  " stuck   extinguish" + targetBuilding.getId());
		this.underlyingAgent.extinguish(targetBuilding.getId(), waterPower);
	}
	
	private StandardEntity chooseBestLocationToExtinguish(FireBrigadeWorld model, FBTarget target, FireBrigade fbAgent) {
		double minDistance = Double.MAX_VALUE;
		double distance;
		Set<EntityID> forbiddenLocations = this.getForbiddenLocation(model, target, fbAgent.getID());
		Set<EntityID> possibleArea = new FastSet<EntityID>();
		StandardEntity locationToExtinguish = null;
		
		Collection<EntityID> areaInExtinguishableRange = target.getSeuBuilding().getAreasInExtinguishableRange();
		possibleArea.addAll(areaInExtinguishableRange);
		possibleArea.removeAll(forbiddenLocations);
		
		
		if (possibleArea.isEmpty())
			possibleArea.addAll(areaInExtinguishableRange);
		
		Set<Building> possibleBuildingsToExtinguish = new FastSet<Building>();
		for (EntityID next : possibleArea) {
			StandardEntity entity = model.getEntity(next);
			if (entity instanceof Road) {
				distance = model.getDistance(target.getSeuBuilding().getId(), entity.getID());
				if (distance < minDistance) {
					minDistance = distance;
					locationToExtinguish = entity;
				}
			} else if (entity instanceof Building) {
				possibleBuildingsToExtinguish.add((Building) entity);
			}
		}
		if (locationToExtinguish == null) {
			for (Building next : possibleBuildingsToExtinguish) {
				distance = model.getDistance(target.getSeuBuilding().getId(), next.getID());
				boolean flag = next.isFierynessDefined() && next.getFieryness() >= 4 && next.getFieryness() < 1;
				if (distance < minDistance && (!next.isFierynessDefined() || flag)) {
					minDistance = distance;
					locationToExtinguish = next;
				}
			}
		}

		return locationToExtinguish;
	}
	
	private StandardEntity chooseBestLocationToExtinguish1(FireBrigadeWorld model, FBTarget target, FireBrigade fbAgent) {
		
		
		
		double minDistance = Double.MAX_VALUE;
		double distance;
		Set<EntityID> forbiddenLocations = this.getForbiddenLocation(model, target, fbAgent.getID());
		Set<EntityID> possibleArea = new FastSet<EntityID>();
		StandardEntity locationToExtinguish = null;
		if(needtoCheckBeforeExtin()
			){//niu  all visible is burned out, on this situation need to move to target to check the burning
//			System.out.println("in time: " + world.getTime() + ", agent: " + world.getControlledEntity() 
//					+"all visible is burned out, need to check");
			locationToExtinguish=world.getEntity(target.getSeuBuilding().getId());
			
		}else{
		  
		Collection<EntityID> areaInExtinguishableRange = target.getSeuBuilding().getAreasInExtinguishableRange();
		possibleArea.addAll(areaInExtinguishableRange);
		possibleArea.removeAll(forbiddenLocations);
		
		
		if (possibleArea.isEmpty())
			possibleArea.addAll(areaInExtinguishableRange);
		
		Set<Building> possibleBuildingsToExtinguish = new FastSet<Building>();
		for (EntityID next : possibleArea) {
			StandardEntity entity = model.getEntity(next);
			if (entity instanceof Road) {
				distance = model.getDistance(target.getSeuBuilding().getId(), entity.getID());
				if (distance < minDistance) {
					minDistance = distance;
					locationToExtinguish = entity;
				}
			} else if (entity instanceof Building) {
				possibleBuildingsToExtinguish.add((Building) entity);
			}
		}
		if (locationToExtinguish == null) {
			for (Building next : possibleBuildingsToExtinguish) {
				distance = model.getDistance(target.getSeuBuilding().getId(), next.getID());
				boolean flag = next.isFierynessDefined() && next.getFieryness() >= 4 && next.getFieryness() < 1;
				if (distance < minDistance && (!next.isFierynessDefined() || flag)) {
					minDistance = distance;
					locationToExtinguish = next;
				}
			}
		}
		}
		return locationToExtinguish;
	}
	
		
	public void currentLocationExtinguish(int marker) throws ActionCommandException {
		StandardEntity locationToExtinguish = null;
		if( needtoCheckBeforeExtin()
			){//niu  marker==1  all visible is burned out, on this situation need to move to target to check the burning
//			System.out.println("in time: " + world.getTime() + ", agent: " + world.getControlledEntity() +
//					"all visible is burned out,need to  check  111111111");
			return;
		}
//		if(marker==1 && world.isMapUncontrollable()){
//			
//		try {
//			System.out.println("prewater "
//					+ "current 1" );
//			underlyingAgent.preWater();
//		} catch (TimeOutException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		}
//		}
		
		SortedSet<SEUBuilding> inRanges = new TreeSet<SEUBuilding>(new Comparator<SEUBuilding>() {

			@Override
			public int compare(SEUBuilding o1, SEUBuilding o2) {
				if (o1.BUILDING_VALUE > o2.BUILDING_VALUE)
					return 1;
				if (o1.BUILDING_VALUE < o2.BUILDING_VALUE)
					return -1;
				return 0;
			}
		});
		
		inRanges.addAll(FbUtilities.getBuildingInExtinguishableRange(world, controlledEntity.getID()));
		
		if (inRanges.isEmpty())
			return;
		SEUBuilding target = inRanges.first();
		
		if (this.controlledEntity.getWater() != 0 && target.getEstimatedFieryness() >= 0 && target.getEstimatedFieryness() <= 4
				&& target.getEstimatedTemperature()>25
				) { 
			
			
			
			this.waterPower = FbUtilities.waterNeededToExtinguish( target); 
			if(waterPower>((FireBrigade) world.getControlledEntity()).getWater() || waterPower> world.getConfig().maxPower){
			this.waterPower = FbUtilities.calculateWaterPower(world, target);
			}
			
			target.setWaterQuantity(target.getWaterQuantity() + waterPower);
			extinguishTargetAndTimeMap.put(world.getTime(), target);
			//System.out.println(world.getTime()+"   "+underlyingAgent+  "  current location extinguish" + target.getId() + "   marker   " + marker);	
			underlyingAgent.extinguish(target.getId(), waterPower);
		} else {
			
		}
	}

	private Set<EntityID> getForbiddenLocation(FireBrigadeWorld model, FBTarget target, EntityID fbAgent) {
		Set<EntityID> forbiddenLocations = new FastSet<EntityID>();
		
		if (target.getCluster() != null) {
			forbiddenLocations.addAll(target.getCluster().getAllEntities());
		}
		forbiddenLocations.addAll(Util.seuBuildingToId(world.getEstimatedBurningBuildings()));
		
		int number = 6;//3
		if (model.isMapMedium()) number = 10;
		if (model.isMapHuge()) number = 15;
		
		for(EntityID next : model.getFireBrigadeIdList()) {
			// -------------- ?????????????????????????? 
			if (fbAgent.getValue() < next.getValue() 
					&& model.getDistance(fbAgent, next) < model.getConfig().viewDistance
					&& --number < 0) {
				FireBrigade fireBrigade = model.getEntity(next, FireBrigade.class);
				StandardEntity entity = world.getEntity(fireBrigade.getPosition());
				if (entity instanceof Building) {
					SEUBuilding buildingOfNearFb = model.getSeuBuilding(entity.getID());
					forbiddenLocations.addAll(buildingOfNearFb.getObservableAreas());
				} else if (entity instanceof Road) {
					SEURoad roadOfNearFb = model.getSeuRoad(entity.getID());
					forbiddenLocations.addAll(roadOfNearFb.getObservableAreas());
				}
			}
		}
		
		return forbiddenLocations;
	}
	
	private boolean needtoCheckBeforeExtin() {
		int buildingcount=0;
			int burnedoutcout=0;
		for(EntityID eid:underlyingAgent.getVisibleEntities()){
			StandardEntity st=world.getEntity(eid);
			
			if(st instanceof Building){
				buildingcount++;
				Building b=(Building)st;
				if(b.isFierynessDefined() && b.getFieryness()==8){
					burnedoutcout++;
				}
			}
		}
		if(burnedoutcout>0 && burnedoutcout==buildingcount){
			return true;
		}
		return false;
	}
}
