package seu.agent.pf;

import java.awt.Point;
import java.util.Collection;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;

import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.AKClear;
import rescuecore2.standard.messages.AKClearArea;
import rescuecore2.standard.messages.AKMove;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import seu.Viewer.layers.SEU_ZonePolygonLayer;
import seu.agent.SEUAgent;
import seu.agent.pf.PFLastTaskType.PFClusterLastTaskEnum;

import seu.agent.pf.cluster.Cluster;
import seu.agent.pf.cluster.Clustering;
import seu.model.AgentConstants;
import seu.model.object.seuZoneEntity.SeuZones;
import seu.model.route.pov.CostFunction;

public abstract class AbstractPoliceForceAgent extends SEUAgent<PoliceForce> {

	/**
	 * A list of availiable clusters of current map.
	 */
	protected List<Cluster> clusters;
	
	/**
	 * Index of the cluster assigned to this agent at initialization.
	 */
	protected int assignedClusterIndex;
	
	/**
	 * Index of the cluster this agent currently work on.
	 */
	protected int currentClusterIndex;
	
	/**
	 * The cluster this agent currently work on.
	 */
	protected Cluster currentCluster;
	
	/**
	 * All zones of current map. And we will clustering those zones into clusters.
	 */
	private SeuZones zones;
	
	/**
	 * A set of unvisited entrances of current cluster.
	 */
	protected Set<Road> traversalEntranceSet;
	
	/**
	 * A set of unvisited critical areas of current cluster.
	 */
	protected Set<Area> traversalCriticalAreas;
	
	/**
	 * A set of unvisited refuges of current cluster.
	 */
	protected Set<Refuge> traversalRefugeSet;

	/**
	 * The task target of this agent in last cycle.
	 */
	protected PFLastTaskTarget taskTarget;
	
	/**
	 * The last task type of this agent's cluster's tasks.
	 */
	protected PFClusterLastTaskEnum clusterLastTaskType;
	
	/**
	 * The path this agent will following when cllearing.
	 */
	protected List<EntityID> lastCyclePath = null;
	
	/**
	 * The clear strategy this agent currently using.
	 */
	//protected SEU_ClearStrategy clearStrategy;
	
	/**
	 * A list of clusters this agent needed to expand to.
	 */
	protected List<Cluster> needToExpandClusters;
	
	/**
	 * A list of clusters this agent has searched or is searching now.
	 */
	protected List<Cluster> expandedClusters;
	
	/**
	 * A set of burning buildings in this agent's current and neighbour clusters.
	 */
	protected Set<Building> nearBurningBuildings;
	
	/**
	 * A set of burning buildings this agent has searched.
	 */
	protected Set<Building> hadSearchBuringBuildings;
	
	/**
	 * A set of entrances of the burning zone that this agent is searching now.
	 */
	protected Set<Road> entrancesOfBurningZone;
	
	/**
	 * The number of buildings of this agent's expaned clusters.
	 */
	protected int buildingCount;
	
	/**
	 * Flags to determines whether needs to search burning building or not.
	 */
	protected boolean searchBurningFlag;
	
	protected Set<Refuge> coincidentRefuge = new HashSet<Refuge>();
	
	protected Set<EntityID> visitedRefuges = new HashSet<EntityID>();
	
	protected Set<Human> coincidentBuriedAgent = new HashSet<Human>();
	
	protected Set<EntityID> visitedBuriedAgent = new HashSet<EntityID>();
	
	@Override
	protected void initialize() {
		super.initialize();

		zones = world.getZones();
		getClusters();
	}
	
	@Override
	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.POLICE_FORCE);
	}

	protected void getClusters() {
		Collection<StandardEntity> allPF = world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE);
		clusters = Clustering.KMeansPlusPlus(allPF.size(), zones);
		Clustering.assignAgentsToClusters(allPF, clusters, world);
		assignedClusterIndex = Clustering.getClusterIndexAgentBelongTo(getID(), clusters);
		currentClusterIndex = assignedClusterIndex;
		SEU_ZonePolygonLayer.CLUSTERS = clusters;
		
		
	}
	 
	@Override
	public void sendMove(int time, List<EntityID> path, int destX, int destY) {
		send(new AKMove(getID(), time, path, destX, destY));
	}
	
	@Override
	public void sendClear(int time, int destX, int destY){
		send(new AKClearArea(getID(), time, destX, destY));
	}
	
	@Override
	public void sendClear(int time, EntityID target) {
		send(new AKClear(getID(), time, target));
	}

	@Override
	protected void moveToRefuge() throws ActionCommandException {
		Collection<StandardEntity> refuges = world.getEntitiesOfType(StandardEntityURN.REFUGE);
		if (refuges.isEmpty()) 
			return;
		
		CostFunction costFunc = router.getNormalCostFunction();
		Point selfL = new Point(((Human) me()).getX(),((Human) me()).getY());
		lastCyclePath = router.getMultiAStar(location(), refuges, costFunc, selfL);
		
		
		sendMove(timeStep, lastCyclePath);
		throw new ActionCommandException(StandardMessageURN.AK_MOVE);
	}
}