package seu.agent.pf.Tasks;

import java.awt.Point;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javolution.util.FastSet;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;
import seu.geom.CompositeConvexHull;
import seu.geom.PolygonScaler;
import seu.model.SEUWorldModel;
import seu.model.object.SEUBuilding;
import seu.model.object.seuZoneEntity.SeuZone;


public class BurningBuildingCluster {
	private List<Building> burningBuldings;
	private List<SeuZone> burningRegions;
	//private int size;
	private List<Building> borderBurningBuildings;
	//private Map<Region,Integer> burningRegionLevel;
	private SEUWorldModel world;
	private Point centerPoint;
	
	public BurningBuildingCluster(List<SeuZone> burningRegion,SEUWorldModel world){
		burningRegions=burningRegion;
		this.world=world;
		burningBuldings=new ArrayList<Building>();		
		borderBurningBuildings=new ArrayList<Building>();
		//burningRegionLevel=new HashMap<Region,Integer>();
		initBurningCluster();	
	}
	
	private void initBurningCluster(){
		if(burningRegions!=null&&!burningRegions.isEmpty()){
			for(SeuZone next:burningRegions){
				List<SEUBuilding> tempB=next.getBuldingsInRegion();
				if(tempB!=null&&!tempB.isEmpty()){
					for(SEUBuilding next1:tempB){
						burningBuldings.add(next1.getSelfBuilding());
					}					
				}
			}
			CompositeConvexHull convexHull = new CompositeConvexHull();
			Set<StandardEntity> allEntities = new HashSet<StandardEntity>();			
	    	try{
	    	for (Building entity : burningBuldings) {
	    		for(EntityID next:entity.getNeighbours()){
	    			StandardEntity s=world.getEntity(next);
	    			allEntities.add(s);
	    		}
				//allEntities.add(entity.getn);
				Pair<Integer, Integer> location = entity.getLocation(world);
				convexHull.addPoint(location.first(), location.second());
	    	}	  
	    	Set<StandardEntity> borderBuilding = 
	    			PolygonScaler.getBurningBorderBuildings(convexHull, allEntities, 1.1, world);
	    	double cx = convexHull.getConvexPolygon().getBounds().getCenterX();
	        double cy = convexHull.getConvexPolygon().getBounds().getCenterY();
	        centerPoint = new Point((int) cx, (int) cy);
	    	borderBuilding.removeAll(world.getBorderBuildings());
	    	for(StandardEntity next:borderBuilding){
	    		if(next instanceof Building){
	    			borderBurningBuildings.add((Building)next);
	    		}
	    	}
	    	}catch(Exception e){
	    		System.out.println("创建着火分区出错");
	    	}
	    	
		}
		
	}
	
	public void setBurningBuildings(List<Building> bs){
		burningBuldings=bs;
	}
	
	public List<Building> getBurningBuildings(){
		return burningBuldings;
	}
	
	public void setBorderBurningBuilding(List<Building> bs){
	    borderBurningBuildings=bs;
	}
	
	public List<Building> getBorderBurningBuilding(){
		return borderBurningBuildings;
	}
	
	public int size(){
		return burningBuldings.size();
	}
	
	public Point getCenterPoint(){
		return centerPoint;
	}

}
