package seu.agent.pf.Tasks;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
import seu.agent.SEUAgent.ActionCommandException;
import seu.agent.pf.PoliceForceAgent;
import seu.agent.pf.Tasks.SEUPFLastTaskType.PFPartitionLastTaskEnum;
import seu.common.BasicTask;
import seu.geom.Distance;
import seu.geom.GeomUtil;
import seu.model.SEUWorldModel;
import seu.model.route.pov.CostFunction;
import seu.standard.DistanceComparator;

public class HelpSelfTask extends BasicTask<PoliceForceAgent>{

	SEUWorldModel world;
	
	public int stuckTime=10;
	public HelpSelfTask(PoliceForceAgent agent) {
		super(agent);
		world=agent.getModel();
		
	}



	@Override
	public void execute() throws ActionCommandException {
		
		if(isCantWork()){
			agent.sendRest();
		}
		if(agent.getTimeStep()<stuckTime){
			if(agent.isStuck(agent.getMeAsHuman())){
				if(agent.commandHistory.get(agent.getTimeStep() - 1)!=null&&
				agent.commandHistory.get(agent.getTimeStep() - 1).equals(StandardMessageURN.AK_CLEAR)){
					agent.selfState=PFState.STUCK;
				    Blockade target = agent.isLocateInBlockade(agent.getMeAsHuman());
					if (target != null) {
				//		System.out.println(agent.getTimeStep()+"   "+agent.getID()+"helpself--stuck"+target.getID());

						agent.clearTools.clearBlockadeID(target);
					}else{
						if(agent.lastPFTarget!=null&&!agent.lastPFTarget.isDone()){
							agent.lastPFTarget.setDone();
							agent.updatePFTarget();
						}
						agent.randomWalk();
					}			
				}				
			}
		}
		
		if(isBlocked()){
			agent.selfState=PFState.STUCK;
			if(agent.location() instanceof Building){
			//	System.out.println(agent.getTimeStep()+"   "+agent.getID()+"escapebuilding");
				agent.randomWalk();  
				//escapeBuilding();
			}
			Blockade target = agent.clearTools.getNearestBlockade();	
			if (target != null) {
			//	System.out.println(agent.getTimeStep()+"   "+agent.getID()+"helpself--block"+target.getID());
				agent.randomWalk();  
				//agent.clearTools.clearBlockadeID(target);
			}else{
				if(agent.lastPFTarget!=null&&!agent.lastPFTarget.isDone()){
					agent.lastPFTarget.setDone();
					agent.updatePFTarget();
				}
			}
			agent.randomWalk();  
		}
		if(agent.location() instanceof Building&&!(agent.location() instanceof Refuge)){
			Building building = (Building) agent.location();
			if ((building.isFierynessDefined() && building.isOnFire())
					|| (building.isFierynessDefined() && building.getFieryness() != 8
							&& building.isTemperatureDefined() && building.getTemperature() > 35)) {
				if(agent.lastPFTarget!=null&&agent.lastPFTarget.getID()!=null){

						if(agent.lastPFTarget.isDone()){
							agent.updatePFTarget();
						}else{
							agent.lastPFTarget.setDone();
							agent.updatePFTarget();
						}
				}
				moveToRefuge();
				//escapeBuilding();
			}
		}
		if(isWillDead()){
			Collection<StandardEntity> refuges = world.getEntitiesOfType(StandardEntityURN.REFUGE);		
			if (refuges.isEmpty()) {
				return;	
			}else{
				if(agent.lastPFTarget!=null&&agent.lastPFTarget.getID()!=null){
					if(agent.lastPFTarget.getTaskRank()==PFTaskRank.SELFHELP){
						if(agent.lastPFTarget.isDone()){
							agent.sendRest();
						}else{
							agent.sendClearPFTarget(agent.lastPFTarget);
							agent.sendRest();
						}
					}
				}else{
					if(agent.location() instanceof Refuge){
						agent.sendRest();
					}else{
						List<StandardEntity> refuges1=new ArrayList<StandardEntity>();
						for(StandardEntity next:refuges){
							refuges1.add(next);
						}
						Comparator<StandardEntity> comparator=new DistanceComparator(agent.getMeAsHuman(),world);
						Collections.sort(refuges1, comparator);
					
						for(StandardEntity next:refuges1){
							PFTarget target=new PFTarget(next.getID(),SEUPFTargetType.SELF_HELP);	
							target.setTaskRank(PFTaskRank.SELFHELP);
							agent.sendClearPFTarget(target);
							agent.sendRest();
						}
					}		
				}										
			}			
		}
	}
	
	public boolean isWillDead(){
		if (!agent.getMeAsHuman().isHPDefined() || agent.getMeAsHuman().getHP() <= 1000) {
			agent.selfState=PFState.HURT;
			agent.partitionLastTaskType = PFPartitionLastTaskEnum.CANNOT_TO_CLEAR;
			return true;
		}
		if (agent.getMeAsHuman().getHP() - agent.getMeAsHuman().getDamage() * (agent.totalTime - agent.getTimeStep()) < 16) {
			agent.selfState=PFState.HURT;
			return true;
		}
		
		return false;
	}
	
	public boolean isBlocked() {
		if (agent.getTimeStep() < world.getConfig().ignoreTime + 3) return false;	
		if (agent.commandHistory.get(agent.getTimeStep() - 1)!=null&&
				agent.commandHistory.get(agent.getTimeStep() - 1).equals(StandardMessageURN.AK_MOVE) &&
				agent.commandHistory.get(agent.getTimeStep() - 2)!=null&&
				agent.commandHistory.get(agent.getTimeStep() - 2).equals(StandardMessageURN.AK_MOVE)&&
				agent.commandHistory.get(agent.getTimeStep() - 3)!=null&&
				agent.commandHistory.get(agent.getTimeStep() - 3).equals(StandardMessageURN.AK_MOVE)) {
			Pair<Integer, Integer> location_0 = agent.locationHistory.get(agent.getTimeStep() - 3);
			Pair<Integer, Integer> location_1 = agent.locationHistory.get(agent.getTimeStep() - 2);
			Pair<Integer, Integer> location_2 = agent.locationHistory.get(agent.getTimeStep() - 1);
			Pair<Integer, Integer> location_3 = agent.locationHistory.get(agent.getTimeStep());		
			if (location_0==null||location_1 == null || location_2 == null || location_3 == null) return false;	
			double distance_0 = Distance.getDistance(location_0, location_1);
			double distance_1 = Distance.getDistance(location_1, location_2);
			double distance_2 = Distance.getDistance(location_2, location_3);	
			if (distance_0<5000&&distance_1 < 5000 && distance_2 < 5000) {
				return true;
			}
		}	
		return false;
	}
	
	public boolean isCantWork(){
		if(agent.getMeAsHuman().isHPDefined()&&agent.getMeAsHuman().getHP()<=10){
			agent.selfState=PFState.DEAD;
			agent.partitionLastTaskType = PFPartitionLastTaskEnum.CANNOT_TO_CLEAR;
			return true;
		}
		if(agent.getMeAsHuman().isBuriednessDefined()&&agent.getMeAsHuman().getBuriedness()>0){
			agent.selfState=PFState.BURIED;
			return true;
		}
		return false;		
	}
	
	public void escapeBuilding() throws ActionCommandException{
		Building selfBuilding=(Building)agent.location();
		Point2D agentPoint=GeomUtil.getPoint(world.getSelfLocation());
		
		List<Road> entrances=world.getEntrance(selfBuilding);
		Road targetRoad=null;
		boolean isFind=false;
		if(entrances!=null&&!entrances.isEmpty()){
			for(Road road:entrances){
				List<EntityID> neighbors=road.getNeighbours();
				for(EntityID next:neighbors){
					StandardEntity s=world.getEntity(next);
					if(s instanceof Road){
						Road r1=(Road)s;
						if(!world.isEntrance(r1)){
							targetRoad=road;
							isFind=true;
							break;
						}						
					}
				}
				if(isFind){
					break;
				}
			}		
		}
		if(agent.lastPFTarget!=null&&!agent.lastPFTarget.isDone()){
			agent.lastPFTarget.setDone();
			agent.updatePFTarget();
		}
		if(targetRoad!=null){
			List<EntityID> path=agent.getPathToAreaInNormal(targetRoad);
			if(path!=null&&path.size()>1){
			//	System.out.println(agent.getTimeStep()+"   "+agent.getID()+"escape"+targetRoad.getID());
               if(path.size()==2){
            	   agent.sendMove(path);
               }else{
				agent.clearTools.clearPathInProcess(path);
				}
				//agent.clearTools.leaveBuilding(path, selfBuilding);
			}else{
				agent.randomWalk();
			}
		}else{
			agent.randomWalk();
		}
		
	}
	public void moveToRefuge() throws ActionCommandException{
		Collection<StandardEntity> refuges = world.getEntitiesOfType(StandardEntityURN.REFUGE);
		if (refuges.isEmpty()) 
			return;
		List<EntityID>path=new ArrayList<>();
		CostFunction costFunc = agent.router.getNormalCostFunction();
		Point selfL = new Point(agent.getMeAsHuman().getX(),agent.getMeAsHuman().getY());
		path =agent.router.getMultiAStar(agent.location(), refuges, costFunc, selfL);
		
		
		
		agent.sendMove(path);
		
	}
}