package seu.agent.pf.Tasks;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import seu.common.BasicTask;
import seu.model.SEUWorldModel;
import seu.model.object.SEUBuilding;
import seu.standard.DistanceComparator;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
import seu.agent.pf.PoliceForceAgent;
import seu.agent.SEUAgent.ActionCommandException;



public class SearchTask extends BasicTask<PoliceForceAgent>{

	SEUWorldModel world;
	
	public SearchTask(PoliceForceAgent agent) {
		super(agent);
		world=agent.getModel();
	
	}


	@Override
	public void execute() throws ActionCommandException {
		//agent.finishCurrent();
		checkEntranceInPartition();
		agent.expandCluster();
		randomWalk();
	}
	
	private void checkEntranceInPartition() throws ActionCommandException {
		if(agent.lastPFTarget!=null&&agent.lastPFTarget.getID()!=null){			
			if(agent.lastPFTarget.getTaskRank()==PFTaskRank.EntranceInPartition){
				if(!agent.lastPFTarget.isDone()){
					StandardEntity target=world.getEntity(agent.lastPFTarget.getID());
					if(agent.needCheckEntranceSet.contains(target)){
					agent.sendClearPFTarget(agent.lastPFTarget);
					}else{
						agent.lastPFTarget.setDone();
					
						}
				}
			}
		}
		double count1=0;
		double count2=0;
		for (SEUBuilding seuBuilding : agent.currentPartition.getBuildingList()) {
			Building bu = seuBuilding.getSelfBuilding();
			if (!bu.isFierynessDefined())
				continue;
			if (bu.getFieryness() == 8 || bu.getFieryness() == 7 || bu.getFieryness() == 3) {
				count2++;
				agent.needCheckEntranceSet.removeAll(world.getEntrance(bu));
			}
			count1++;
		}
		if(count2/count1>0.8){
			agent.needCheckEntranceSet.clear();
		}
		if(agent.needCheckEntranceSet!=null&&!agent.needCheckEntranceSet.isEmpty()){
			Comparator<StandardEntity> comparator=new DistanceComparator(agent.getMeAsHuman(),world);
			SortedSet<StandardEntity> sort = new TreeSet<StandardEntity>(comparator);
			sort.addAll(agent.needCheckEntranceSet);
			List<Road> needCheckEntranceSet1=new ArrayList<Road>();
			for(StandardEntity next:sort){
				needCheckEntranceSet1.add((Road)next);
			}
	
			PFTarget target=new PFTarget(needCheckEntranceSet1.get(0).getID(),SEUPFTargetType.BLOCKADE_ENTRANCE);
			target.setTaskRank(PFTaskRank.EntranceInPartition);
			agent.lastPFTarget=target;
	//		System.out.println(agent.getTimeStep()+"   "+agent.getID()+"search"+target.getID());
			agent.sendClearPFTarget(target);
			
		}		
	}
	

	public void randomWalk() throws ActionCommandException {		
		if(agent.lastPFTarget!=null&&agent.lastPFTarget.getID()!=null){			
			if(agent.lastPFTarget.getTaskRank()==PFTaskRank.RANDOMWALK){
				if(!agent.lastPFTarget.isDone()){
					agent.sendClearPFTarget(agent.lastPFTarget);
				}
			}
		}
		Collection<StandardEntity> inRnage = world.getObjectsInRange( agent.getMeAsHuman().getID(), 50000);
		List<Area> target = new ArrayList<Area>();
		Collection<Area> blockadeDefinedArea = new ArrayList<Area>();
		
		for (StandardEntity entity : inRnage) {
			if (!(entity instanceof Area))
				continue;
			if (entity.getID().getValue() == agent.location().getID().getValue())
				continue;
			Area nearArea = (Area) entity;
			if (nearArea.isBlockadesDefined() && nearArea.getBlockades().size() > 0) {
				blockadeDefinedArea.add(nearArea);
				continue;
			}
				
			target.add(nearArea);
		}
		
		if (target.isEmpty())
			target.addAll(blockadeDefinedArea);
		Area randomTarget = target.get( agent.getRandom().nextInt(target.size()));
		PFTarget pftarget=new PFTarget(randomTarget.getID(),SEUPFTargetType.RANDOMWALK);
		pftarget.setTaskRank(PFTaskRank.RANDOMWALK);
		agent.lastPFTarget=pftarget;
//		System.out.println(agent.getTimeStep()+"   "+agent.getID()+"random"+randomTarget.getID());

		agent.sendClearPFTarget(pftarget);						
	}

}
