package seu.model;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import javolution.util.FastMap;
import javolution.util.FastSet;
import rescuecore2.config.Config;
import rescuecore2.misc.Pair;
import rescuecore2.registry.Registry;
import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.GasStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.Viewer.layers.SEU_BuildingLayer;
import seu.Viewer.layers.SEU_RoadLayer;
import seu.agent.SEUAgent;
import seu.agent.SEUCenter;
import seu.agent.fb.FireBrigadestools.toolsForOldExtinguish.ClusterManager;
import seu.agent.pf.cluster.Clustering;
//import seu.agent.fb.FireBrigadestools.ClusterManager;
import seu.common.TimeOutException;
import seu.communication.CivilianVoiceListener;
import seu.geom.CompositeConvexHull;
import seu.geom.Distance;
import seu.geom.PolygonScaler;
import seu.model.flow.EnergyFlow;
import seu.model.object.SEUBuilding;
import seu.model.object.SEUWall;
import seu.model.object.SEURoad;
import seu.model.object.seuZoneEntity.SeuZoneFactory;
import seu.model.object.seuZoneEntity.SeuZones;
import seu.model.route.pov.POVRouter;
import seu.standard.simplePartition.GroupingType;
import seu.standard.simplePartition.Line;
import seu.util.EntityIdComparator;


public class SEUWorldModel extends StandardWorldModel {
	public static final double MEAN_VELOCITY_OF_MOVING = 31445.392;
	protected long thinkStartTime;
	protected String exceptionMessage = "";
	
	// owner agent part
	public Human me = null;					// for platoon agent
	public SEUAgent<Human> selfHuman;
	public Building building;				// for center agent
	public SEUCenter centre;
	public EntityID selfId;
	public SEUAgent<Human> agent=null;
	// message handlers
	protected BurningBuildings burningBuildings = null;
	protected ExtinguishedBuilding extinguishBuildings = null;
	protected CollapsedBuildings collapsedBuildings = null;
//	protected StuckHandler stuckHandler = null;
	protected BuriedHumans buriedHumans = null;
	protected RemainCluster remainCluster = null;
	protected SearchedBuildings searchedBuildings = null;
	protected CriticalArea criticalArea = null;
	protected WaterPort waterPort = null;
	
	// map informs
	protected Set<StandardEntity> mapBorderBuildings;
	protected Dimension mapDimension = null;
	protected double mapWidth;
	protected double mapHeight;
	protected double mapDiameter;
	protected boolean isMapHuge = false;
	protected boolean isMapMedium = false;
	protected boolean isMapSmall = false;
	protected Area mapCenter;
	
	// communication conditions
	protected boolean communicationLess = true;
	protected boolean communicationLow = false;
	protected boolean communicationMedium = false;
	protected boolean communicationHigh = false;
	
	protected boolean isNoRadio = false;
	
	// agent informs
	protected List<FireBrigade> fireBrigadeList = new ArrayList<FireBrigade>();
	protected List<PoliceForce> policeForceList = new ArrayList<PoliceForce>();
	protected List<AmbulanceTeam> ambulanceTeamList = new ArrayList<AmbulanceTeam>();
	protected List<FireStation> fireStationList = new ArrayList<FireStation>();
	protected List<PoliceOffice> policeOfficeList = new ArrayList<PoliceOffice>();
	protected List<AmbulanceCentre> ambulanceCenterList = new ArrayList<AmbulanceCentre>();
	protected List<EntityID> fireBrigadeIdList = new ArrayList<EntityID>();
	protected List<EntityID> policeForceIdList = new ArrayList<EntityID>();
	protected List<EntityID> ambulanceTeamIdList = new ArrayList<EntityID>();
	protected List<EntityID> fireStationIdList = new ArrayList<EntityID>();
	protected List<EntityID> policeOfficeIdList = new ArrayList<EntityID>();
	protected List<EntityID> ambulanceCenterIdList = new ArrayList<EntityID>();
	protected Set<EntityID> stuckedAgentList = new HashSet<EntityID>();
	
	// SEUBuilding and SEURoad part
	protected Map<EntityID, SEUBuilding> seuBuildingMap;
	protected Map<EntityID, SEURoad> seuRoadMap;
	protected Set<SEURoad> seuRoads;
	protected Map<String, Building> buildingXYMap = new FastMap<String, Building>();
	
	protected Set<Road> allEntrance;
	protected Map<Road, List<Building>>  roadToBuildingMap ;///
	protected Map<Building, List<Road>>  buildingToRoadMap ;///
	// others
	// protected RegionModel regionModel;
	// protected Exploration<Building> buildingExploration;
	// protected Set<BuildingInfo> buildingInfos;
	protected int time;
	protected ConfigConstants config = null;
	protected POVRouter router = null;
	protected Entrance entrance = null;
	protected Uniform uniform = null;
	protected TimeStamp timestamp = null;
	protected EnergyFlow energyFlow = null;
	protected IsolatedArea isolated = null;
	protected Set<EntityID> visibleEntity = null;
	protected SeuZones zones = null;
	
	//NIU SEU MESSAGE
	protected List<EntityID> visibleATs = new ArrayList<EntityID>();
	protected List<EntityID> visibleHealthyATs = new ArrayList<EntityID>();
	protected List<EntityID> visibleFBs = new ArrayList<EntityID>();
	protected List<EntityID> visiblePFs = new ArrayList<EntityID>();
	protected List<EntityID> visibleAgents = new ArrayList<EntityID>();
	protected List<EntityID> visibleHumansToLoad = new ArrayList<EntityID>();
	protected List<EntityID> newDiedHuman=new ArrayList<EntityID>();
	protected List<EntityID> visualBuriedHuman=new ArrayList<EntityID>();
	protected List<Building> visualBurningBuliding = new ArrayList<Building>();
	protected List<Building> visualBurnOutBuilding = new ArrayList<Building>();
	protected List<Building> visualExtinguishedBuilding = new ArrayList<Building>();
	protected List<Blockade> visualBlockade = new ArrayList<Blockade>();
	protected Set<EntityID> RescuedorDiedHuman=new HashSet<EntityID>();
	protected Set<Building> BurningBuliding=new HashSet<Building>();
	protected Set<Building> CollapsedBuildings=new HashSet<Building>();
	protected Set<Building> extinguishedBuilding=new HashSet<Building>();
	private List<EntityID> firstBuriedAgents=new ArrayList<EntityID>();
	private Clustering partitionModel;
	private ClusterManager clusterManager;
	List<EntityID> team=new ArrayList<EntityID>();

/* ----------------------------------- constructor and initialize ----------------------------------------- */
	public SEUWorldModel() {
		super();
	}
	
	@SuppressWarnings("unchecked")
	public void initialize(SEUAgent<? extends StandardEntity> selfAgent, Config conf, GroupingType type) {
		// others
		this.config = new ConfigConstants(conf, this);
		this.uniform = new Uniform(this);
		this.timestamp = new TimeStamp(this);
		this.router = new POVRouter(this);
		this.energyFlow = new EnergyFlow(this);
        // this.regionModel = new RegionModel(this, type);
        // this.buildingExploration = new Exploration<Building>(this);
		
        // handle owner agent informs
		this.selfId = selfAgent.getID();
		
		// create SEUBuildings and SEURoads
		this.createSeuBuildings();
		this.createSeuRoads();
		
		if (selfAgent instanceof SEUCenter) {
			this.centre = (SEUCenter)selfAgent;
			this.building = (Building) getEntity(selfAgent.getID());
		} else if (selfAgent instanceof SEUAgent) {
			this.agent = (SEUAgent)selfAgent;
			this.selfHuman = (SEUAgent)selfAgent;
			this.me = (Human)getEntity(selfAgent.getID());
			this.isolated = new IsolatedArea(this.me, this);
			this.entrance = new Entrance(this);
			initializeEntrance();
			updateAgentStates();
		}
		
		
		
		// message handlers
		this.burningBuildings = new BurningBuildings(this);
		this.collapsedBuildings = new CollapsedBuildings(this);
		this.extinguishBuildings = new ExtinguishedBuilding(this);
		this.buriedHumans = new BuriedHumans(this);
		this.waterPort = new WaterPort(this);
//		this.stuckHandler = new StuckHandler(this, 80, 6);
		this.criticalArea = new CriticalArea(this);
		this.searchedBuildings = new SearchedBuildings(this);
		this.remainCluster = new RemainCluster(this);
		
		// handle map informs
		this.mapBorderBuildings = this.findBorderBuilding(0.9, this);
		this.mapDimension = this.calculateMapDimension();
		this.verifyMap();
        
		// get all agents
		for (StandardEntity next : getEntitiesOfType(AgentConstants.PLATOONS)) {
			if (next instanceof FireBrigade) {
				fireBrigadeList.add((FireBrigade) next);
				fireBrigadeIdList.add(next.getID());
			} else if (next instanceof PoliceForce) {
				policeForceList.add((PoliceForce) next);
				policeForceIdList.add(next.getID());
			} else if (next instanceof AmbulanceTeam) {
				ambulanceTeamList.add((AmbulanceTeam) next);
				ambulanceTeamIdList.add(next.getID());
			}
		}
		for (StandardEntity next : getEntitiesOfType(AgentConstants.CENTRES)) {
			if (next instanceof FireStation) {
				fireStationList.add((FireStation) next);
				fireStationIdList.add(next.getID());
			} else if (next instanceof PoliceOffice) {
				policeOfficeList.add((PoliceOffice) next);
				policeOfficeIdList.add(next.getID());
			} else if (next instanceof AmbulanceCentre) {
				ambulanceCenterList.add((AmbulanceCentre) next);
				ambulanceCenterIdList.add(next.getID());
			}
		}
		if(getEntity(selfId) instanceof PoliceForce || getEntity(selfId) instanceof FireBrigade){
			partitionModel=new Clustering();
		}
        // SEUZone part
		if (me instanceof Human) {
			SeuZoneFactory zoneFactory = new SeuZoneFactory(this);
			zones = zoneFactory.createZones();
		}
		
		
	}
	
	
/* ----------------------------------------- update and merge --------------------------------------------- */
	/** Update world model, and be invoked in Agent's BeforeAct() method.*/
	public void update(Human me, ChangeSet changed) throws TimeOutException{
		
	
		this.visibleEntity = changed.getChangedEntities();
		this.criticalArea.update(router);
		this.burningBuildings.update(this, changed);
		this.extinguishBuildings.update(this, changed);
		this.waterPort.update(this);
		this.collapsedBuildings.update(this, changed);
		this.buriedHumans.update(changed);
		this.energyFlow.update(changed);
		this.searchedBuildings.update(this, me);
		
		this.burningBuildings.remove();
		this.extinguishBuildings.remove();
		this.collapsedBuildings.remove();
		
		
		visibleAgents.clear();
		visibleHumansToLoad.clear();
		visibleATs.clear();
		visibleHealthyATs.clear();
		visiblePFs.clear();
		visibleFBs.clear();
		visualBuriedHuman.clear();
		newDiedHuman.clear();
		visualExtinguishedBuilding.clear();
		visualBurnOutBuilding.clear();
		visualBurningBuliding.clear();
		visualBlockade.clear();
		
		
		
		if(agent!=null){
			
			for (StandardEntity se : getEntitiesOfType(AgentConstants.HUMANOIDS)) {
				Human hm = (Human) se;
				if (me.getPosition(this).equals(hm.getPosition(this))
						&& !changed.getChangedEntities().contains(hm.getID())) {
					hm.undefinePosition();
				}
			}
			
			for (StandardEntity se : getEntitiesOfType(StandardEntityURN.BLOCKADE)) {
				Blockade block = (Blockade) se;
				if (me.getPosition().equals(block.getPosition())
						&& !changed.getChangedEntities().contains(block.getID())) {
					block.undefinePosition();
				}
			}
		
		
		}
		
		
	for (EntityID next : changed.getChangedEntities()) {
		 if(next.equals(selfId))
				continue;	///////////////////排除自己
		
		 StandardEntity entity = getEntity(next);

			if (entity instanceof Road) {
				SEURoad road = getSeuRoad(next);
				road.update();
			}
			
			if (entity instanceof AmbulanceTeam || entity instanceof FireBrigade) {
				if (getAgent().isStucked((Human)entity))
					stuckedAgentList.add(next);
				else
					stuckedAgentList.remove(next);
			}
		
		if(entity instanceof Blockade){
			Blockade block = (Blockade) entity;
			visualBlockade.add(block);
		}
		
   if(entity instanceof Human){
		 
     if(entity instanceof AmbulanceTeam ){ 
			 visibleAgents.add(next);
			 visibleATs.add(next); 
			 if(((AmbulanceTeam) entity).getBuriedness()<=0)
				 visibleHealthyATs.add(next);
			 if (getAgent().isStuck((Human)entity))
					stuckedAgentList.add(next);
				else
					stuckedAgentList.remove(next);
			 
		 }
			
		 if(entity instanceof FireBrigade ){
			 
			 visibleAgents.add(next);
			 visibleFBs.add(next); 
			 if (getAgent().isStuck((Human)entity))
					stuckedAgentList.add(next);
				else
					stuckedAgentList.remove(next);
			 
		 }
		 if(entity instanceof PoliceForce ){
			 visibleAgents.add(next);
			 visiblePFs.add(next);
		 }
		 
		   Human human = (Human)entity;
			if(human.isHPDefined()&&human.getHP()>0&&human.getHP()<10000){
				if(human.isBuriednessDefined()&&human.getBuriedness()==0){
					if(human.isDamageDefined()&&human.getDamage()>0){
					if(human.getStandardURN()==StandardEntityURN.CIVILIAN){
						visibleHumansToLoad.add(next);
					  }
				
					}
				}
				
			}
			
			
			
			if (human.getBuriedness() > 0){
				visualBuriedHuman.add(next);
			}
			if (human instanceof Civilian&&human.isDamageDefined()&&human.getDamage()>0){
				visualBuriedHuman.add(next);
			}
			if (RescuedorDiedHuman.contains(human.getID()))
			
			if (human.getHP() == 0&&!RescuedorDiedHuman.contains(human.getID())) {
				newDiedHuman.add(human.getID());
				RescuedorDiedHuman.add(human.getID());
		        }
			   StandardEntity position = human.getPosition(this);
			  if (position instanceof Refuge){
				RescuedorDiedHuman.add(human.getID());
			}
			RescuedorDiedHuman.add(human.getID());
		}

	if (entity instanceof Building) {
			Building building = (Building) entity;
			
			if (building.isFierynessDefined() && building.getFieryness() > 3 && building.getFieryness() < 8){
				extinguishedBuilding.add(building);
					visualExtinguishedBuilding.add(building);
			} 
	  
			if (building.isFierynessDefined() && building.getFierynessEnum() == Fieryness.BURNT_OUT) {
				CollapsedBuildings.add(building);
					visualBurnOutBuilding.add(building);
			}
			
			if (building.isOnFire()) {
				
				if (getTime() < getConfig().ignoreTime + 1) {
					BurningBuliding.add(building);
					visualBurningBuliding.add(building);
				} else {
					BurningBuliding.add(building); 
						visualBurningBuliding.add(building);
					
				}
			} else BurningBuliding.remove(building);
		
		}

			
		}
	
	
	}
	
	/**
	 * The merge() method do the following things:
	 * 
	 * <pre>
	 * First, get all entities I can see from ChangeSet. For each visible entity, if it's
	 * not in the collection of entities I can see last cycle, add it. If it is in, update its 
	 * property.
	 * 
	 * Second, delete those entities I can see this cycle from the collection of entities
	 * I can see last cycle.
	 * 
	 * Third, update human rectangles.
	 * </pre>
	 */
	@Override
	public void merge(ChangeSet changeSet) {
		
		List<EntityID> changedEntityID = new ArrayList<EntityID>();
		for (EntityID id : changeSet.getChangedEntities()) {
			if (getEntity(id) == null) {
				changedEntityID.add(id);
			}
		}
		
		super.merge(changeSet);
		if (timestamp != null) {
			timestamp.merge(changeSet);
		}
		for (EntityID id : changedEntityID) {
			StandardEntity se = getEntity(id);
			if (se instanceof Civilian) {
				timestamp.addStateChangeListener(se);
			}
		}
		
		for (EntityID entityId : changeSet.getChangedEntities()) {
			try {
				String entityUrn = changeSet.getEntityURN(entityId);
				if (isBuilding(entityUrn)) {
					Building building = getEntity(entityId, Building.class);
					if (building == null) {
						building = (Building) Registry.getCurrentRegistry().createEntity(entityUrn, entityId);
						this.addEntity(building);
					}
					if (this.me != null && this.me instanceof FireBrigade) {
						updateBuildingFuelForFireBrigade(building);
					}
					 getSeuBuilding(building).setSensed(getTime());//niu
					 if (building.isOnFire()) {
						 getSeuBuilding(building).setIgnitionTime(getTime());
	                  }
				}
			} catch (NullPointerException e) {
				System.out.println("NullPointException in SEUWorldModel's merge " +
						"method in time: " + time);
				e.printStackTrace();
			} catch (ClassCastException ex) {
				System.out.println("CLassCastException in SEUWorldModel's merge method " +
						"in time: " + time);
				ex.printStackTrace();
			}
		}
	}
	
	/**
	 * Determines whether an entity in a building.
	 * 
	 * @param entityUrn
	 *            the URN of this entity
	 * @return true when this entity is a building. Otherwise, false.
	 */
	private boolean isBuilding(String entityUrn) {
		return entityUrn.equals(StandardEntityURN.BUILDING.toString())
				|| entityUrn.equals(StandardEntityURN.REFUGE.toString())
				|| entityUrn.equals(StandardEntityURN.AMBULANCE_CENTRE.toString())
				|| entityUrn.equals(StandardEntityURN.FIRE_STATION.toString())
				|| entityUrn.equals(StandardEntityURN.POLICE_OFFICE.toString())
				|| entityUrn.equals(StandardEntityURN.GAS_STATION.toString());
	}
	
	/**
	 * Update the building fuel and energy for the given building for
	 * FireBrigade Agent.
	 * 
	 * @param building
	 *            the target building currently within the eyeside of this
	 *            FireBrigade Agent
	 */
	private void updateBuildingFuelForFireBrigade(Building building) {
		SEUBuilding seuBuilding = this.getSeuBuilding(building.getID());
		seuBuilding.setVisible(true);
		if (building.isFierynessDefined() && building.isTemperatureDefined()) {
			int temperature = building.getTemperature();
			seuBuilding.setEnergy(temperature * seuBuilding.getCapacity(), "update changeSet");
			switch (building.getFieryness()) {
			case 0:
				seuBuilding.setFuel(seuBuilding.getInitialFuel());
				if (seuBuilding.getEstimatedTemperature() >= seuBuilding.getIgnitionPoint()) {
					seuBuilding.setEnergy(seuBuilding.getIgnitionPoint() / 2, "update changeSet case 0");
				}
				break;
			case 1:
				if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.66) {
					seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.75));
				} else if (seuBuilding.getFuel() == seuBuilding.getInitialFuel()) {
					seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.90));
				}
				break;
			case 2:
				if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.33
						|| seuBuilding.getFuel() > seuBuilding.getInitialFuel() * 0.66) {
					seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.50));
				}
				break;
			case 3:
				if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.01
						|| seuBuilding.getFuel() > seuBuilding.getInitialFuel() * 0.33) {
					seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.15));
				}
				break;
			case 8:
				seuBuilding.setFuel(0);
				break;
			default:
				break;
			}
		}
	}
	
	
/* ----------------------------------- getEntitiesOfType and getEntity ------------------------------------- */

	/**
	 * Returns a collection of entity specified by <b>StandardEntityURN urn<b>.
	 * ANd then cast those entity to the specified type <b>&lt;T extends
	 * StandardEntity&gt;</b>
	 */
	public <T extends StandardEntity> List<T> getEntitiesOfType(Class<T> c, StandardEntityURN urn) {
		Collection<StandardEntity> entities = getEntitiesOfType(urn);
		List<T> list = new ArrayList<T>();
		for (StandardEntity entity : entities) {
			if (c.isInstance(entity)) {
				list.add(c.cast(entity));
			}
		}
		return list;
	}

	/** Get all entities of specified types stores in this world model. */
	public Collection<StandardEntity> getEntitiesOfType(EnumSet<StandardEntityURN> urns) {
		Collection<StandardEntity> res = new HashSet<StandardEntity>();
		for (StandardEntityURN urn : urns) {
			res.addAll(getEntitiesOfType(urn));
		}
		return res;
	}

	/**
	 * Get an object of Entity according to its ID and cast this object to
	 * <b>&lt;T extends StandardEntity&gt;</b>.
	 */
	public <T extends StandardEntity> T getEntity(EntityID id, Class<T> c) {
		StandardEntity entity;
		entity = getEntity(id);
		if (c.isInstance(entity)) {
			T castedEntity = c.cast(entity);

			return castedEntity;
		} else {
			return null;
		}
	}

	public SEUBuilding getSeuBuilding(StandardEntity entity) {
		return this.seuBuildingMap.get(entity.getID());
	}

	public SEUBuilding getSeuBuilding(EntityID entityId) {
		return this.seuBuildingMap.get(entityId);
	}

	public SEURoad getSeuRoad(StandardEntity entity) {
		return this.seuRoadMap.get(entity.getID());
	}

	public SEURoad getSeuRoad(EntityID entityId) {
		return this.seuRoadMap.get(entityId);
	}

	public Collection<SEUBuilding> getSeuBuildings() {
		return this.seuBuildingMap.values();
	}
	
	public Map<EntityID, SEUBuilding> getSeuBuildingMap() {
		return seuBuildingMap;
	}
	
    public Set<SEURoad> getSeuRoads() {
    	return this.seuRoads;
    }
    
    
/* ---------------------------------------------------------------------------------------------------------- */
    /**
     * This method used to find all border building of this world.
     * 
     * @return a set of border building of this world
     */
    private Set<StandardEntity> findBorderBuilding (double scale, SEUWorldModel world) {
    	CompositeConvexHull convexHull = new CompositeConvexHull();
    	Set<StandardEntity> allEntities = new FastSet<StandardEntity>();
    	for (StandardEntity entity : world) {
			if (entity instanceof Building) {
				allEntities.add(entity);
				Pair<Integer, Integer> location = entity.getLocation(world);
				convexHull.addPoint(location.first(), location.second());
			}
    	}
  
    	Set<StandardEntity> borderBuilding = 
    			PolygonScaler.getMapBorderBuildings(convexHull, allEntities, scale, world);
    	return borderBuilding;
    }
    
//    /**
//     * Remove the map border buildings from all building set to save the computing time.
//     */
//    private void removeBorderBuilding() {
//    	for (StandardEntity next : borderBuildings) {
//    		if (next instanceof Building) {
//    			seuBuildingsCopy.remove(this.getSeuBuilding(next.getID()));
//    		}
//    	}
//    }
    
	/**
	 * Calculate the dimension of current map. The result dimension is slightly
	 * smaller than the real dimention. But it can save lots of computing time.
	 */
    private Dimension calculateMapDimension() {
    	int minX = Integer.MAX_VALUE;
    	int maxX = Integer.MIN_VALUE;
    	int minY = Integer.MAX_VALUE;
    	int maxY = Integer.MIN_VALUE;
        Pair<Integer, Integer> pos;
        for (StandardEntity standardEntity : this.getAllEntities()) {
            pos = standardEntity.getLocation(this);
            if (pos.first() < minX)
                minX = pos.first();
            if (pos.second() < minY)
                minY = pos.second();
            if (pos.first() > maxX)
                maxX = pos.first();
            if (pos.second() > maxY)
                maxY = pos.second();
        }
        Dimension mapDimension = new Dimension(maxX - minX, maxY - minY);
        this.mapWidth = mapDimension.getWidth();
        this.mapHeight = mapDimension.getHeight();
        
        return mapDimension;
    }
    
    /**
     * Thie method determines whether current map is huge, medium or small.
     */
    private void verifyMap() {
    	double mapWidth = this.getMapDimension().getWidth();
    	double mapHeight = this.getMapDimension().getHeight();
    	double mapDiagonalLength = Math.hypot(mapWidth, mapHeight);
    	double rate = mapDiagonalLength / MEAN_VELOCITY_OF_MOVING;
    	if (rate > 60) {
    		this.isMapHuge = true;
    	} else if (rate > 30) {
    		this.isMapMedium = true;
    	} else {
    		this.isMapSmall = true;
    	}
    }
    
    private void calculateMapDiameter() {
    	this.mapDiameter = Math.sqrt(Math.pow(this.mapWidth, 2.0) + Math.pow(this.mapHeight, 2.0));
    }
    public double getMapDiameter() {
    	if (mapDiameter == 0)
    		this.calculateMapDiameter();
    	return this.mapDiameter;
    }
    
    public boolean isMapHuge() {
    	return this.isMapHuge;
    }
    public boolean isMapMedium() {
    	return this.isMapMedium;
    }
    public boolean isMapSmall() {
    	return this.isMapSmall;
    }
    
    /** Find the center area of current map.*/
    public Area getMapCenter() {
    	if (this.mapCenter != null) {
    		return this.mapCenter;
    	}
    	
    	double ret;
        int min_x = Integer.MAX_VALUE;
        int max_x = Integer.MIN_VALUE;
        int min_y = Integer.MAX_VALUE;
        int max_y = Integer.MIN_VALUE;

        Collection<StandardEntity> areas = getEntitiesOfType(AgentConstants.AREAS);

        long x = 0, y = 0;
        Area result;

        for (StandardEntity entity : areas) {
            Area area1 = (Area) entity;
            x += area1.getX();
            y += area1.getY();
        }

        x /= areas.size();
        y /= areas.size();
        result = (Area) areas.iterator().next();
        for (StandardEntity entity : areas) {
            Area temp = (Area) entity;
            double a = Distance.getDistance((int) x, (int) y, result.getX(), result.getY());
            double b = Distance.getDistance((int) x, (int) y, temp.getX(), temp.getY());
            if (a > b) {
                result = temp;
            }

            if (temp.getX() < min_x) {
                min_x = temp.getX();
            } else if (temp.getX() > max_x)
                max_x = temp.getX();

            if (temp.getY() < min_y) {
                min_y = temp.getY();
            } else if (temp.getY() > max_y)
                max_y = temp.getY();
        }
        ret = (Math.pow((min_x - max_x), 2) + Math.pow((min_y - max_y), 2));
        ret = Math.sqrt(ret);
        this.mapCenter = result;

        return result;
    }
    
    public Point getMapCenterPoint() {
    	return new Point((int)mapWidth >> 2, (int)mapHeight >> 2);
    }
    
    public List<Building> findNearBuildings(Building centerbuilding, int distance) {
        List<Building> result;
        Collection<StandardEntity> allObjects;
        int radius;

        Rectangle rect = centerbuilding.getShape().getBounds();
        radius = (int) (distance + rect.getWidth() + rect.getHeight());

        allObjects = getObjectsInRange(centerbuilding, radius);
        result = new ArrayList<Building>();
        for (StandardEntity next : allObjects) {
            if (next instanceof Building) {
                Building building;

                building = (Building) next;
                if (!building.equals(centerbuilding)) {
                    if (Distance.getDistance(centerbuilding, building) < distance) {
                        result.add(building);
                    }
                }
            }
        }
        return result;
    }
    
/* -------------------------------------------------------------------------------------------------------- */
    
    public int distance(java.awt.Point p1, java.awt.Point p2) {
        return distance(p1.x, p1.y, p2.x, p2.y);
    }

    public int distance(int x1, int y1, int x2, int y2) {
        double dx = x1 - x2;
        double dy = y1 - y2;
        return (int) Math.hypot(dx, dy);
    }
   
    public Line getLine(Area node1, Area node2) {
        int x1 = node1.getX();
        int y1 = node1.getY();
        int x2 = node2.getX();
        int y2 = node2.getY();
        return new Line(x1, y1, x2, y2);
    }
    
    public Area getPositionFromCoordinates(int x, int y) {
		for (StandardEntity entity : getObjectsInRange(x, y, 1000)) {
			if (entity instanceof Area) {
				Area area = (Area) entity;
				if (area.getShape().contains(x, y)) {
					return area;
				}
			}
		}
		return null;
	}
    
    public StandardEntity getPositionFromCoordinates(Pair<Integer, Integer> coordinate) {
		return getPositionFromCoordinates(coordinate.first(), coordinate.second());
	}
    
    public Area getNearestRoad(int x, int y) {
		Area result = getPositionFromCoordinates(x, y);
		if (result != null) 
			return result;
		
		int l = 1;
		int diameter = (int) Math.hypot(
				Math.max(config.MAX_X, x) - Math.min(config.MIN_X, x),
				Math.max(config.MAX_Y, y) - Math.min(config.MIN_Y, y));
		Collection<StandardEntity> minRange = null;
		while (diameter - l > 1000) {
			int m = (diameter + l) / 2;
			Collection<StandardEntity> range = getObjectsInRange(x, y, m);
			if (0 < range.size()) {
				diameter = m;
				if (minRange == null || minRange.size() < range.size()) {
					minRange = range;
				}
			} else {
				l = m;
			}
		}
		if (minRange == null) 
			return null;
		double minDist = Double.MAX_VALUE;
		for (StandardEntity se : minRange) {
			if (se instanceof Area) {
				Area area = (Area) se;
				double dist = Math.hypot(area.getX() - x, area.getY() - y);
				if (dist < minDist) {
					minDist = dist;
					result = area;
				}
			}
		}
		return result;
	}
    
/* -------------------------------------------------------------------------------------------------------- */
    
    public int getDeathTime(Human human) {
		if (human.isDamageDefined()) {
			if (human.getDamage() == 0) {
				return getConfig().timestep + 1;
			}
			else {
				int lastChangedTime = timestamp.getLastChangedTime(human.getID());
				if (lastChangedTime < 0) {
					lastChangedTime = time;
				}
				if (human.isHPDefined()) {
					return lastChangedTime + human.getHP() / human.getDamage() - 1;
				}
				else {
					return getConfig().timestep + 1;
				}
			}
		}
		return getConfig().timestep + 1;
	}
    
	/**
	 * Get the Agent this world model belongs to.
	 * 
	 * @return the Agent who own this world model
	 */
    public SEUAgent<? extends StandardEntity> getAgent() {
    	if (this.selfHuman != null)
    		return this.selfHuman;
    	if (this.centre != null)
    		return this.centre;
    	return null;
    }
    
	/**
	 * Get the controlled entity of this world model's owner Agent
	 * 
	 * @return the entity controlled by this world model's owner Agent
	 */
    public StandardEntity getControlledEntity() {
    	if (this.me != null)
    		return this.me;
    	if (this.building != null)
    		return this.building;
    	return null;
    }

//    public BuildingInfo getBuildingInfo(EntityID id) {
//		BuildingInfo info;
//
//		info = buildingInfoMap.get(id);
//		if (info == null) {
//			Building building;
//			building = getEntity(id, Building.class);
//			if (building != null) {
//				info = new BuildingInfo(building, null, time);
//				buildingInfoMap.put(id, info);
//			}
//		}
//		return info;
//	}
    
	public Set<EntityID> getVisibleEntitiesID() {
		return visibleEntity;
	}
	
	public boolean isVisible(EntityID id) {
		return getVisibleEntitiesID().contains(id);
	}
	
	protected boolean isVisible(StandardEntity entity) {
		return isVisible(entity.getID());
	}
	
    
    
/* -------------------------------------------------------------------------------------------------------- */
	public Set<StandardEntity> getNeighbours(StandardEntity e, int Ext){
		Set<StandardEntity> Neighbours = new HashSet<StandardEntity>();
		if (e instanceof Building){
			for(EntityID tmp2:((Building) e).getNeighbours()){
				Neighbours.add(getEntity(tmp2));
			}
			if(((Building) e).isEdgesDefined()){
				Ext*=1000;
				List<Edge> Edges = new ArrayList<Edge>();
				Edges=((Building) e).getEdges();
				Polygon ExtArea =new Polygon();
				Polygon baseArea = new Polygon();
				int n=1;
				Point2D.Double tmp1=new Point2D.Double();
				Point2D.Double tmp2=new Point2D.Double();
				Point2D.Double tmp3=new Point2D.Double();
				Point2D.Double tmp4=new Point2D.Double();
				Point2D.Double tmp5=new Point2D.Double();
				for(Edge Ee : Edges){baseArea.addPoint(Ee.getStartX(),Ee.getStartY());}
				for(Edge Ee : Edges)
				{
					for(Edge Ee1:Edges){
						if(Ee.getStart().equals(Ee1.getEnd())){
							tmp1.setLocation
							(
									Ee.getStartX()+(n*(Ee.getEndX()-Ee.getStartX()))/Math.hypot((double)Ee.getEndX()-(double)Ee.getStartX() , (double)Ee.getEndY()-(double)Ee.getStartY() )
									,
									Ee.getStartY()+(n*(Ee.getEndY()-Ee.getStartY()))/Math.hypot((double)Ee.getEndX()-(double)Ee.getStartX() , (double)Ee.getEndY()-(double)Ee.getStartY() )
							);
							tmp2.setLocation
							(
									Ee1.getEndX()+(n*(Ee1.getStartX()-Ee1.getEndX()))/Math.hypot((double)Ee1.getStartX()-(double)Ee1.getEndX() , (double)Ee1.getStartY()-(double)Ee1.getEndY() )
									,
									Ee1.getEndY()+(n*(Ee1.getStartY()-Ee1.getEndY()))/Math.hypot((double)Ee1.getStartX()-(double)Ee1.getEndX() , (double)Ee1.getStartY()-(double)Ee1.getEndY() )
							);

							tmp3.setLocation
							(
									(tmp1.x+tmp2.x)/2
									,
									(tmp1.y+tmp2.y)/2
							);
							tmp4.setLocation(Ee.getStartX(), Ee.getStartY());
							if(tmp3.x==tmp4.x&&tmp3.y==tmp4.y){continue;}
							else if(baseArea.contains(tmp3)){
								tmp5.setLocation(
										tmp4.x+(Ext*(tmp4.x-tmp3.x)/Math.hypot(tmp4.x-tmp3.x,tmp4.y-tmp3.y))
										,
										tmp4.y+(Ext*(tmp4.y-tmp3.y)/Math.hypot(tmp4.x-tmp3.x, tmp4.y-tmp3.y)));

							}
							else if (!baseArea.contains(tmp3)){
								tmp5.setLocation(
										tmp4.x+(Ext*(tmp3.x-tmp4.x)/Math.hypot(tmp3.x-tmp4.x,tmp3.y-tmp4.y))
										,
										tmp4.y+(Ext*(tmp3.y-tmp4.y)/Math.hypot(tmp3.x-tmp4.x, tmp3.y-tmp4.y)));
							}
							ExtArea.addPoint((int)tmp5.x, (int)tmp5.y);
						}
					}
				}
				for(StandardEntity checker:getObjectsInRange(e, 100*1000)){
					if(checker instanceof Building){
						if(((Building)checker).isEdgesDefined()){
							for(Edge Edger:((Building)checker).getEdges()){
								if(ExtArea.contains(Edger.getStartX(),Edger.getStartY())){
									Neighbours.add(checker);
								}
								else	if(ExtArea.contains((Edger.getStartX()+Edger.getEndX())/2,(Edger.getStartY()+Edger.getEndY())/2)){
									Neighbours.add(checker);
								}
							}
						}
					}
				}
			}
		}
		else if (e instanceof Road){
			for(EntityID Neighs:((Road) e).getNeighbours()){
				if(getEntity(Neighs) instanceof Road){Neighbours.add(getEntity(Neighs));}
				else if (getEntity(Neighs) instanceof Building){Neighbours.add(getEntity(Neighs));}
			}
		}
		return Neighbours;
	}
	
	
/* ------------------------------------------- handle civilians ------------------------------------------- */
	public CivilianVoiceListener createNewCivilianListener() {
		return new CivilianVoiceListener() {
			@Override
			public void hear(AKSpeak message) {
				final EntityID id = message.getAgentID();
				addNewCivilian(id);
			}
		};
	}
	
	public Civilian addNewCivilian(EntityID id) {
		StandardEntity existing = getEntity(id);
		if (existing == null) {
			Civilian civilian = new Civilian(id);
			timestamp.addStateChangeListener(civilian);
			addEntity(civilian);
			return civilian;
		}
		return null;
	}
    
	public Building getBuildingInPoint(int x, int y) {
		String xy = x + "," + y;
		return buildingXYMap.get(xy);
	}
   
	
	
	
/* ------------------------------------------- SEUBuilding part -------------------------------------------- */
	
	/**
	 * Create {@link SEUBuilding} object for this world mdel.
	 */
	private void createSeuBuildings() {
		this.seuBuildingMap = new FastMap<EntityID, SEUBuilding>();
		
		for (StandardEntity entity : this.getEntitiesOfType(AgentConstants.BUILDINGS)) {
			SEUBuilding seuBuilding;
			Building building = (Building) entity;
			String xy = building.getX() + "," + building.getY();
			buildingXYMap.put(xy, building);
			
			seuBuilding = new SEUBuilding(entity, this);
			
			if (entity instanceof Refuge || entity instanceof PoliceOffice
					|| entity instanceof FireStation || entity instanceof AmbulanceCentre)
				seuBuilding.setInflammable(false);
			this.seuBuildingMap.put(building.getID(), seuBuilding);
		}
		
		for (SEUBuilding next : seuBuildingMap.values()) {
			Collection<StandardEntity> neighbour = getObjectsInRange(next.getId(), SEUWall.MAX_SAMPLE_DISTANCE);
			
			for (StandardEntity entity : neighbour) {
				if (entity instanceof Building) {
					next.addNeighbourBuilding(this.seuBuildingMap.get(entity.getID()));
				}
			}
		}
		
		if (AgentConstants.LAUNCH_VIEWER && SEU_BuildingLayer.SEU_BUILDING_MAP.isEmpty()) {
			SEUBuilding.VIEWER_BUILDING_MAP.put(selfId, seuBuildingMap);
			SEU_BuildingLayer.SEU_BUILDING_MAP.putAll(seuBuildingMap);
		}
	}
	
	/**
	 * Create {@link SEURoad} object for this world model.
	 */
	private void createSeuRoads() {
		this.seuRoadMap = new FastMap<EntityID, SEURoad>();
		this.seuRoads = new FastSet<SEURoad>();
		
		SEURoad seuRoad;
		Road road;
		
		for (StandardEntity entity : this.getEntitiesOfType(AgentConstants.ROADS)){
			road = (Road)entity;
			seuRoad = new SEURoad(road, this);
			this.seuRoadMap.put(entity.getID(), seuRoad);
			this.seuRoads.add(seuRoad);
		}
		
		if (AgentConstants.LAUNCH_VIEWER) {
			if (this.selfId.getValue() == 485278126)
				SEU_RoadLayer.SEU_ROAD_MAP.putAll(seuRoadMap);
		}
	}
	
/* ----------------------------------------- getters and setters ------------------------------------------- */

//	public void addBuildingInfo(BuildingInfo info) {
//		this.buildingInfos.add(info);
//	}
//	
//	public RegionModel getRegionModel() {
//        return regionModel;
//    }
//	public void setRegionModel(RegionModel regionModel) {
//        this.regionModel = regionModel;
//    }
//   
//    public Exploration<Building> getBuildingExploration() {
//        return this.buildingExploration;
//    }
   
	public ConfigConstants getConfig() {
		return config;
	}
	
	public POVRouter getRouter() {
		return router;
	}

	public Entrance getEntrance() {
		return entrance;
	}
	
	public int getTime() {
		return time;
	}

	public Uniform getUniform() {
		return uniform;
	}
	
	public BurningBuildings getBurningBuildings() {
		return this.burningBuildings;
	}
	
	public CollapsedBuildings getCollapsedBuildings() {
		return this.collapsedBuildings;
	}
	
	public ExtinguishedBuilding getExtinguishedBuildings() {
		return this.extinguishBuildings;
	}
	
	/**
	 * Get all border building of this world.
	 * 
	 * @return all border building of this world
	 */
	public Set<StandardEntity> getBorderBuildings() {
		return this.mapBorderBuildings;
	}

//	public StuckHandler getStuckHandle() {
//		return stuckHandler;
//	}
	
	public BuriedHumans getBuriedHumans() {
		return buriedHumans;
	}
	
	public SearchedBuildings getSearchedBuildings() {
		return searchedBuildings;
	}
	public void setSearchBuildings(SearchedBuildings searchBuildings) {
		this.searchedBuildings = searchBuildings;
	}
	
	public CriticalArea getCriticalArea() {
		return criticalArea;
	}
	
	public EnergyFlow getEnergyFlow() {
		return energyFlow;
	}
	
	public IsolatedArea getIsolated() {
		return isolated;
	}

	public TimeStamp getTimestamp() {
		return timestamp;
	}

	public void setTime(int time) {
		this.time = time;
	}
	
	public Dimension getMapDimension() {
		return this.mapDimension;
	}
	public double getMapWidth() {
		return this.mapWidth;
	}
	public double getMapHeight() {
		return this.mapHeight;
	}
	
	public boolean isCommunicationLess() {
		return this.communicationLess;
	}
	public void setCommunicationLess(boolean communicationLess) {
		this.communicationLess = communicationLess;
	}
	
	public boolean isCommunicationLow() {
		return this.communicationLow;
	}
	public void setCommunicationLow(boolean communicationLow) {
		this.communicationLow = communicationLow;
	}
	
	public boolean isCommunicationMedium() {
		return this.communicationMedium;
	}
	public void setCommunicationMedium(boolean communicationMedium) {
		this.communicationMedium = communicationMedium;
	}
	
	public boolean isCommunicationHigh() {
		return this.communicationHigh;
	}
	public void setCommunicationHigh(boolean communicationHigh) {
		this.communicationHigh = communicationHigh;
	}
	
	public WaterPort getWaterPort() {
		return this.waterPort;
	}

	public List<FireBrigade> getFireBrigadeList() {
		return fireBrigadeList;
	}

	public List<PoliceForce> getPoliceForceList() {
		return policeForceList;
	}
	
	public List<AmbulanceTeam> getAmbulanceTeamList() {
		return ambulanceTeamList;
	}

	public List<FireStation> getFireStationList() {
		return fireStationList;
	}

	public List<PoliceOffice> getPoliceOfficeList() {
		return policeOfficeList;
	}

	public List<AmbulanceCentre> getAmbulanceCenterList() {
		return ambulanceCenterList;
	}
	
	public List<EntityID> getFireBrigadeIdList() {
		return fireBrigadeIdList;
	}
	
	public List<EntityID> getPoliceForceIdList() {
		return policeForceIdList;
	}

	public List<EntityID> getAmbulanceTeamIdList() {
		return ambulanceTeamIdList;
	}

	public List<EntityID> getFireStationIdList() {
		return fireStationIdList;
	}

	public List<EntityID> getPoliceOfficeIdList() {
		return policeOfficeIdList;
	}
	
	public List<EntityID> getAmbulanceCenterIdList() {
		return ambulanceCenterIdList;
	}
	
	public Set<EntityID> getStuckedAgents() {
		return this.stuckedAgentList;
	}

	public StandardEntity getSelfPosition() {
		if (building != null)
			return building;
		else {
			return me.getPosition(this);
		}
	}
	
	public Pair<Integer, Integer> getSelfLocation() {
        if (me != null) {
            return me.getLocation(this);
        } else {
            return building.getLocation(this);
        }
    }
	
	public void setThinkStartTime(long thinkStartTime) {
		this.thinkStartTime = thinkStartTime;
	}
	
	public long getThinkStartTime() {
		return thinkStartTime;
	}
	
	public void setExceptionMessage(String message) {
		this.exceptionMessage = message;
	}
	public String getExceptionMessage() {
		return this.exceptionMessage;
	}
	
	public RemainCluster getRemainCluster(){
		return this.remainCluster;
	}
	
	public SeuZones getZones() {
		return zones;
	}
	
	public boolean isNoRadio() {
		return this.isNoRadio;
	}
	
	public void setNoRadio(boolean noRadio) {
		this.isNoRadio = noRadio;
	}
	
	 public int getMinDistanceToRefuge(EntityID humanID){
			if (getEntitiesOfType(StandardEntityURN.REFUGE).isEmpty())
				return 1;
			Human human = (Human)getEntity(humanID);
			int minDistance = Integer.MAX_VALUE;
			for (StandardEntity refuge : getEntitiesOfType(StandardEntityURN.REFUGE)){
				int distance =  getDistance((Area)refuge , (Area)human.getPosition(this));
				if (distance < minDistance)
					minDistance = distance;
			}
			return minDistance;
		}
	 
		  public Collection<StandardEntity> getRefuges() {
		        return getEntitiesOfType(StandardEntityURN.REFUGE);
		    }
		  public List<Refuge> getAllRefuges() {
				List<Refuge> result = getEntitiesOfType(Refuge.class,
						StandardEntityURN.REFUGE);
				return result;
			}
		  public List<GasStation> getAllGasStations() {
				List<GasStation> result = getEntitiesOfType(GasStation.class,
						StandardEntityURN.GAS_STATION);
				return result;
			}
		  public Collection<StandardEntity> getAllHydrants() {
			  Collection<StandardEntity> result = getEntitiesOfType(
						StandardEntityURN.HYDRANT);
				return result;
			}
		    public Collection<StandardEntity> getCentres() {
		        return getEntitiesOfType(
		                StandardEntityURN.AMBULANCE_CENTRE,
		                StandardEntityURN.POLICE_OFFICE,
		                StandardEntityURN.FIRE_STATION);
		    }

		    public Collection<StandardEntity> getRoads() {
		        return getEntitiesOfType(StandardEntityURN.ROAD, StandardEntityURN.HYDRANT);
		    }
	      public Collection<StandardEntity> getBuildings(){
	      	 return getEntitiesOfType(
	      			 StandardEntityURN.BUILDING,
			            StandardEntityURN.REFUGE,
			            StandardEntityURN.AMBULANCE_CENTRE,
			            StandardEntityURN.POLICE_OFFICE,
			            StandardEntityURN.FIRE_STATION,
			            StandardEntityURN.GAS_STATION);
	      }
	      
	      public Collection<StandardEntity> getUniforms() {
	      	return getEntitiesOfType(
	  			StandardEntityURN.AMBULANCE_TEAM,
	  			StandardEntityURN.FIRE_BRIGADE,
	  			StandardEntityURN.POLICE_FORCE,
	  			StandardEntityURN.BUILDING,
	  			StandardEntityURN.ROAD,
	  			StandardEntityURN.REFUGE,
	  			StandardEntityURN.GAS_STATION,
	  			StandardEntityURN.HYDRANT);
	 }
	      
		    public Collection<StandardEntity> getAreas() {
		        return getEntitiesOfType(
		                StandardEntityURN.BUILDING,
		                StandardEntityURN.REFUGE,
		                StandardEntityURN.ROAD,
		                StandardEntityURN.AMBULANCE_CENTRE,
		                StandardEntityURN.POLICE_OFFICE,
		                StandardEntityURN.FIRE_STATION,
		                StandardEntityURN.HYDRANT,
		                StandardEntityURN.GAS_STATION);
		    }

		    public Collection<StandardEntity> getHumans() {
		        return getEntitiesOfType(
		                StandardEntityURN.CIVILIAN,
		                StandardEntityURN.FIRE_BRIGADE,
		                StandardEntityURN.POLICE_FORCE,
		                StandardEntityURN.AMBULANCE_TEAM);
		    }

		    public Collection<StandardEntity> getAgents() {
		        return getEntitiesOfType(
		                StandardEntityURN.FIRE_BRIGADE,
		                StandardEntityURN.POLICE_FORCE,
		                StandardEntityURN.AMBULANCE_TEAM);
		    }

		    public Collection<StandardEntity> getPoliceForces() {
		        return getEntitiesOfType(StandardEntityURN.POLICE_FORCE);
		    }

		    public Collection<StandardEntity> getPoliceOffices() {
		        return getEntitiesOfType(StandardEntityURN.POLICE_OFFICE);
		    }

		    public Collection<StandardEntity> getAmbulanceTeams() {
		        return getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM);
		    }

		    public Collection<StandardEntity> getAmbulanceCentres() {
		        return getEntitiesOfType(StandardEntityURN.AMBULANCE_CENTRE);
		    }

		    public Collection<StandardEntity> getFireBrigades() {
		        return getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE);
		    }

		    public Collection<StandardEntity> getFireStations() {
		        return getEntitiesOfType(StandardEntityURN.FIRE_STATION);
		    }

		    public Collection<StandardEntity> getCivilians() {//////// initial  return  size()=0;
		        return getEntitiesOfType(StandardEntityURN.CIVILIAN);
		    }

		    public StandardEntity getMe() {
		    	if (this.me != null)
		    		return this.me;
		    	if (this.building != null)
		    		return this.building;
		    	return null;
		    }
		    
		    
			public List<Road> getEntrance(Building building) {
					return buildingToRoadMap.get(building);
				}

		    public List<Building> getBuilding(Road road) {
					return  roadToBuildingMap.get(road);
				}

			
			public boolean isEntrance(Road road) {
					return allEntrance.contains(road);
				}
			
			public Set<Road> getAllEntrance() {
				return allEntrance;
		    }
			
			public Map<Road, List<Building>> getRoadToBuildingMap(){
				return roadToBuildingMap;
			}
			public Map<Building, List<Road>> getBuildingToRoadMap(){
				return buildingToRoadMap;
			}
			
			 public ClusterManager getClusterManager() {
			        return clusterManager;
			    }
			 
			 private void createClusterManager() {
			       //clusterManager = new seu.agent.fb.FireBrigadestools.CivilianClusterManager(this);
			    }
			
			public void updateClusters() {
				clusterManager.updateClusters();
			}
			
			public List<EntityID> getTeam() {
				return team;
			}
			
			@SuppressWarnings("unused")
			public void addTeamMembers(StandardEntityURN... urn) {
				Collection<StandardEntity> teamMembers = getEntitiesOfType(urn);
				for (StandardEntity next : teamMembers) {
					Human agent = (Human) next;
					EntityID member = next.getID();
					addToTeam(member);
				
				}

			}
			
			public void addToTeam(EntityID entity) {
				if (!team.contains(entity)) {
					team.add(entity);
					EntityIdComparator comparator = new EntityIdComparator();
					Collections.sort(team, comparator);
				}
			}

			public Set<Building> getBurningBuildingsSEU(){
				 
				 return BurningBuliding;
				 
			 }
			 public Set<Building> getExtinguishedBuildingsSEU(){
				 return  extinguishedBuilding;
				 
				 
			 }
			 public List<Building> getVisualExtinguishedBuildings(){
				 
				 return visualExtinguishedBuilding;
				 
			 }
			 public List<Building> getVisualBurningBuildings(){
				 
				 return visualBurningBuliding;
				 
			 }

			public Set<Building> getBurningOutBuildings(){
				  
				  return CollapsedBuildings;
				  
			}
			public List<Building> getVisualBurnOutBuildings(){
				  
				  return visualBurnOutBuilding;
				  
			}
			  public int getVisualATsNumber(){
				  return visibleATs.size();
				  
			  }
			   
			  public  List<EntityID> getVisualATs(){

				  return visibleATs;
				  
			  }
			  public  List<EntityID> getVisualHealthyATs(){

				  return visibleHealthyATs;
				  
			  }
			  public  List<EntityID> getVisualFBs(){

				  return visibleFBs;
				  
			  }
			  public  List<EntityID> getVisualPFs(){

				  return visiblePFs;
				  
			  }
			  
			  public  List<EntityID> getVisualAgents(){

					  return visibleAgents;
			  }
			  public  List<EntityID> getVisibleHumansToLoad(){

				  return visibleHumansToLoad;
			}
			  public List<EntityID> getFirstBuriedAgents(){
					
					 return firstBuriedAgents;
					 
				 }
			  public List<Blockade> getVisualBlockade(){
					
					return visualBlockade;
				}
				public  List<EntityID> getVisualBuriedHuman(){
					
					return visualBuriedHuman;
				}
				
				public List<EntityID> getNewDiedHuman(){
					
					return newDiedHuman;
				}
			   public Set<EntityID> getRescuedorDiedHuman(){
					
					return  RescuedorDiedHuman;
				}
			   
			   public List<StandardEntity> getEntities(List<EntityID> entityIDs) {
				     List<StandardEntity> result = new ArrayList<StandardEntity>();
				     for (EntityID next : entityIDs) {
				         result.add(getEntity(next));
				      }
				     return result;
				}
			   
			   public Clustering getPartitionModel(){
					return partitionModel;
				}
			   
			   private void initializeEntrance(){
					 allEntrance = new HashSet<Road>();
						roadToBuildingMap = new  FastMap<Road, List<Building>>();
						buildingToRoadMap = new  FastMap<Building, List<Road>>();
						for (StandardEntity entity : getBuildings()) {
							Building building = (Building) entity;
							List<Road> entrances = setEntrance(building);
							for (Road road : entrances) {
								if(roadToBuildingMap.get(road)==null){
									List<Building> buildings=new ArrayList<Building>();
									 buildings.add(building);
									roadToBuildingMap.put(road,buildings);
								}
									
								roadToBuildingMap.get(road).add(building);
							}
							allEntrance.addAll(entrances);
							buildingToRoadMap.put((Building) entity, entrances);
						}
					 
					 
				 }
			   
			   private List<Road> setEntrance(Building building) {
					List<Road> roads = new ArrayList<Road>();
					Stack<Area> stack = new Stack<Area>();
					List<Area> visited = new ArrayList<Area>();
					stack.add(building);
					do {
						Area entity = stack.pop();
						visited.add(entity);
						if (entity instanceof Road) {
							roads.add((Road) entity);
							getSeuRoad(entity.getID()).setEntrance(true);
						} else if (entity instanceof Building) {
							Building bld = (Building) entity;
							for (EntityID id : bld.getNeighbours()) {
								Area e = (Area) getEntity(id);
								if (!visited.contains(e)) {
									stack.push(e);
								}
							}
						}
					} while (!stack.isEmpty());
					
					return roads;
				}
			   
			   private void updateAgentStates() {
				      // System.out.println("111111111111111");
					                 
					        if(agent.getMeAsHuman().getPosition(this) instanceof Building)  return;
					        
					                  Human human;
					            for (StandardEntity entity : getAgents()) {
					                human = (Human) entity;
					                if (human.getPosition(this) instanceof Building) {
					                	Building building =(Building) human.getPosition(this);
					                	firstBuriedAgents.add(human.getID());
					                	human.setHP(9000);
					                	human.setDamage(30);
					                	human.setBuriedness(20);
					                	human.setPosition(building.getID());
					                	human.setX(building.getX());
					                	human.setY(building.getY());
					                	
					                }

					    }
					 }
			   
			  
			 
}
