 package seu.model.object;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import seu.util.BitArrayInputStream;
import seu.util.IntPropertySorter;
import rescuecore2.messages.Command;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardPropertyURN;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
import seu.communication.CivilianVoiceListener;
import seu.communication.CommunicationUtil;
import seu.communication.MessageBitSection;
import seu.communication.Port;
import seu.communication.MessageConstant.MessageReportedType;
import seu.message.MessagePriority;
import seu.message.Type.HumanIsSavedOrDeadMessage;
import seu.message.Type.HumanisBuriedfromChangesetorVoiceMessage;
import seu.message.Type.RescuingMessage;
import seu.model.SEUWorldModel;
import seu.standard.HumanInformation;
import seu.util.BitUtil;


final public class SEUBuriedHumans {
	
	private ChangeSet changed;

	public Map<EntityID , Map<Integer , HumanInformation>> fromCommunication;
	
	public Set<EntityID> fromVoiceOfCivilian;
	
	public Set<EntityID> totalBuriedHuman ;
	

	public EntityID newRescuedHumanFromMe;

	public Set<EntityID> newRescuedHumanFromCommunication;

	
	public Map<EntityID,EntityID>firstPosition;
	
	private final SEUWorldModel world;
	private static final int MAX_N = 31;
	private static final int N_BIT_SIZE = BitUtil.needBitSize(MAX_N + 1);
	private static final int BURIEDNESS_SEND_MAX = 255;
	private static final int BURIEDNESS_BIT_SIZE = 8;
	//private final int MAX_HP;
	private final int MAX_HP_BIT;
	//private final int MAX_DAMAGE;
	private final int MAX_DAMAGE_BIT;
   
	@SuppressWarnings("rawtypes")
	private SEUAgent agent;
	
	public SEUBuriedHumans(SEUAgent agent) {
		this.agent=agent;
		this.world = agent.getModel();
		fromCommunication = new HashMap<EntityID , Map<Integer , HumanInformation>>();
		fromVoiceOfCivilian = new HashSet<EntityID>();
		newRescuedHumanFromMe = null;
		newRescuedHumanFromCommunication = new HashSet<EntityID>();
		
		totalBuriedHuman = new HashSet<EntityID>();
		firstPosition=new HashMap<EntityID,EntityID>();
		if(agent.agentType==StandardEntityURN.POLICE_FORCE){
		}else{
		totalBuriedHuman.addAll(world.getFirstBuriedAgents());
		}
		//MAX_HP = world.getConfig().maxRoundHP / world.getConfig().hpPrecision;
		MAX_HP_BIT = 5;
		//MAX_DAMAGE = world.getConfig().timestep / world.getConfig().damagePrecision;
		MAX_DAMAGE_BIT = 6;
		initionalFirstPosition();
	}

	public void update() {
		this.changed = agent.getChangeSet();
		if (world.getTime() < world.getConfig().ignoreTime) {
			return;
		}
		filter();
		updateBuiredHumanInformation();
		updateTotalBuriedHuman(changed);
		
		
	}
	

	public void filter(){
		 for (Iterator<EntityID> it = fromCommunication.keySet().iterator() ; it.hasNext() ;){
				EntityID humanID = it.next();
				if (world.getVisualBuriedHuman().contains(humanID)){
					it.remove();
					continue;
				}
				if (agent.getID().equals(humanID)){
					it.remove();
					continue;
				}
				
				if (world.getRescuedorDiedHuman().contains(humanID)){
					it.remove();
					continue;
				}
			}
		 for (Iterator<EntityID> it = fromVoiceOfCivilian.iterator() ; it.hasNext() ; ){
				EntityID humanID = it.next();

				if (world.getVisualBuriedHuman().contains(humanID)){
					it.remove();
					continue;
				}	
		 }
		 
	 }
	
	public void updateTotalBuriedHuman(ChangeSet changed){
		
	
		totalBuriedHuman.addAll(world.getVisualBuriedHuman());
		
		totalBuriedHuman.addAll(fromVoiceOfCivilian);
		
		totalBuriedHuman.addAll(fromCommunication.keySet());
		
		totalBuriedHuman.remove(newRescuedHumanFromCommunication);
		
		totalBuriedHuman.removeAll(world.getRescuedorDiedHuman());
		 
		totalBuriedHuman.remove(agent.getID());
		if(totalBuriedHuman.size()!=0){
			Set<EntityID> removeBuriedHuman=new HashSet<EntityID>();
		    for(EntityID next :changed.getChangedEntities()){
			  if(totalBuriedHuman.contains(next)){
				  Human human=(Human) world.getEntity(next);
				  if(human.isBuriednessDefined()&&human.getBuriedness()==0){
					removeBuriedHuman.add(next);
			       }
				  if(human.isHPDefined()&&human.getHP()==0){
					  removeBuriedHuman.add(next);
				  }
		    }
		}
		  for(EntityID next:firstPosition.keySet()){
			  if(agent.getMeAsHuman().getPosition().equals(firstPosition.get(next))){
				  if(!agent.canSee(next)){
					  removeBuriedHuman.add(next); 
					  world.getRescuedorDiedHuman().add(next);
				  }
	 
			  }
		
		  }
		 
		    totalBuriedHuman.removeAll(removeBuriedHuman);  
		  
		}
		

		
		
	}
	public void updateBuiredHumanInformation(){
		
		if(fromCommunication.keySet().size()!=0)
		for (EntityID next:fromCommunication.keySet()){
			
			boolean newest = true;
			HumanInformation nearestHumanInfo = null;
			try{
				Human human = (Human) world.getEntity(next);
				if (human == null) {
					human = world.addNewCivilian(next);
					newest = false;
				}
				int lastSeenTime = world.getTimestamp().getLastSeenTime(next);
			
				
				int nearestTime = 0;
				Map<Integer, HumanInformation> humanInfos = fromCommunication.get(next);
				for (Integer timeStep : humanInfos.keySet()) {
					if (nearestTime < timeStep.intValue()) {
						nearestTime = timeStep.intValue();
						nearestHumanInfo = humanInfos.get(timeStep);
					}
				}
				if (nearestHumanInfo == null)
					continue ;
				
				if (nearestTime > lastSeenTime)
					newest = false;
				if (!newest){
					human.setBuriedness(nearestHumanInfo.getBuriedness());
					human.setDamage(nearestHumanInfo.getDamage());
					human.setHP(nearestHumanInfo.getHP());
					human.setPosition(nearestHumanInfo.getPositionID());
					if (!human.isXDefined() || !human.isYDefined()){
						Area position = (Area)human.getPosition(world);
						human.setX(position.getX());
						human.setY(position.getY());
					}

				}
			}catch(ClassCastException e){
				e.printStackTrace(System.out);
				continue;
			}
		}
		
	
	}
	
 
	public void addRescuedHuman(EntityID humanID) {
		Human human = (Human) world.getEntity(humanID);
		int visualATs = world.getVisualHealthyATs().size()+1;
		
			if (human.getBuriedness()/visualATs<=3){
				newRescuedHumanFromMe = humanID;
			  if(!world.getRescuedorDiedHuman().contains(humanID))
					world.getRescuedorDiedHuman().add(newRescuedHumanFromMe);
			}
		
	}
	
	public void addRescuedHuman1(EntityID humanID) {
		if(!world.getRescuedorDiedHuman().contains(humanID))
			world.getRescuedorDiedHuman().add(newRescuedHumanFromMe);
	}
	

	 

	public List<EntityID> getBuriedHumanFromChangeSet(){
		return world.getVisualBuriedHuman();
	}
	
	public Set<EntityID> getBuriedHumanFromVoiceOfCivilian(){
		return fromVoiceOfCivilian;
	}
	
	public Map<EntityID , Map<Integer , HumanInformation>> getBuriedHumanFromCommunication(){
		return fromCommunication;
	}
	
	public Set<EntityID> getTotalBuriedHuman(){  
		return totalBuriedHuman;
	}
	
	public Set<EntityID> getRescuedorDiedHuman(){
		return world.getRescuedorDiedHuman();
	}

 
	
 
	public void processHumanisBuriedfromChangesetorVoiceMessage(HumanisBuriedfromChangesetorVoiceMessage message){
		
		int LocationNumber=message.getLocationNumber();
		int isBuilding=getSelectedBit(LocationNumber,16);
		 EntityID positionID;
		if(isBuilding==1)
			positionID=world.getUniform().toID(StandardEntityURN.BUILDING, ToSelectedBinary(LocationNumber,15)); 
		else
			positionID=world.getUniform().toID(StandardEntityURN.ROAD, ToSelectedBinary(LocationNumber,15));
		
		Area position=(Area) world.getEntity(positionID);
		 EntityID humanID=message.getHumanID();
		if(agent.getID().equals(humanID))
			return;
		 
		 int HPBDD=message.getHPBDD();
		 int reciveHP = (ToSelectedBinary(HPBDD,5)+1)*300;
		  int buriedness =(ToSelectedBinary(HPBDD,6,10)+1)*6;
		  int damage = (ToSelectedBinary(HPBDD,11,16)+1)*6;
		  int timeStep = message.getReceivedStep();
		  

		  
		if (position==null||positionID == null||buriedness > 80 
				|| buriedness <= 0||damage > 400 || damage <0)
			return;
 
		StandardEntity se = world.getEntity(humanID);
		if(se == null)
			{
			Civilian civilian = new Civilian(humanID);
			civilian.setHP(reciveHP);
			civilian.setDamage(damage);
			civilian.setBuriedness(buriedness);
			civilian.setPosition(positionID,position.getX(),position.getY());
		
			}
		if(position!=null&&position instanceof Building
				&&world.getSeuBuilding(position)!=null&&!world.getSeuBuilding(position).getCivilians().contains((Human)se)){
			world.getSeuBuilding(position).getCivilians().add((Human)se);
			
		}
		
		if (!fromCommunication.containsKey(humanID)){
			Map<Integer , HumanInformation> humanInfos = new HashMap<Integer , HumanInformation>();
			fromCommunication.put(humanID , humanInfos);
		}
		if (!fromCommunication.get(humanID).containsKey(new Integer(timeStep))){
			HumanInformation humanInfo = new HumanInformation(humanID , positionID , buriedness , damage , reciveHP , timeStep);
			fromCommunication.get(humanID).put(humanInfo.getTimeStep() , humanInfo);
		}
	
			firstPosition.put(humanID, positionID);
		
	}
		
	public void sendBuriedHumanSeenHeared(){
		List<EntityID> sendBuriedHumans=new ArrayList<EntityID>(world.getVisualBuriedHuman());
		if(agent.getMeAsHuman()!=null&&agent.getMeAsHuman().getBuriedness()>0){
			sendBuriedHumans.add(agent.getID());
		}
		if(sendBuriedHumans.size()!=0){
			
			for(EntityID next:sendBuriedHumans){
			
				int LocationNumber=0;
				int HPBDD=0;
				Human human = (Human)world.getEntity(next);
				  if(human.getBuriedness()<=0) continue;
				 EntityID id =human.getPosition();
			
				
			   firstPosition.put(next,id);
				
				StandardEntity position=human.getPosition(world);
				if(position instanceof Building){
					LocationNumber=(int) (world.getUniform().toUniform(id)+Math.pow(2, 15));
				}else
					LocationNumber= world.getUniform().toUniform(id);
	
				int HPbit=0;
				if(human.isHPDefined())
					{
				  if(human.getHP()%300>150)
				    HPbit=human.getHP()/300+1;
				   else
					HPbit=human.getHP()/300;
					}
				if(HPbit>32) HPbit=32;
				if(HPbit<1) HPbit=1;
				HPbit=HPbit-1;
				
				int BDbit=0;
				if(human.isBuriednessDefined()){
					if(human.getBuriedness()%6>3)
						BDbit=human.getBuriedness()/6+1;
					else
						BDbit=human.getBuriedness()/6;
				}
				if(BDbit>16) BDbit=16;
				if(BDbit<1) BDbit=1;
				BDbit=BDbit-1;
			
				int Dbit=0;
				if(human.isDamageDefined()){
					if(human.getDamage()%6>3)
						Dbit=human.getDamage()/6+1;
					else
						Dbit=human.getDamage()/6;
				}
				if(Dbit>64) Dbit=64;
				if(Dbit<1) Dbit=1;
				Dbit=Dbit-1;

				HPBDD=TOHPBDD(HPbit,BDbit,Dbit);
				
				HumanisBuriedfromChangesetorVoiceMessage message=new 
						HumanisBuriedfromChangesetorVoiceMessage(LocationNumber,next,HPBDD);
				
				  StandardEntity pos=human.getPosition(world);
		 			if(pos instanceof Building&&!world.getSeuBuilding(pos).getCivilians().contains(human)){
		 				world.getSeuBuilding(pos).getCivilians().add(human);
		 				
		 			}
		 			if((!world.isNoRadio())&&(!agent.isLimited))
        		 		agent.sendMessage(message, MessagePriority.High);
		}
		}
		
		
	if(fromVoiceOfCivilian.size()!=0){
		for (EntityID next : fromVoiceOfCivilian){
			int LocationNumber=0;
			int HPBDD=0;
			Human human = (Human)world.getEntity(next);
			  if(human.getBuriedness()<=0) continue;
			EntityID id =human.getPosition();
			
			 
				firstPosition.put(next,id);
		
			StandardEntity position=human.getPosition(world);
			if(position instanceof Building){
				LocationNumber=(int) (world.getUniform().toUniform(id)+Math.pow(2, 15));
			}else
				LocationNumber= world.getUniform().toUniform(id);
			int HPbit=0;
			if(human.isHPDefined())
				{
			  if(human.getHP()%300>150)
			    HPbit=human.getHP()/300+1;
			   else
				HPbit=human.getHP()/300;
				}
			if(HPbit>32) HPbit=32;
			if(HPbit<1) HPbit=1;
			HPbit=HPbit-1;
			
			int BDbit=0;
			if(human.isBuriednessDefined()){
				if(human.getBuriedness()%6>3)
					BDbit=human.getBuriedness()/6+1;
				else
					BDbit=human.getBuriedness()/6;
			}
			if(BDbit>16) BDbit=16;
			if(BDbit<1) BDbit=1;
			BDbit=BDbit-1;
		
			int Dbit=0;
			if(human.isDamageDefined()){
				if(human.getDamage()%6>3)
					Dbit=human.getDamage()/6+1;
				else
					Dbit=human.getDamage()/6;
			}
			if(Dbit>64) Dbit=64;
			if(Dbit<1) Dbit=1;
			Dbit=Dbit-1;

			HPBDD=TOHPBDD(HPbit,BDbit,Dbit);
			
			HumanisBuriedfromChangesetorVoiceMessage message=new 
					HumanisBuriedfromChangesetorVoiceMessage(LocationNumber,next,HPBDD);
			
			  StandardEntity pos=human.getPosition(world);
	 			if(pos instanceof Building&&!world.getSeuBuilding(pos).getCivilians().contains(human)){
	 				world.getSeuBuilding(pos).getCivilians().add(human);
	 				
	 			}

	 			if((!world.isNoRadio())&&(!agent.isLimited))
    		 		agent.sendMessage(message, MessagePriority.High);
		}
	    }
	}
	
		
	public void processHumanIsSavedOrDeadMessage(HumanIsSavedOrDeadMessage message){
	 
		
		StandardEntity se = world.getEntity(message.getHumanId());
		if (se != null && !(se instanceof Human))
			return;
		Human human = (Human)se;
		if (human == null) {
			human = world.addNewCivilian(message.getHumanId());
		}
		
		StandardEntity pos=world.getEntity(message.getHumanId());
		if(pos instanceof Building&&world.getSeuBuilding(pos).getCivilians().contains(human)){
			world.getSeuBuilding(pos).getCivilians().remove((Human)se);
			
		}
		if (!world.getRescuedorDiedHuman().contains(message.getHumanId())){
			 world.getRescuedorDiedHuman().add(message.getHumanId());
		}
		firstPosition.remove(human.getID());
		
	}
	public void sendHumanIsSavedOrDeadMessage(){
		 if(newRescuedHumanFromMe!=null){
			 
			 
			 HumanIsSavedOrDeadMessage message =new HumanIsSavedOrDeadMessage(newRescuedHumanFromMe);
			 
			 StandardEntity se = world.getEntity(newRescuedHumanFromMe);
 			if (se != null && !(se instanceof Human))
 				return;
 			Human human = (Human)se;
 		   StandardEntity pos=human.getPosition(world);
 			if(pos instanceof Building&&world.getSeuBuilding(pos).getCivilians().contains(human)){
 				world.getSeuBuilding(pos).getCivilians().remove(human);
 				
 			}
			
 			if((!world.isNoRadio())&&(!agent.isLimited))
		 		agent.sendMessage(message, MessagePriority.High);
 			
 			firstPosition.remove(human.getID());
			
			 
		 }
        if(!world.getNewDiedHuman().isEmpty()){
        	for(EntityID next:world.getNewDiedHuman()){
        		 HumanIsSavedOrDeadMessage message =new HumanIsSavedOrDeadMessage(next);
        		 
        		 StandardEntity se = world.getEntity(next);
        			if (se != null && !(se instanceof Human))
        				return;
        			Human human = (Human)se;
        		 StandardEntity pos=human.getPosition(world);
        			if(pos instanceof Building&&world.getSeuBuilding(pos).getCivilians().contains(human)){
        				world.getSeuBuilding(pos).getCivilians().remove(human);
        				
        			}
        
        			if((!world.isNoRadio())&&(!agent.isLimited))
        		 		agent.sendMessage(message, MessagePriority.High);
        			
        			firstPosition.remove(human.getID());
        		
        	}
        	
  
        }
        newRescuedHumanFromMe=null;
        world.getNewDiedHuman().clear();
	} 
	
	

	
	public void processCivilianCommand(AKSpeak speakCommand) { 
		if (!world.getRescuedorDiedHuman().contains(speakCommand.getAgentID())){
			Human human = (Human)world.getEntity(speakCommand.getAgentID());
			if(human==null) {
				
				return;
			}
			if (human.isBuriednessDefined() && human.getBuriedness()>0&&
					human.isHPDefined() &&
					human.isDamageDefined() &&
					human.isPositionDefined() 
					)// niu human.getPosition(world) instanceof Building
				fromVoiceOfCivilian.add(speakCommand.getAgentID());
				firstPosition.put(human.getID(), human.getPosition());
		}
	}
 
	public int TOHPBDD(int HPbit,int BDbit,int Dbit)
	{
		         int total=0;
				 int w[]= {0,0,0,0,0,0,0,0};
				  int t=0;
					if(BDbit<=0) BDbit= 0;
					while(BDbit!=0){
							w[t]=BDbit%2;
							BDbit=BDbit/2;
							t++;
						}
					 for(int n=0;n<w.length;n++){
						 total=(int) (total+w[n]*Math.pow(2,n+5));
				
					 }
				
					 
						 int m[]= {0,0,0,0,0,0,0,0};
						  int a=0;
							if(Dbit<=0) Dbit= 0;
							while(Dbit!=0){
									m[a]=Dbit%2;
									Dbit=Dbit/2;
									a++;
								}
							 for(int n=0;n<m.length;n++){
								 total=(int) (total+m[n]*Math.pow(2,n+10));
							
							 }
					 return HPbit+total;
	
	}
	
	 public int getSelectedBit(int n,int bit){ 
		 int m[]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
			int i=0;
		while(n!=0){
				m[i]=n%2;
				n=n/2;
				i++;
			}
		 
		  return m[bit-1];	 
			 
	 }
	 public int ToSelectedBinary(int n,int bit){
		 int m[]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};//48
			int i=0;
		int total=0;
		while(n!=0){
				m[i]=n%2;
				n=n/2;
				i++;
			}
		 for(int j=0;j<bit;j++){
			 total=(int) (total+m[j]*Math.pow(2,j));
		 }
		  return total;	 
			 
	 }
	 public int ToSelectedBinary(int n,int a,int b){
		 int m[]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
			int i=0;
		int total=0;
		while(n!=0){
				m[i]=(n%2);
				n=n/2;
				i++;
			}
		 for(int j=a-1;j<b;j++){
			 total=(int) (total+m[j]*Math.pow(2,j-a+1));
		 }
		  return total;	
	 }
	public void initionalFirstPosition(){
		if(world.getFirstBuriedAgents().isEmpty())
			return;
		for(EntityID next:world.getFirstBuriedAgents())
		 {
			Human human=(Human) world.getEntity(next);
		
			 firstPosition.put(next,human.getPosition());
		 }
	}
	
	 public int getDeathTime(Human human){
	    	if(human.isDamageDefined()&&human.isHPDefined()){
	    		if(human.getDamage()==0) human.setDamage(1);
	    		return (int) (human.getHP()/human.getDamage()*0.8);
	    	}
	    	 else return -1;
	    
	    	
	    }
	 public Port createAddPortForNoRadio(final CommunicationUtil comUtil, final int timeToLive) {
			return new Port() {
				private List<MessageBitSection> secList = new ArrayList<MessageBitSection>();
				private List<Human> addCivilianList;
				private List<Human> addAgentList;
				private int sentCount;
				private boolean firstRead = true;
				private boolean firstSent = true;
				
				
				@Override
				public void resetCounter() {
					// counter = 0;
				}
				
				@Override
				public void init(ChangeSet changed) {
					if (firstSent){
						addCivilianList = new ArrayList<Human>();
						addAgentList = new ArrayList<Human>();
						
						List<EntityID> sendBuriedHumans=new ArrayList<EntityID>(world.getVisualBuriedHuman());
						if(agent.getMeAsHuman()!=null&&agent.getMeAsHuman().getBuriedness()>0){
							sendBuriedHumans.add(agent.getID());
						}
						
						for (EntityID id : sendBuriedHumans){
							Human human = (Human)world.getEntity(id);
							if (!(human.getPosition(world) instanceof Building)){
//								System.out.println(world.getControlledEntity() + "   Time : " + world.getTime() + "   wrong #3#");
								continue;
							}
							if (human.getURN().equals(StandardEntityURN.CIVILIAN.toString()))
								addCivilianList.add(human);
							else addAgentList.add(human);
						}
						for (EntityID id : fromVoiceOfCivilian){
							Human human = (Human)world.getEntity(id);
							if (!(human.getPosition(world) instanceof Building)){
//								System.out.println(world.getControlledEntity() + "   Time : " + world.getTime() + "   wrong #4#");
								continue;
							}
							if (human.getURN().equals(StandardEntityURN.CIVILIAN.toString()))
								addCivilianList.add(human);
							else addAgentList.add(human);
						}
						Collections.sort(addCivilianList, 
								new IntPropertySorter<StandardEntity>(
										StandardPropertyURN.BURIEDNESS));
						Collections.sort(addAgentList, 
								new IntPropertySorter<StandardEntity>(
										StandardPropertyURN.BURIEDNESS));
						
						for (Iterator<MessageBitSection> iterator = secList.iterator() ; iterator.hasNext() ; ){
							MessageBitSection sec = iterator.next();
							if (sec.getTimeToLive() <= 0)
								iterator.remove();
						}
						if (!addCivilianList.isEmpty()){
							MessageBitSection newSec = createNewSection(addCivilianList , 25);
							newSec.setTimeToLive(timeToLive);
							secList.add(newSec);
						}
						else if (!addAgentList.isEmpty()){
							MessageBitSection newSec = createNewSection(addAgentList , 25);
							newSec.setTimeToLive(timeToLive);
							secList.add(newSec);
						}
						
					}
					sentCount = secList.size();
				}
				
				
				@Override
				public boolean hasNext() {
					if (sentCount > 0){
						firstSent = false;
						firstRead = true;
					}
					return sentCount > 0;
				}
				@Override
				public MessageBitSection next() {
					sentCount--;
					return secList.get(sentCount);
				}
				@Override
				public void read(EntityID sender, int time, BitArrayInputStream stream) {
					try {
						if (firstRead){
							firstSent = true;
							firstRead = false;
							fromCommunication.clear();
						}
						final int n = stream.readBit(N_BIT_SIZE);
						 
						for (int i = 0; i < n; i++) {
							final EntityID humanID = new EntityID(stream.readBit(comUtil.UINT_BIT_SIZE));
							final int uniform = stream.readBit(comUtil.BUILDING_UNIFORM_BIT_SIZE);
							EntityID positionID = world.getUniform().toID(StandardEntityURN.BUILDING, uniform);
//							final EntityID positionID = comUtil.readArea(stream, world.getUniform());
							final int buriedness = stream.readBit(BURIEDNESS_BIT_SIZE);
							final int reciveHP = (stream.readBit(MAX_HP_BIT)+1)*300;
							final int damage = (stream.readBit(MAX_DAMAGE_BIT)+1)*6;
							final int timeStep = stream.readBit(BitUtil.needBitSize(400));
							 
							// System.out.println("  heard    humanID "+humanID);
							
							 
							 if (positionID == null)
								continue;
							if (buriedness > 80 || buriedness <= 0)
								continue;
							if (reciveHP > 10000 || reciveHP <= 0)
								continue;
							if (damage > 300 || damage < 0)
								continue;
							Area position=(Area) world.getEntity(positionID);
							StandardEntity se = world.getEntity(humanID);
							if(se == null)
							{
							Civilian civilian = new Civilian(humanID);
							civilian.setHP(reciveHP);
							civilian.setDamage(damage);
							civilian.setBuriedness(buriedness);
							civilian.setPosition(positionID,position.getX(),position.getY());
							}
							
			
							if (!fromCommunication.containsKey(humanID)){
								Map<Integer , HumanInformation> humanInfos = new HashMap<Integer , HumanInformation>();
								fromCommunication.put(humanID , humanInfos);
							}
							if (!fromCommunication.get(humanID).containsKey(new Integer(timeStep))){
								HumanInformation humanInfo = new HumanInformation(humanID , positionID , buriedness , damage , reciveHP , timeStep);
								fromCommunication.get(humanID).put(humanInfo.getTimeStep() , humanInfo);

							}
							
							if(!firstPosition.containsKey(humanID)){
								
								firstPosition.put(humanID,positionID);
								
							}
						}
					}catch(ArrayIndexOutOfBoundsException e){
						e.printStackTrace(System.out);
					}
				}
				
				
				private MessageBitSection createNewSection(List<Human> addList , int priority){
					MessageBitSection sec = new MessageBitSection(priority);
					final int n = Math.min(addList.size(), MAX_N);
					sec.add(n, N_BIT_SIZE);
					for (int i = 0; i < n; i++) {
						Human human = addList.get(i);
						final int humanID = human.getID().getValue();
						int uniform = world.getUniform().toUniform(human.getPosition());
						sec.add(humanID , comUtil.UINT_BIT_SIZE);
						sec.add(uniform, comUtil.BUILDING_UNIFORM_BIT_SIZE);
//						comUtil.writeArea(sec , (Area)human.getPosition(world), world.getUniform());
						if (human.isBuriednessDefined()) {
							sec.add(Math.min(human.getBuriedness(), BURIEDNESS_SEND_MAX), BURIEDNESS_BIT_SIZE);
						}
						else {
							sec.add(BURIEDNESS_SEND_MAX, BURIEDNESS_BIT_SIZE);
						}
						int hp=10000;
						int HPBit=0;
						if(human.isHPDefined()){
						  hp=human.getHP();
						}
						if(hp/300>150) {
							HPBit=hp/300+1;	
							
						}else{
						HPBit=hp/300;
						
						}
						
						int damage=10000;
						int damageBit=0;
						if(human.isDamageDefined()){
							damage=human.getDamage();
						}
						if(damageBit/6>3) {
							damageBit=damage/6+1;	
							
						}else{
							damageBit=hp/300;
							
						}
						HPBit=Math.min(HPBit, 31);
						damageBit=Math.min(HPBit, 63);
						 
					
						sec.add(HPBit, MAX_HP_BIT);
						sec.add(damageBit, MAX_DAMAGE_BIT);
						sec.add(world.getTime() - 1 , BitUtil.needBitSize(400));
						
					}
					return sec;
				}
				
				@Override
				public MessageReportedType getMessageReportedType() {
					return MessageReportedType.REPORTRD_TO_ALL;
				}

				@Override
				public void printWrite(MessageBitSection packet, int channel) {
					// TODO Auto-generated method stub
					
				}

				@Override
				public void printRead(int channel) {
					// TODO Auto-generated method stub
					
				}
		
				
			};
		}
	 
	 public Port createRescuedOrDiedPortForNorRadio(final CommunicationUtil comUtil) {
			return new Port() {
				private List<EntityID> removeList = new ArrayList<EntityID>(world.getRescuedorDiedHuman());
				private boolean removeSent;
				private int MAX_REMOVE = 3;
				private boolean firstRead = true;
				private String readString = "";
				
				@Override
				public void resetCounter() {
					// counter = 0;
				}
				
				@Override
				public void init(ChangeSet changed) {
					removeSent = removeList.isEmpty();
					firstRead = true;
				}
				@Override
				public boolean hasNext() {
					return !removeSent;
				}
				@Override
				public MessageBitSection next() {
					final int len = removeList.size();
					removeSent = true;
					MessageBitSection sec = new MessageBitSection(50);
					final int n = Math.min(len , MAX_REMOVE);
					sec.add(n, N_BIT_SIZE);
					for (int i = len - n ; i <= len - 1; i++) {
						sec.add(removeList.get(i).getValue(), comUtil.UINT_BIT_SIZE);

					}
					return sec;
				}
				@Override
				public void read(EntityID sender, int time, BitArrayInputStream stream) {
					try {
						if (firstRead){
							firstRead = false;
							newRescuedHumanFromCommunication.clear();
						}
						final int n = stream.readBit(N_BIT_SIZE);
				
						for (int i = 0; i < n; i++) {
							final EntityID humanID = new EntityID(stream.readBit(comUtil.UINT_BIT_SIZE));
							 
							StandardEntity se = world.getEntity(humanID);
							if (se != null && !(se instanceof Human))
								continue;
							Human human = (Human)se;
							if (human == null) {
								human = world.addNewCivilian(humanID);
							}
							if (!world.getRescuedorDiedHuman().contains(humanID)){
								newRescuedHumanFromCommunication.add(humanID);
								world.getRescuedorDiedHuman().add(humanID);
							}
							if(firstPosition.containsKey(sender)){
								firstPosition.remove(sender);
							}
						}
					}catch(ArrayIndexOutOfBoundsException e){
						e.printStackTrace(System.out);
					}
				}
				@Override
				public MessageReportedType getMessageReportedType() {
					return MessageReportedType.REPORTRD_TO_ALL;
				}

				@Override
				public void printWrite(MessageBitSection packet, int channel) {
					// TODO Auto-generated method stub
					
				}

				@Override
				public void printRead(int channel) {
					// TODO Auto-generated method stub
					
				}
				
			};
		}
	 
	 private int divide(int value , int precision){
		   if(value%precision>precision/2) return value/precision+1;
		   else return value/precision;
		}
	
	 public CivilianVoiceListener createCivilianVoiceListener() {
			return new CivilianVoiceListener() {
				@Override
				public void hear(AKSpeak message) {
					if (!world.getRescuedorDiedHuman().contains(message.getAgentID())){
						Human human = (Human)world.getEntity(message.getAgentID());
						if (human.isBuriednessDefined() && 
								human.isHPDefined() &&
								human.isDamageDefined() &&
								human.isPositionDefined() &&
								human.getPosition(world) instanceof Building)
							fromVoiceOfCivilian.add(message.getAgentID());
						if(!firstPosition.containsKey(message.getAgentID()))
							firstPosition.put(message.getAgentID(), human.getPosition());
					}
				}
			};
		}
	
}
