package seu.model.object;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import seu.util.BitArrayInputStream;
import javolution.util.FastSet;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardPropertyURN;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
import seu.communication.CivilianVoiceListener;
import seu.communication.CommunicationUtil;
import seu.communication.MessageBitSection;
import seu.communication.Port;
import seu.communication.MessageConstant.MessageReportedType;
import seu.message.MessagePriority;
import seu.message.Type.BuildingIsBurningMessage;
import seu.model.SEUWorldModel;
import seu.util.BitUtil;


@SuppressWarnings("serial")
public class SEUBurningBuildings {
	private SEUWorldModel world;
	
	 
	private SEUAgent agent;
	
	private Comparator<Pair<Building, Integer>> timeComparator = new Comparator<Pair<Building,Integer>>() {

		@Override
		public int compare(Pair<Building, Integer> o1, Pair<Building, Integer> o2) {
			if (o1.second() > o2.second()) 
				return -1;
			if (o1.second() < o2.second())
				return 1;
			if (o1.second() == o2.second()) {
				if (o1.first().getID().getValue() > o2.first().getID().getValue())
					return -1;
				if (o1.first().getID().getValue() < o2.first().getID().getValue())
					return 1;
			}
			return 0;
		}
	};
	
	private Set<Pair<Building, Integer>> allShouldSend = new TreeSet<Pair<Building, Integer>>(timeComparator);
	private Set<Pair<Building, Integer>> newlyAddFromVoice = new FastSet<Pair<Building, Integer>>();
	private final int MAX_N = 31;
	private final int N_BIT_SIZE = BitUtil.needBitSize(MAX_N);
	
	public SEUBurningBuildings(SEUAgent  agent) {
		this.agent=agent;
		this.world = agent.getModel();
	}
	
	
	public void processBuildingIsBurningMessage(BuildingIsBurningMessage message){
		
		 
		EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, message.getBuildingNumber());
		int FandT = message.getFandT();
		
	
		//System.out.println(agent.getID()+"  process     id  "+id+"   time   "+agent.getTimeStep());
		int temperature=ToSelectedBinary(FandT,6)*25;
		int fireyness=ToSelectedBinary(FandT,7,8)+1;
		
		if (temperature == 0)
			return;
		if (0 > fireyness || fireyness >= 9) 
			return;
		if (temperature >=2000)
			return;
			
		if (agent.canSee(id))
			return;
		Building building = (Building) world.getEntity(id);
		building.setFieryness(fireyness);
		building.setTemperature(temperature);
		world.getBurningBuildings().add(building);
		
		SEUBuilding seuBuilding = world.getSeuBuilding(building.getID());
		 
		if (world.getMe() instanceof FireBrigade) {
			switch (building.getFieryness()) {
			case 0:
				seuBuilding.setFuel(seuBuilding.getInitialFuel());
				break;
			case 1:
				if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.66) {
					seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.75));
                } else if (seuBuilding.getFuel() == seuBuilding.getInitialFuel()) {
                	seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.90));
                }
				break;
			case 2:
				if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.33
                        || seuBuilding.getFuel() > seuBuilding.getInitialFuel() * 0.66) {
                    seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.50));
                }
				break;
			case 3:
				if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.01
                        || seuBuilding.getFuel() > seuBuilding.getInitialFuel() * 0.33) {
                    seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.15));
                }
				break;
			case 8:
				seuBuilding.setFuel(0);
				break;
			default:
				break;
			}
			seuBuilding.setEnergy(building.getTemperature() * seuBuilding.getCapacity(), "BurningBuildings");
		}
			
	}
		
		
	public void sendBuildingIsBurningMessage(){
		if (world.getVisualBurningBuildings().isEmpty())
			return ;
		for(Building next:world.getVisualBurningBuildings()){
		int number =world.getUniform().toUniform(next.getID());
		
		
		int FandT=0;
		int Fbit=next.getFieryness()-1;
		int Tbit=0;
		if(next.getTemperature()%25>12)
			Tbit=next.getTemperature()/25+1;
		else
			Tbit=next.getTemperature()/25;
		if(Tbit>63) Tbit=63;
		
		FandT=ToFandT(Tbit,Fbit);
	
		
		
			BuildingIsBurningMessage message=new BuildingIsBurningMessage(number,FandT);
		
			if((!world.isNoRadio()))
		 		agent.sendMessage(message, MessagePriority.High);
	
		}
		 
		world.getVisualBurningBuildings().clear();
		
		
	}
	 
	 public int ToFandT(int T ,int F){
			
			 int k[]= {0,0,0,0,0,0,0,0};
			int total=0;
			int i=0;
			if(F<=0) F= 0;
			while(F!=0){
					k[i]=F%2;
					F=F/2;
					i++;
				}
			 for(int j=0;j<k.length;j++){
				 total=(int) (total+k[j]*Math.pow(2,j+6));
			 }
		
			  return total+T;	  
		 }
	 
		 
		 
		 
		 public int ToSelectedBinary(int n,int bit){ 
			 int m[]= {0,0,0,0,0,0,0,0};
				int i=0;
			int total=0;
			while(n!=0){
					m[i]=n%2;
					n=n/2;
					i++;
				}
			 for(int j=0;j<bit;j++){
			
				 total=(int) (total+m[j]*Math.pow(2,j));
			
			 }
			  return total;	 
				 
		 }
		 public int ToSelectedBinary(int n,int a,int b){
			 int m[]= {0,0,0,0,0,0,0,0};
				int i=0;
			int total=0;
			while(n!=0){
					m[i]=(n%2);
					n=n/2;
					i++;
				}
			 for(int j=a-1;j<b;j++){
				 total=(int) (total+m[j]*Math.pow(2,j-a+1));
		
			 }
			  return total;	
		 }
		 public Port createNoComPort(final CommunicationUtil comUtil, final int timeToLive) {
				return new Port() {
					
					private ChangeSet changed;
					private List<MessageBitSection> packetList = new ArrayList<MessageBitSection>();
					private int counter = 0;
					private String readString = "";

					@Override
					public void init(ChangeSet changed) {
						counter = 0;
						this.changed = changed;
						
						allShouldSend.addAll(newlyAddFromVoice);
						for (Building next : world.getVisualBurningBuildings()) {
							allShouldSend.add(new Pair<Building, Integer>(next, world.getTime()));
						}
						
						newlyAddFromVoice.clear();
						
						MessageBitSection newPacket = createNewPacket();
						if (newPacket != null)
							packetList.add(newPacket);
						for (Iterator<MessageBitSection> itor = packetList.iterator(); itor.hasNext(); ) {
							if (itor.next().getTimeToLive() <= 0)
								itor.remove();
						}
					}
					
					private MessageBitSection createNewPacket() {
						if (allShouldSend.isEmpty())
							return null;
						MessageBitSection sec = new MessageBitSection(5);
						sec.setTimeToLive(timeToLive);
						final int n = Math.min(allShouldSend.size(), MAX_N + 1);
						sec.add(n, N_BIT_SIZE);
						int count = 0;
						for (Pair<Building, Integer> next : allShouldSend) {
							count++;
							if (count > n)
								break;
							Building building = next.first();
							int uniform = world.getUniform().toUniform(building.getID());
							sec.add(uniform, comUtil.BUILDING_UNIFORM_BIT_SIZE);
							sec.add(building.getFieryness(), comUtil.FIRERYNESS_BIT_SIZE);
							sec.add(building.getTemperature(), comUtil.UINT_BIT_SIZE);
							sec.add(next.second().intValue(), comUtil.TIME_BIT_SIZE);
						}
						allShouldSend.clear();
						return sec;
					}

					@Override
					public boolean hasNext() {
						if (counter < packetList.size())
							return true;
						return false;
					}

					@Override
					public MessageBitSection next() {
						MessageBitSection next = packetList.get(counter);
						counter++;
						return next;
					}

					@Override
					public void read(EntityID sender, int time, BitArrayInputStream stream) {
						SEUBuilding seuBuilding;
						final int n = stream.readBit(N_BIT_SIZE);
						 
						for (int i = 0; i < n; i++) {
							final int uniform = stream.readBit(comUtil.BUILDING_UNIFORM_BIT_SIZE);
							EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, uniform);
							final int fieryness = stream.readBit(comUtil.FIRERYNESS_BIT_SIZE);
							final int temperature = stream.readBit(comUtil.UINT_BIT_SIZE);
							final int findTime = stream.readBit(comUtil.TIME_BIT_SIZE);
							 
							
							if (id == null)
								continue;
							if (changed.getChangedEntities().contains(id)) 
								continue;
							Building building = (Building)world.getEntity(id);
							
							int temperatureLastChangeTime = 
									world.getTimestamp().getPropertyTimeStamp(id, StandardPropertyURN.TEMPERATURE);
							int fierynessLastChangeTime = 
									world.getTimestamp().getPropertyTimeStamp(id, StandardPropertyURN.FIERYNESS);
							if (temperatureLastChangeTime >= findTime || fierynessLastChangeTime >= findTime)
								continue;
							
							seuBuilding = world.getSeuBuilding(id);
							if (0 <= fieryness && fieryness < 9)
								building.setFieryness(fieryness);
							else
								continue;
							if (temperature == 0)
								continue;
							if (temperature < 2000)
								building.setTemperature(temperature);
							else
								continue;
							
							newlyAddFromVoice.add(new Pair<Building, Integer>(building, findTime));
							
							// for fb's simulator
							if (agent.agentType==StandardEntityURN.FIRE_BRIGADE) {
								switch (building.getFieryness()) {
								case 0:
									seuBuilding.setFuel(seuBuilding.getInitialFuel());
									break;
								case 1:
									if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.66) {
										seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.75));
				                    } else if (seuBuilding.getFuel() == seuBuilding.getInitialFuel()) {
				                    	seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.90));
				                    }
									break;
								case 2:
									if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.33
				                            || seuBuilding.getFuel() > seuBuilding.getInitialFuel() * 0.66) {
				                        seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.50));
				                    }
									break;
								case 3:
									if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.01
				                            || seuBuilding.getFuel() > seuBuilding.getInitialFuel() * 0.33) {
				                        seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.15));
				                    }
									break;
								case 8:
									seuBuilding.setFuel(0);
									break;
								default:
									break;
								}
								seuBuilding.setEnergy(building.getTemperature() * seuBuilding.getCapacity(), "BurningBuildings"); 
							}
						}
					}

					@Override
					public MessageReportedType getMessageReportedType() {
						return MessageReportedType.REPORTRD_TO_ALL;
					}

					@Override
					public void resetCounter() {
						counter = 0;
					}

					@Override
					public void printWrite(MessageBitSection packet, int channel) {
						// TODO Auto-generated method stub
						
					}

					@Override
					public void printRead(int channel) {
						// TODO Auto-generated method stub
						
					}
				};
			}
	
	public Port createVeryLimitComPort(final CommunicationUtil comUtil, final int timeToLive) {
		return new Port() {
			
			private ChangeSet changed;
			private List<MessageBitSection> packetList = new ArrayList<MessageBitSection>();
			private int counter = 0;
			private String readString = "";

			@Override
			public void init(ChangeSet changed) {
				counter = 0;
				this.changed = changed;
				
				// allShouldSend.addAll(newlyAddFromVoice);
				for (Building next : world.getVisualBurningBuildings()) {
					allShouldSend.add(new Pair<Building, Integer>(next, world.getTime()));
				}
				
				//fromChangeSetVeryLimit.clear();
				// newlyAddFromVoice.clear();
				
				MessageBitSection newPacket = createNewPacket();
				if (newPacket != null)
					packetList.add(newPacket);
				for (Iterator<MessageBitSection> itor = packetList.iterator(); itor.hasNext(); ) {
					if (itor.next().getTimeToLive() <= 0)
						itor.remove();
				}
			}
			
			private MessageBitSection createNewPacket() {
				if (allShouldSend.isEmpty())
					return null;
				MessageBitSection sec = new MessageBitSection(1);
				sec.setTimeToLive(timeToLive);
				final int n = Math.min(allShouldSend.size(), MAX_N + 1);
				sec.add(n, N_BIT_SIZE);
				int count = 0;
				for (Pair<Building, Integer> next : allShouldSend) {
					count++;
					if (count > n)
						break;
					Building building = next.first();
					int uniform = world.getUniform().toUniform(building.getID());
					sec.add(uniform, comUtil.BUILDING_UNIFORM_BIT_SIZE);
					sec.add(building.getFieryness(), comUtil.FIRERYNESS_BIT_SIZE);
					sec.add(building.getTemperature(), comUtil.UINT_BIT_SIZE);
					sec.add(next.second().intValue(), comUtil.TIME_BIT_SIZE);
				}
				allShouldSend.clear();
				return sec;
			}

			@Override
			public boolean hasNext() {
				if (counter < packetList.size())
					return true;
				return false;
			}

			@Override
			public MessageBitSection next() {
				MessageBitSection next = packetList.get(counter);
				counter++;
				return next;
			}

			@Override
			public void read(EntityID sender, int time, BitArrayInputStream stream) {
				SEUBuilding seuBuilding;
				final int n = stream.readBit(N_BIT_SIZE);
				for (int i = 0; i < n; i++) {
					final int uniform = stream.readBit(comUtil.BUILDING_UNIFORM_BIT_SIZE);
					EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, uniform);
					final int fieryness = stream.readBit(comUtil.FIRERYNESS_BIT_SIZE);
					final int temperature = stream.readBit(comUtil.UINT_BIT_SIZE);
					final int findTime = stream.readBit(comUtil.TIME_BIT_SIZE);
					
					System.out.println(agent.getID()+" read     id  "+id+"   time   "+agent.getTimeStep());
					if (id == null)
						continue;
					if (changed.getChangedEntities().contains(id)) 
						continue;
					Building building = (Building)world.getEntity(id);
					
					int temperatureLastChangeTime = 
							world.getTimestamp().getPropertyTimeStamp(id, StandardPropertyURN.TEMPERATURE);
					int fierynessLastChangeTime = 
							world.getTimestamp().getPropertyTimeStamp(id, StandardPropertyURN.FIERYNESS);
					if (temperatureLastChangeTime >= findTime || fierynessLastChangeTime >= findTime)
						continue;
					
					seuBuilding = world.getSeuBuilding(id);
					if (0 <= fieryness && fieryness < 9)
						building.setFieryness(fieryness);
					else
						continue;
					if (temperature == 0)
						continue;
					if (temperature < 2000)
						building.setTemperature(temperature);
					else
						continue;
					
					newlyAddFromVoice.add(new Pair<Building, Integer>(building, findTime));
					
					if (agent.agentType==StandardEntityURN.FIRE_BRIGADE) {
						switch (building.getFieryness()) {
						case 0:
							seuBuilding.setFuel(seuBuilding.getInitialFuel());
							break;
						case 1:
							if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.66) {
								seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.75));
		                    } else if (seuBuilding.getFuel() == seuBuilding.getInitialFuel()) {
		                    	seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.90));
		                    }
							break;
						case 2:
							if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.33
		                            || seuBuilding.getFuel() > seuBuilding.getInitialFuel() * 0.66) {
		                        seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.50));
		                    }
							break;
						case 3:
							if (seuBuilding.getFuel() < seuBuilding.getInitialFuel() * 0.01
		                            || seuBuilding.getFuel() > seuBuilding.getInitialFuel() * 0.33) {
		                        seuBuilding.setFuel((float) (seuBuilding.getInitialFuel() * 0.15));
		                    }
							break;
						case 8:
							seuBuilding.setFuel(0);
							break;
						default:
							break;
						}
						seuBuilding.setEnergy(building.getTemperature() * seuBuilding.getCapacity(), "BurningBuildings"); 
					}
				}
			}

			@Override
			public MessageReportedType getMessageReportedType() {
				return MessageReportedType.REPORTRD_TO_ALL;
			}

			@Override
			public void resetCounter() {
				counter = 0;
			}

			@Override
			public void printWrite(MessageBitSection packet, int channel) {
				// TODO Auto-generated method stub
				
			}

			@Override
			public void printRead(int channel) {
				// TODO Auto-generated method stub
				
			}
		};
	}
	
	


}
