package seu.model.object;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import seu.util.BitArrayInputStream;
import rescuecore2.messages.Command;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardPropertyURN;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
import seu.communication.CommunicationUtil;
import seu.communication.MessageBitSection;
import seu.communication.Port;
import seu.communication.MessageConstant.MessageReportedType;
import seu.message.MessagePriority;
import seu.message.Type.BuildingIsBurnedOutMessage;
import seu.model.SEUWorldModel;
import seu.model.TimeStamp;
import seu.util.BitUtil;



@SuppressWarnings("serial")
public class SEUBurningOutBuildings{

	private SEUWorldModel world;
	private SEUAgent agent;
	public ChangeSet changed;
	private ArrayList<Building> addRecoreds = new ArrayList<Building>();
	private final int MAX_NUMBER = 31;
	private final int NUMBER_BIT_SIZE = BitUtil.needBitSize(MAX_NUMBER);
	
	public SEUBurningOutBuildings(SEUAgent agent){
		this.agent=agent;
		this.world = agent.getModel();
	}
	
	public void processBuildingIsBurnedOutMessage(BuildingIsBurnedOutMessage message){
		EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, message.getBuildingNumber());
		if (id == null)
			return;
	
		TimeStamp timestamp = world.getTimestamp();
		if (timestamp.getPropertyTimeStamp(id, StandardPropertyURN.FIERYNESS) >= message.getReceivedStep())
			return;
		if (agent.canSee(id))
			return;
		Building building = world.getEntity(id, Building.class);
		building.setFieryness(8);
		SEUBuilding seuBuilding = world.getSeuBuilding(id);
		seuBuilding.setFuel(0);
		
		world.getBurningOutBuildings().add(building);
		
	}
	
	
	public void sendBuildingIsBurnedOutMessage(){
		if(world.getVisualBurnOutBuildings().size()!=0){
			for(Building next:world.getVisualBurnOutBuildings()){
				int number = world.getUniform().toUniform(next.getID());
			
				BuildingIsBurnedOutMessage message=new BuildingIsBurnedOutMessage(number);
				
				if((!world.isNoRadio())&&(!agent.isLimited))
    		 		agent.sendMessage(message, MessagePriority.High);
				
			
			}
		}
	}
	 public int ToBurnedOutMessage(int number ,int Fieryness){ 
		
			 int k[]= {0,0,0,0,0,0,0,0};
			int total=0;
			int i=0;
			if(Fieryness<=0) Fieryness= 0;
			while(Fieryness!=0){
					k[i]=Fieryness%2;
					Fieryness=Fieryness/2;
					i++;
				}
			 for(int j=0;j<k.length;j++){
				 total=(int) (total+k[j]*Math.pow(2,j+6));
			 }
			
			  return total+number;	  
		 }
	 
		 
		 
		 
		 public int ToSelectedBinary(int n,int bit){
			 int m[]= {0,0,0,0,0,0,0,0};
				int i=0;
			int total=0;
			while(n!=0){
					m[i]=n%2;
					n=n/2;
					i++;
				}
			 for(int j=0;j<bit;j++){
			
				 total=(int) (total+m[j]*Math.pow(2,j));
			 }
			  return total;	 
				 
		 }
		 public int ToSelectedBinary(int n,int a,int b){
			 int m[]= {0,0,0,0,0,0,0,0};
				int i=0;
			int total=0;
			while(n!=0){
					m[i]=(n%2);
					n=n/2;
					i++;
				}
			 for(int j=a-1;j<b;j++){
				 total=(int) (total+m[j]*Math.pow(2,j-a+1));
			
			 }
			  return total;	
		 }
		 
		 public Port createAddPortNoRadio(final CommunicationUtil comUtil, final int timeToLive, final int priority) {
				return new Port() {
					private ChangeSet changed;
					private List<MessageBitSection> packetList = new ArrayList<MessageBitSection>();
					private int counter = 0;
					private String readString = "";
					
					@Override
					public void resetCounter() {
						counter = 0;
					}
					
					@Override
					public void init(ChangeSet changed) {
						this.changed = changed;
						
						MessageBitSection newPacket = createNewPacket();
						if (newPacket != null)
							packetList.add(newPacket);
						for (Iterator<MessageBitSection> it = packetList.iterator(); it.hasNext();)
							if (it.next().getTimeToLive() <= 0)
								it.remove();
					}
					
					private MessageBitSection createNewPacket() {
						if (addRecoreds.isEmpty())
							return null;
						MessageBitSection sec = new MessageBitSection(priority);
						sec.setTimeToLive(timeToLive);
						final int n = Math.min(addRecoreds.size(), MAX_NUMBER);
						sec.add(n, NUMBER_BIT_SIZE);
						
						for (int i = 0; i < n; i++) {
							Building building = addRecoreds.get(i);
							int uniform = world.getUniform().toUniform(building.getID());
							sec.add(uniform, comUtil.BUILDING_UNIFORM_BIT_SIZE);
							sec.add(building.getFieryness(), comUtil.FIRERYNESS_BIT_SIZE);
							sec.add(world.getTime(), comUtil.TIME_BIT_SIZE);
						}
						addRecoreds.clear();
						return sec;
					}
					
					@Override
					public boolean hasNext() {
						if (counter < packetList.size())
							return true;
						else
							return false;
					}
					
					@Override
					public MessageBitSection next() {
						MessageBitSection next = packetList.get(counter);
						// next.decrementTTL();
						counter++;
						return next;
						
					}
					
					@Override
					public void read(EntityID sender, int time, BitArrayInputStream stream) {
						final int n = stream.readBit(NUMBER_BIT_SIZE);
						readString = "sender: " + sender + ", sendTime = " + time + ", numberToRead = " + n + "\n";
						SEUBuilding seuBuilding;
						for (int i = 0; i < n; i++) {
							final int uniform = stream.readBit(comUtil.BUILDING_UNIFORM_BIT_SIZE);
							readString = readString + "uniform = " + uniform + "\n";
							final int fireyness = stream.readBit(comUtil.FIRERYNESS_BIT_SIZE);
							readString = readString + "fireyness = " + fireyness + "\n";
							final int findTime = stream.readBit(comUtil.TIME_BIT_SIZE);
							readString = readString + "findTime = " + findTime + "\n";
							
							EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, uniform);
							if (id == null)
								continue;
							TimeStamp timestamp = world.getTimestamp();
							if (timestamp.getPropertyTimeStamp(id, StandardPropertyURN.FIERYNESS) >= findTime)
								continue;
							if (changed.getChangedEntities().contains(id))
								continue;
							Building building = world.getEntity(id, Building.class);
							if (0 <= fireyness && fireyness < 9) {
								building.setFieryness(fireyness);
								seuBuilding = world.getSeuBuilding(id);
								seuBuilding.setFuel(0);
							}
							world.getBurningOutBuildings().add(building);
						}
					}
					
					@Override
					public MessageReportedType getMessageReportedType() {
						return MessageReportedType.REPORTRD_TO_ALL;
					}

					@Override
					public void printWrite(MessageBitSection packet, int channel) {
						// TODO Auto-generated method stub
						
					}

					@Override
					public void printRead(int channel) {
						// TODO Auto-generated method stub
						
					}
					

					
				};
			}
	
	public boolean contains(Building building) {
		return world.getBurningOutBuildings().contains(building);
	}
	
	public boolean contains(EntityID id) {
		return world.getBurningOutBuildings().contains(world.getEntity(id, Building.class));
	}

	
}
