package seu.model.object;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

 

import seu.message.MessagePriority;
import seu.message.Type.CriticalAreaisClearedMessage;
import seu.model.SEUWorldModel;
import rescuecore2.messages.Command;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;


public class SEUCriticalArea {
	private SEUWorldModel world;
	private SEUAgent agent;
	private List<Area> criticalAreas = new ArrayList<Area>();
	private List<Area> sendRemovedAreas = new ArrayList<Area>();
	
	public SEUCriticalArea( SEUAgent agent) {
		this.agent=agent;
		this.world = agent.getModel();
		
		for (StandardEntity se : world.getEntitiesOfType(SEUAgentConstants.ROADS)) {
			Area area = (Area) se;
			List<Edge> edges = area.getEdges();
			if (edges.size() < 3)
				continue;
			for (Edge edge : edges) {
				if (!edge.isPassable())
					continue ;
				if (world.getEntity(edge.getNeighbour()) instanceof Building)
					continue ;
			}
			
			criticalAreas.add(area);
		}
		
		for (Iterator<Area> itor = criticalAreas.iterator(); itor.hasNext();) {
			Area area = (Area)itor.next();
			List<EntityID> neighbours = area.getNeighbours();
			CriticalArea: for (EntityID entityId : neighbours) {
				Area neighbour = (Area) world.getEntity(entityId);
				List<EntityID> n_neighbours = neighbour.getNeighbours();
				if (n_neighbours.size() <= 2) {
					for (EntityID next : n_neighbours) {
						if (world.getEntity(next) instanceof Building) {
							itor.remove();
							break CriticalArea;
						}
					}
				} else if (neighbour.getEdges().size() == n_neighbours.size()) {
					itor.remove();
					break CriticalArea;
				}
			}
		}
		

	}
	
	

	
//	public void update(Search router) {
//		sendRemovedAreas.clear();
//		for (Iterator<Area> it = criticalAreas.iterator(); it.hasNext();) {
//			Area area = (Area) it.next();
//			
//			if (router.isSureReachable(area)) {
//				if (sendRemovedAreas.size() <= 1) {
//					sendRemovedAreas.add(area);
//				}
//				it.remove();
//			}
//		}
//	}

	
	
	public void processCriticalAreaisClearedMessage(CriticalAreaisClearedMessage message){
		
		EntityID id=world.getUniform().toID(StandardEntityURN.ROAD, message.getCriticalAreaNumber());
	//	System.out.println(agent.getID()+" processCritical   next.getID()  "+id+"    number  "+message.getCriticalAreaNumber());
		if(criticalAreas.contains(id))
		{
		Area area=(Area) world.getEntity(id);
			criticalAreas.remove(area);	
		}
		if(world.isNoRadio()||agent.isLimited)	{////////////////wan
		Human sendHuman = (Human) world.getEntity(message.getSender());
		if (world.getTimestamp().getLastChangedTime(message.getSender()) < message.getReceivedStep()) {
			if (world.getEntity(message.getSender()) instanceof Human&&message.getSender().getValue() != agent.getID().getValue()) {
				Area area=(Area) world.getEntity(id);
				sendHuman.setPosition(id, area.getX(), area.getY()); 
			}
		}
		}

	}
	
	public void sendCriticalAreaisClearedMessage(){//////  unuseed  niu 
		if (sendRemovedAreas.isEmpty())
			return;
		for(Area next:sendRemovedAreas){
		   int number=world.getUniform().toUniform(next.getID());
			CriticalAreaisClearedMessage message=new CriticalAreaisClearedMessage(number);
			//System.out.println(agent.getID()+" sendCritical  next.getID()  "+next.getID()+"    number  "+number);
			if(world.isNoRadio()||agent.isLimited)	{
		 		 message.setAddTime(agent.getTimeStep());
		 		 agent.mKeeper.addMessage(message);
		 	}else
		 		agent.sendMessage(message, MessagePriority.Medium);
			
			
		}
		
	
		sendRemovedAreas.clear();
	}
 

	public boolean isCriticalArea(Area area) {
		return criticalAreas.contains(area);
	}
 
}
