package seu.model.object;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import seu.util.BitArrayInputStream;
import rescuecore2.messages.Command;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardPropertyURN;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
import seu.communication.CommunicationUtil;
import seu.communication.MessageBitSection;
import seu.communication.Port;
import seu.communication.MessageConstant.MessageReportedType;
import seu.message.MessagePriority;
import seu.message.Type.BuildingIsExtinguishedMessage;
import seu.model.SEUWorldModel;
import seu.model.TimeStamp;
import seu.util.BitUtil;


@SuppressWarnings("serial")
public class SEUExtinguishedBuilding{
	/** An instance of world model.*/
	private SEUWorldModel world;
	private SEUAgent agent;
	private ChangeSet changed;
	private final int MAX_NUMBER = 31;
	private final int NUMBER_BIT_SIZE = BitUtil.needBitSize(MAX_NUMBER);
	private List<Building> addRecords = new ArrayList<Building>();
	public SEUExtinguishedBuilding(SEUAgent agent) {
		this.agent=agent;
		this.world = agent.getModel();
	
	}
	
	
	public void processBuildingIsExtinguishedMessage(BuildingIsExtinguishedMessage message){
		
        int ExtinguishedMessage=message.getExtinguishedMessage();
		EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, ToSelectedBinary(ExtinguishedMessage,12));
		
		int temperature=(ToSelectedBinary(ExtinguishedMessage,13,14)+1)*15;//(1-4)
		int fireyness=ToSelectedBinary(ExtinguishedMessage,15,16)+4;
		
		
		if (agent.canSee(id))
			return;
		TimeStamp timestamp = world.getTimestamp();
		if (timestamp.getPropertyTimeStamp(id, StandardPropertyURN.TEMPERATURE) >= message.getReceivedStep()
				|| timestamp.getPropertyTimeStamp(id, StandardPropertyURN.FIERYNESS) >= message.getReceivedStep())
			return;
		Building building = world.getEntity(id, Building.class);
		SEUBuilding seuBuilding = world.getSeuBuilding(id);
		if (0 <= fireyness && fireyness < 9)
			building.setFieryness(fireyness);
		else
			building.setFieryness(StandardEntityConstants.Fieryness.WATER_DAMAGE.ordinal());
		if (temperature > 0 && temperature < seuBuilding.getIgnitionPoint())
			building.setTemperature(temperature);
		else
			building.setTemperature((int)seuBuilding.getIgnitionPoint());
		seuBuilding.setEnergy(0, "extinguishedBuildings");
		seuBuilding.setWasEverWatered(true);
		
		world.getExtinguishedBuildings().add(building);
		
	}
	
	
	public void sendBuildingIsExtinguishedMessage(){
		
		if (world.getVisualExtinguishedBuildings().isEmpty())
			return;
		
		
		for(Building next:world.getVisualExtinguishedBuildings()){
			int number = world.getUniform().toUniform(next.getID());
			int Fbit=next.getFieryness()-4;//// 4 5 6 7 (2bit)
			
			int Tbit=0;
			if(next.getTemperature()%15>10)//////getTemperature below 60
				Tbit=next.getTemperature()/15+1;///(2bit)
			else
				Tbit=next.getTemperature()/15;
			
			if(Tbit>4) Tbit=4;
			Tbit=Tbit-1;////(0-3)
		
			int ExtinguishedMessage=ToExtinguishedMessage(number,Tbit,Fbit);////12bit 2bit 2bit
			
			
			BuildingIsExtinguishedMessage message=new BuildingIsExtinguishedMessage(ExtinguishedMessage);
			
			
			if((!world.isNoRadio()))
		 		agent.sendMessage(message, MessagePriority.High);
			
			
		}
		 
		world.getVisualExtinguishedBuildings().clear();
	}
		
	
	 public int ToExtinguishedMessage(int number ,int Tbit,int Fbit){
		 int k[]= {0,0,0,0,0,0,0,0};
		int total=0;
		int i=0;
		if(Tbit<=0) Tbit= 0;
		while(Tbit!=0){///bit  13 14
				k[i]=Tbit%2;
				Tbit=Tbit/2;
				i++;
			}
		 for(int j=0;j<k.length;j++){
			 total=(int) (total+k[j]*Math.pow(2,j+12));
		 }
		 int h[]= {0,0,0,0,0,0,0,0};
		  int m=0;
			if(Fbit<=0) Fbit= 0;
			while(Fbit!=0){///bit  15 16
					h[m]=Fbit%2;
					Fbit=Fbit/2;
					m++;
				}
			 for(int n=0;n<h.length;n++){
				 total=(int) (total+h[n]*Math.pow(2,n+14));
			 }
	
		  return total+number;	  
	 }
 
	 
	 
	 
	 public int ToSelectedBinary(int n,int bit){////from 0 to  bit  
		 int m[]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
			int i=0;
		int total=0;
		while(n!=0){
				m[i]=n%2;
				n=n/2;
				i++;
			}
		 for(int j=0;j<bit;j++){
			// if(j>=m.length)  break; ///////high  bit maybe zero
			 total=(int) (total+m[j]*Math.pow(2,j));
			// System.out.println("total11111111111111   "+total);
		 }
		  return total;	 
			 
	 }
	 public int ToSelectedBinary(int n,int a,int b){///from a to b
		 int m[]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
			int i=0;
		int total=0;
		while(n!=0){
				m[i]=(n%2);
				n=n/2;
				i++;
			}
		 for(int j=a-1;j<b;j++){
			 total=(int) (total+m[j]*Math.pow(2,j-a+1));
			// System.out.println("total2222222222222222222  "+total);
		 }
		  return total;	
	 }
	 
	 
	 public Port createAddPortNoRadio(final CommunicationUtil comUtil, final int timeToLive) {
			return new Port() {
				private ChangeSet changed;
				private List<MessageBitSection> packetList = new ArrayList<MessageBitSection>();
				private int counter = 0;
				private String readString = "";
				
				@Override
				public void resetCounter() {
					counter = 0;
				}
				
				@Override
				public void read(EntityID sender, int time, BitArrayInputStream stream) {
					final int n = stream.readBit(NUMBER_BIT_SIZE);
					readString = "sender: " + sender + ", sendTime = " + time + ", numberToRead = " + n + "\n";
					SEUBuilding seuBuilding;
					for (int i = 0; i < n; i++) {
						final int uniform = stream.readBit(comUtil.BUILDING_UNIFORM_BIT_SIZE);
						EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, uniform);
						final int fireyness = stream.readBit(comUtil.FIRERYNESS_BIT_SIZE);
						final int temperature = stream.readBit(comUtil.UINT_BIT_SIZE);
						final int findTime = stream.readBit(comUtil.TIME_BIT_SIZE);
						readString = readString + "uniform = " + uniform + ", id = " + id + "\n";
						readString = readString + "fieryness = " + fireyness + "\n";
						readString = readString + "temperature = " + temperature + "\n";
						readString = readString + "findTime = " + findTime + "\n";
						
						if (id == null)
							continue;
						if (changed.getChangedEntities().contains(id))
							continue;
						TimeStamp timestamp = world.getTimestamp();
						if (timestamp.getPropertyTimeStamp(id, StandardPropertyURN.TEMPERATURE) >= findTime
								|| timestamp.getPropertyTimeStamp(id, StandardPropertyURN.FIERYNESS) >= findTime)
							continue;
						Building building = world.getEntity(id, Building.class);
						seuBuilding = world.getSeuBuilding(id);
						if (0 <= fireyness && fireyness < 9)
							building.setFieryness(fireyness);
						else
							building.setFieryness(StandardEntityConstants.Fieryness.WATER_DAMAGE.ordinal());
						if (temperature > 0 && temperature < seuBuilding.getIgnitionPoint())
							building.setTemperature(temperature);
						else
							building.setTemperature((int)seuBuilding.getIgnitionPoint());
						seuBuilding.setEnergy(0, "extinguishedBuildings");
						seuBuilding.setWasEverWatered(true);
						
						world.getExtinguishedBuildings().add(building);
					}
				}
				
				@Override
				public MessageBitSection next() {
					MessageBitSection next = packetList.get(counter);
					// next.decrementTTL();
					counter++;
					return next;
				}
				
				/** This method used to create this cycle's packet.*/
				private MessageBitSection createNewPacket() {
					if (addRecords.isEmpty())
						return null;
					MessageBitSection sec = new MessageBitSection(9); 
					sec.setTimeToLive(timeToLive);
					final int n = Math.min(MAX_NUMBER, addRecords.size());
					sec.add(n, NUMBER_BIT_SIZE);
					for (int i = 0; i < n; i++) {
						Building building = addRecords.get(i);
						int uniform = world.getUniform().toUniform(building.getID());
						sec.add(uniform, comUtil.BUILDING_UNIFORM_BIT_SIZE);
						sec.add(building.getFieryness(), comUtil.FIRERYNESS_BIT_SIZE);
						sec.add(building.getTemperature(), comUtil.UINT_BIT_SIZE);
						sec.add(world.getTime(), comUtil.TIME_BIT_SIZE);
					}
					addRecords.clear();
					return sec;
				}
				
				@Override
				public void init(ChangeSet changed) {
					this.changed = changed;
					
					MessageBitSection newPacket = createNewPacket();
					if (newPacket != null)
					    this.packetList.add(newPacket);
					for (Iterator<MessageBitSection> it = packetList.iterator(); it.hasNext();) {
						if (it.next().getTimeToLive() <= 0)
							it.remove();
					}
				}
				
				@Override
				public boolean hasNext() {
					if (counter < packetList.size())
						return true;
					else
						return false;
				}
				
				@Override
				public MessageReportedType getMessageReportedType() {
					return MessageReportedType.REPORTRD_TO_ALL;
				}

				@Override
				public void printWrite(MessageBitSection packet, int channel) {
					// TODO Auto-generated method stub
					
				}

				@Override
				public void printRead(int channel) {
					// TODO Auto-generated method stub
					
				}
				

				
				
			};
		}
	 
//	 public int ToFandT(int T ,int F){////7 8 bit
//		// int m=ToBinary(n);////shi  stand er
//		// System.out.println(agent.getID()+" m   "+ m);
//		 int k[]= {0,0,0,0,0,0,0,0};
//		int total=0;
//		int i=0;
//		if(F<=0) F= 0;
//		while(F!=0){
//				k[i]=F%2;
//				F=F/2;
//				i++;
//			}
//		 for(int j=0;j<k.length;j++){
//			 total=(int) (total+k[j]*Math.pow(2,j+6));
//		 }
//		// System.out.println(agent.getID()+" (total+T) "+ (total+T));
//		  return total+T;	  
//	 }
// 
//	 
//	 
//	 
//	 public int ToSelectedBinary(int n,int bit){////from 0 to  bit  
//		 int m[]= {0,0,0,0,0,0,0,0};
//			int i=0;
//		int total=0;
//			if(n==0) return 0;
//		while(n!=0){
//				m[i]=n%2;
//				n=n/2;
//				i++;
//			}
//		 for(int j=0;j<bit;j++){
//			// if(j>=m.length)  break; ///////high  bit maybe zero
//			 total=(int) (total+m[j]*Math.pow(2,j));
//			// System.out.println("total11111111111111   "+total);
//		 }
//		  return total;	 
//			 
//	 }
//	 public int ToSelectedBinary(int n,int a,int b){///from a to b
//		 int m[]= {0,0,0,0,0,0,0,0};
//			int i=0;
//		int total=0;
//			if(n==0) return 0;
//		while(n!=0){
//				m[i]=(n%2);
//				n=n/2;
//				i++;
//			}
//		 for(int j=a-1;j<b;j++){
//			 total=(int) (total+m[j]*Math.pow(2,j-a+1));
//		//	 System.out.println("total2222222222222222222  "+total);
//		 }
//		  return total;	
//	 }
//	 
 
}
