package seu.model.object;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import seu.util.BitArrayInputStream;
import rescuecore2.messages.Command;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
import seu.agent.at.AmbulanceTeamsTools.partionTools.Cluster;
import seu.communication.CommunicationUtil;
import seu.communication.MessageBitSection;
import seu.communication.Port;
import seu.communication.MessageConstant.MessageReportedType;
import seu.message.MessagePriority;
import seu.message.Type.RemovedClusterMessage;
import seu.model.SEUWorldModel;
import seu.util.BitUtil;

 

public class SEUManageCluster {

	private Set<Integer> removeClusterIndex;

	private Set<Cluster> remainClusters;
	
	private ArrayList<Cluster> allClusters;

	private static final int CLUSTER_BIT = BitUtil.needBitSize(50);

	private SEUWorldModel world;
	private SEUAgent agent;
	
	private final int MAX_CLUSTER_INDEX ;
	
	public SEUManageCluster(SEUAgent agent){
		this.agent=agent;
		this.world = agent.getModel();
		removeClusterIndex = new HashSet<Integer>();
		MAX_CLUSTER_INDEX = world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM).size();
	}

	public void init(ArrayList<Cluster> allClusters){
		this.allClusters = allClusters;
		remainClusters = new HashSet<Cluster>(allClusters);
	}
	
	public void remove(int index){
		removeClusterIndex.add(index);
		remainClusters.remove(allClusters.get(index));

	}

	public Set<Cluster> getRemainClusters(){
		return remainClusters;
	}
	
	public void processRemovedClusterMessage(RemovedClusterMessage message){
		int removeIndex=message.getClusterIndex();
		
		if (world.getMe().getURN().equals(StandardEntityURN.AMBULANCE_TEAM.toString())){
			if (removeIndex >= MAX_CLUSTER_INDEX || removeIndex < 0)
				return ;
			else {
				remainClusters.remove(allClusters.get(removeIndex));
			}
		}
		
	}
	
	public void sendRemovedClusterMessage(){
		 if(removeClusterIndex.isEmpty())
			 return;
		 for(Integer next:removeClusterIndex){
			 RemovedClusterMessage message=new RemovedClusterMessage(next.intValue());
			
			 if((!world.isNoRadio())&&(!agent.isLimited))
 		 		agent.sendMessage(message, MessagePriority.Medium);
		 }
		  removeClusterIndex.clear();
		
	}
	
	public Port createRemovePortForNoRadio(final CommunicationUtil comUtil){
		return new Port(){

			private ArrayList<Integer> removeList = new ArrayList<Integer>();
			private boolean removeSent = false;
			private static final int MAX_SENT = 5;
			private MessageBitSection secToSent;
		 
			
			@Override
			public void resetCounter() {
				// counter = 0;
			}
			
			@Override
			public void init(ChangeSet changed) {
				if (!removeClusterIndex.isEmpty()){
					removeList.addAll(removeClusterIndex);
					removeClusterIndex.clear();
				}
				
				if (!removeList.isEmpty()){
					secToSent = createMessageBitSection(20);
					removeSent = true;
				}
				else removeSent = false;
			}

			@Override
			public boolean hasNext() {
				return removeSent;
			}

			@Override
			public MessageBitSection next() {
				removeSent = false;
				return secToSent;
			}

			@Override
			public void read(EntityID sender, int time, BitArrayInputStream stream) {
				final int n = stream.readBit(CLUSTER_BIT);
			 
				for (int i = 0 ; i < n ; i++){
					final int removeIndex = stream.readBit(CLUSTER_BIT);
					 
					if (removeIndex >= MAX_CLUSTER_INDEX || removeIndex < 0)
						return ;
					else if (!removeList.contains(removeIndex)){
						removeList.add(removeIndex);
						if (world.me.getURN().equals(StandardEntityURN.AMBULANCE_TEAM.toString()))
							remainClusters.remove(allClusters.get(removeIndex));
					}
				}
				 
			}
			
			@Override
			public MessageReportedType getMessageReportedType() {
				return MessageReportedType.REPORTRD_TO_ALL;
			}
			
			public MessageBitSection createMessageBitSection(int prioriy){
				MessageBitSection sec = new MessageBitSection(prioriy);
				int sentCount = Math.min(removeList.size() , MAX_SENT);
				sec.add(sentCount , CLUSTER_BIT);
				for (int i = 0 , j = removeList.size() - 1 ; i < sentCount ; i++ , j--){
					sec.add(removeList.get(j).intValue() , CLUSTER_BIT);
				}
				return sec;
			}

			@Override
			public void printWrite(MessageBitSection packet, int channel) {
				// TODO Auto-generated method stub
				
			}

			@Override
			public void printRead(int channel) {
				// TODO Auto-generated method stub
				
			}

		};
	} 

}
