package seu.model.object;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;

import rescuecore2.messages.Command;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
 
import seu.message.MessagePriority;
import seu.message.Type.AgentPositionMessage;
import seu.model.SEUWorldModel;

@SuppressWarnings("serial")
public class SEUSearchedBuildings extends HashSet<EntityID> {
	
	private SEUWorldModel world;
	private SEUAgent agent;

	public SEUSearchedBuildings(SEUAgent agent) {
		this.agent=agent;
		this.world = agent.getModel();			
	}
	
	public void update() {
		Human agent;
		for (StandardEntity next : world.getEntitiesOfType(SEUAgentConstants.PLATOONS)) {
			if (next instanceof Human) {
				agent = (Human) next;
				if (agent.isPositionDefined()) {
					if (world.getEntity(agent.getPosition()) instanceof Building) {
						this.add(agent.getPosition());
					}
				}
			}
		}

	}
	
	public boolean contains(Building building) {
		return super.contains(building);
	}
	public boolean contains(EntityID id) {
		return super.contains(world.getEntity(id));
	}
	public void processAgentPositionMessage(AgentPositionMessage message){
		EntityID id;
		int PositionMessage =message.getPositionMessage();
		
		int number=ToSelectedBinary(PositionMessage,14);
		int isBuilding=getSelectedBinary(PositionMessage,15);
		 if(isBuilding==1)
			 id=world.getUniform().toID(StandardEntityURN.BUILDING,number);
		 else 
			 id=world.getUniform().toID(StandardEntityURN.ROAD,number);
	
		 if(!(world.getEntity(message.getSender()) instanceof Human)) 
			 return;
		 Human sendHuman = (Human) world.getEntity(message.getSender());
		 
		 if (id == null)
				return;
		 Area area = (Area) world.getEntity(id);
		 
		
		 int isBlocked=getSelectedBinary(PositionMessage,16);
		 if(isBlocked==1){
			 world.getStuckedAgents().add(sendHuman.getID());
		 }else{
			 world.getStuckedAgents().remove(sendHuman.getID());
		 }

			if (message.getSender() == null || area == null)
				return;
			if (world.getTimestamp().getLastChangedTime(message.getSender()) < message.getReceivedStep()) {
				if (world.getEntity(message.getSender()) instanceof Human&&message.getSender().getValue() != agent.getID().getValue()) {
					
					sendHuman.setPosition(area.getID(), area.getX(), area.getY()); 
				}
			}
		 
		 
	}
	public void sendAgentPositionMessage(){
		
		StandardEntity area= world.getEntity(agent.getMeAsHuman().getPosition()) ;
		 int PositionMessage=0;
		 int isBuilidng=0;
		 int isBlocked=0;
		 
		 if(area instanceof Building){
			 isBuilidng=1;
			PositionMessage+=world.getUniform().toUniform(area.getID())+Math.pow(2, 14);
		}else
			PositionMessage+=world.getUniform().toUniform(area.getID());
		
		 
		 if(agent.agentType!=StandardEntityURN.POLICE_FORCE){
	    
		 if(agent.getMeAsHuman().getBuriedness()<=0&&(agent.isBlocked()||agent.isStuck(agent.getMeAsHuman()))){
			isBlocked=1;
			PositionMessage+=Math.pow(2, 15);
		}
		 }
		
		
		AgentPositionMessage message=new AgentPositionMessage(PositionMessage);
		

		if((!world.isNoRadio())&&(!agent.isLimited))
	 		agent.sendMessage(message, MessagePriority.High);
	
	}
	
	
	 public int ToSelectedBinary(int n,int bit){  
		 int m[]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
			int i=0;
		int total=0;
		while(n!=0){
				m[i]=n%2;
				n=n/2;
				i++;
			}
		 for(int j=0;j<bit;j++){
			
			 total=(int) (total+m[j]*Math.pow(2,j));
		
		 }
		  return total;	 
			 
	 }
	 public int getSelectedBinary(int n,int a){
		 int m[]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
			int i=0;
		while(n!=0){
				m[i]=(n%2);
				n=n/2;
				i++;
			}
		return m[a-1];
		 
		 
	 }
	 public int ToSelectedBinary(int n,int a,int b){///from a to b
		 int m[]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
			int i=0;
		int total=0;
		while(n!=0){
				m[i]=(n%2);
				n=n/2;
				i++;
			}
		 for(int j=a-1;j<b;j++){
			 total=(int) (total+m[j]*Math.pow(2,j-a+1));
			// System.out.println("total2222222222222222222  "+total);
		 }
		  return total;	
	 }
//	 public int ToBinary(int n,int bit){////from 0 to  bit  
//	 List<Integer> m=new ArrayList<Integer>();
//			int i=0;
//		int total=0;
//			if(n==0) return 0;
//		while(n!=0){
//				m.add(n%2);
//				n=n/2;
//				i++;
//			}
//		 for(int j=0;j<bit;j++){
//			 total=total+m.get(j)*10^j;
//		 }
//		  return total;	 
//			 
//	 }
//	 public int ToBinary(int n,int a,int b){///from a to b
//		 List<Integer> m=new ArrayList<Integer>();
//			int i=0;
//		int total=0;
//			if(n==0) return 0;
//		while(n!=0){
//				m.add(n%2);
//				n=n/2;
//				i++;
//			}
//		 for(int j=a;j<b;j++){
//			 total=total+m.get(j)*10^(j-a);
//		 }
//		  return total;	
//	 }
//	

}
