package seu.model.object;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import seu.util.BitArrayInputStream;
import rescuecore2.messages.Command;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.agent.SEUAgent;
import seu.agent.fb.FireBrigadeAgent;
import seu.communication.CommunicationUtil;
import seu.communication.MessageBitSection;
import seu.communication.Port;
import seu.communication.MessageConstant.MessageReportedType;
import seu.message.MessagePriority;
import seu.message.Type.WaterMessage;
import seu.model.SEUWorldModel;

public class SEUWaterPort {
	private SEUWorldModel world;
	private SEUAgent agent;
	private EntityID target = null;
	private int waterPower;
 

	Map<Integer, List<EntityID>> WaterLog = new TreeMap<Integer, List<EntityID>>();
	
	public SEUWaterPort(SEUAgent agent) {
		this.agent=agent;
		this.world = agent.getModel();
	}
	
	public void update() {
		for(Iterator<Integer> itor = WaterLog.keySet().iterator(); itor.hasNext(); ) {
			int next = itor.next().intValue();
			if (world.getTime() - next > 10)
				itor.remove();
		}
	}
	
	
	public void processWaterMessage(WaterMessage message){
		 
		int BNAWAT=message.getBuildingNumberAndWaterAndTimeStep();
		int buildingNumber=ToSelectedBinary(BNAWAT,12);
		int waterPower=625*(ToSelectedBinary(BNAWAT,13,15)+1);
		int  SendTime  =ToSelectedBinary(BNAWAT,16,24);
		EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, buildingNumber);
		 
		 
		if (agent.getID().getValue() == message.getSender().getValue()){
		
					return;
		}
			
		 

		List<EntityID> values = WaterLog.get(new Integer(SendTime));
		if (values != null) {
			if (values.contains(message.getSender()))
				return;
			else
				values.add(message.getSender());
		} else {
			List<EntityID> list = new ArrayList<EntityID>();
			list.add(message.getSender());
			WaterLog.put(new Integer(SendTime), list);
		}

		
		
		if (id == null)
			return;
		SEUBuilding seuBuilding = world.getSeuBuilding(id);
		if (waterPower >= world.getConfig().maxPower) {
			waterPower = world.getConfig().maxPower;
		}
	
		seuBuilding.setWaterQuantity(seuBuilding.getWaterQuantity() + waterPower);
	
	}
	
	public void sendWaterMessage(){
	
		if (agent instanceof FireBrigadeAgent && world.getTime() >=  world.getConfig().ignoreTime) {
			FireBrigadeAgent fbAgent = (FireBrigadeAgent)world.getAgent();
			StandardMessageURN lastTimeMessage = fbAgent.getCommandHistory(agent.getTimeStep()-1);
			StandardMessageURN extinguishMessage = StandardMessageURN.AK_EXTINGUISH;
		
			if (lastTimeMessage != null && lastTimeMessage.equals(extinguishMessage)) {
			
				target = fbAgent.extinguishTarget;
				waterPower = fbAgent.waterPower;
				fbAgent.waterPower = 0;
			} else {
				target = null;
				waterPower = 0;
				return;
			}
		} else {
			target = null;
			waterPower = 0;
			return;
		}
		int number = world.getUniform().toUniform(target);
		int water=0;
		if(waterPower%625>300)
			water=waterPower/625+1;
		else 
			water=waterPower/625;
		 if(water>8)
			water=8;
		if(water<1)water=1;
		int waterbit=water-1;
		 
		int BNAWAT=ToBNAWAT(number,waterbit,agent.getTimeStep());
		
		
		WaterMessage message=new WaterMessage(BNAWAT);
		
		if((!world.isNoRadio()))
	 		agent.sendMessage(message, MessagePriority.High);
		
		
		
	}
	 public int ToBNAWAT(int number ,int F,int time){
			 int k[]= {0,0,0,0,0,0,0,0};
			int total=0;
			int i=0;
			if(F<=0) F= 0;
			while(F!=0){
					k[i]=F%2;
					F=F/2;
					i++;
				}
			 for(int j=0;j<k.length;j++){
				 total=(int) (total+k[j]*Math.pow(2,j+12));
			 }
			 int h[]= {0,0,0,0,0,0,0,0,0};
			  int m=0;
				if(time<=0) time= 0;
				while(time!=0){
						h[m]=time%2;
						time=time/2;
						m++;
					}
				 for(int n=0;n<h.length;n++){
					 total=(int) (total+h[n]*Math.pow(2,n+15));
				 }
			 
		
			  return total+number;	  
		 }
	 
		 
		 
		 
		 public int ToSelectedBinary(int n,int bit){  
			 int m[]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
				int i=0;
			int total=0;
			while(n!=0){
					m[i]=n%2;
					n=n/2;
					i++;
				}
			 for(int j=0;j<bit;j++){
				
				 total=(int) (total+m[j]*Math.pow(2,j));
			
			 }
			  return total;	 
				 
		 }
		 public int ToSelectedBinary(int n,int a,int b){
			 int m[]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
				int i=0;
			int total=0;
			while(n!=0){
					m[i]=(n%2);
					n=n/2;
					i++;
				}
			 for(int j=a-1;j<b;j++){
				 total=(int) (total+m[j]*Math.pow(2,j-a+1));
	
			 }
			  return total;	
		 }

		 public Port createWaterPort(final CommunicationUtil comUtil, final int timeToLive) {
				return new Port() {

					private List<MessageBitSection> packetList = new ArrayList<MessageBitSection>();
					private int counter = 0;
					private String readString = "";
					
					@Override
					public void resetCounter() {
						counter = 0;
					}
					
					@Override
					public void init(ChangeSet changed) {
						MessageBitSection newPacket = createNewPacket();
						if (newPacket != null)
							packetList.add(newPacket);
						for (Iterator<MessageBitSection> it = packetList.iterator(); it.hasNext();) {
							if (it.next().getTimeToLive() <= 0) {
								it.remove();
							}
						}
					}
					
					private  MessageBitSection createNewPacket() {
						if (agent.agentType==StandardEntityURN.FIRE_BRIGADE && world.getTime() >= world.getConfig().ignoreTime) {
							FireBrigadeAgent fbAgent = (FireBrigadeAgent)world.getAgent();
							StandardMessageURN lastTimeMessage = fbAgent.getCommandHistory(world.getTime());
							StandardMessageURN extinguishMessage = StandardMessageURN.AK_EXTINGUISH;
							
							if (lastTimeMessage != null && lastTimeMessage.equals(extinguishMessage)) {
								target = fbAgent.extinguishTarget;
								waterPower = fbAgent.waterPower;
								fbAgent.waterPower = 0;
							} else {
								target = null;
								waterPower = 0;
							}
						} else {
							target = null;
							waterPower = 0;
						}
						
						if (target == null || waterPower <= 0)
							return null;
						MessageBitSection sec = new MessageBitSection(3);
						sec.setTimeToLive(timeToLive);
						int uniform = world.getUniform().toUniform(target);
						sec.add(uniform, comUtil.BUILDING_UNIFORM_BIT_SIZE);
						sec.add(waterPower, comUtil.WATER_POWER_BIT_SIZE);
						sec.add(world.getTime(), comUtil.TIME_BIT_SIZE);
						return sec;
					}

					@Override
					public boolean hasNext() {
						if (counter < packetList.size())
							return true;
						else
							return false;
					}

					@Override
					public MessageBitSection next() {
						MessageBitSection next = packetList.get(counter);
						// next.decrementTTL();
						counter++;
						return next;
					}

					@Override
					public void read(EntityID sender, int time, BitArrayInputStream stream) {
						
						int uniform = stream.readBit(comUtil.BUILDING_UNIFORM_BIT_SIZE);
						EntityID id = world.getUniform().toID(StandardEntityURN.BUILDING, uniform);
						
						int waterPower = stream.readBit(comUtil.WATER_POWER_BIT_SIZE);
			
						int firstSendTime = stream.readBit(comUtil.TIME_BIT_SIZE);
						
				
						if (agent.getID().equals(sender))
							return;
						
						List<EntityID> values = WaterLog.get(new Integer(firstSendTime));
						if (values != null) {
							if (values.contains(sender))
								return;
							else
								values.add(sender);
						} else {
							List<EntityID> list = new ArrayList<EntityID>();
							list.add(sender);
							WaterLog.put(new Integer(firstSendTime), list);
						}

						if (id == null)
							return;
						SEUBuilding seuBuilding = world.getSeuBuilding(id);
						if (waterPower >= world.getConfig().maxPower) {
							waterPower = world.getConfig().maxPower;
						}
						
					
						
						seuBuilding.setWaterQuantity(seuBuilding.getWaterQuantity() + waterPower);
					}

					@Override
					public MessageReportedType getMessageReportedType() {
						return MessageReportedType.REPORTED_TO_FB;
					}

					@Override
					public void printWrite(MessageBitSection packet, int channel) {
						// TODO Auto-generated method stub
						
					}

					@Override
					public void printRead(int channel) {
						// TODO Auto-generated method stub
						
					}
					
				
		
				};
			}

	
}
