package seu.model.object.seuZoneEntity;

import java.awt.Point;
import java.awt.Polygon;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javolution.util.FastSet;
import rescuecore2.standard.entities.Road;
import rescuecore2.worldmodel.EntityID;
import seu.geom.CompositeConvexHull;
import seu.geom.Distance;
import seu.model.SEUWorldModel;
import seu.model.object.SEUBuilding;



@SuppressWarnings("serial")
public class SeuZone extends ArrayList<SEUBuilding> implements Comparable<SeuZone>{
	/** The Id of this zone.*/
	private int id;
	/** A reference of world model.*/
	private SEUWorldModel world;
	
	/** The polygon of this zone.*/
	private Polygon zonePolygon;
	/** The center point of this polygon.*/
	private Point centerPoint;
	private double distanceToMapCenter;
	private double radiusLength;
	/**
	 * The number of buildings this zone has.
	 */
	private int buildingCount = 0;
	
	private List<SeuZone> neighbourZones = new ArrayList<SeuZone>();
	private List<Integer> neighbourZoneIds = new ArrayList<Integer>();
	private List<SeuZone> neighbourRegions = new ArrayList<SeuZone>();
	private Set<EntityID> unvisitBuildings = new HashSet<EntityID>();
	
	private Set<EntityID> unclearedEntrances = new HashSet<EntityID>();
	
	private Set<EntityID> buriedHumans = new FastSet<EntityID>();
	
	private double totalInitialFuel = 0.0d;
	private double totalGroundArea = 0.0d;
	private double totalArea = 0.0d;
	private List<SEUBuilding> buildingsInRegion;    //this(SeuZone)
//	private double totalRemainFuel;
//	private List<SEUBuilding> burningBuildings = new ArrayList<>();
//	private List<SEUBuilding> unburnedBuildings = new ArrayList<>();
//	
//	private boolean isOnFire;
//	private boolean isBurned;
	
	private double localValue;
	private double globalValue;
	private double zoneValue;
//	private double searchValue;
//	
//	private int neededAgentToExtinguish;
//	
//	private Map<EntityID, Integer> agentDistanceMap = new FastMap<EntityID, Integer>();
//	private int fireBrigadeSize;
//	private int maxPower;
	
	private Set<EntityID> surroundingRoad = new FastSet<EntityID>();
	private Set<Road> allEntranceRoad = null;
	
	int belongPfClusterIndex = -1;
	
	public SeuZone(SEUWorldModel world, int id) {
		this.world = world;
		this.id = id;
		buildingsInRegion=new ArrayList<SEUBuilding>();
		// this.fireBrigadeSize = world.getFireBrigadeIdList().size();
		// this.maxPower = world.getConfig().maxPower;
	}
	
	public void initialiseZone() {
		calculateCenterPoint();
		createZonePolygon();
		
		for (SEUBuilding next : this) {
			unvisitBuildings.add(next.getId());
		}
//		for (SEUBuilding next : buildingsInRegion) {//buildingsInRegion ?= this
//			unvisitBuildings.add(next.getId());
//		}
		for (Road next : getAllEntranceRoad()) {
			unclearedEntrances.add(next.getID());
		}
	}
	
	private void calculateCenterPoint() {
		centerPoint = new Point();
		for (SEUBuilding next : this) {
			centerPoint.x += next.getSelfBuilding().getX();
			centerPoint.y += next.getSelfBuilding().getY();
		}
//		for (SEUBuilding next : buildingsInRegion) {
//			centerPoint.x += next.getSelfBuilding().getX();
//			centerPoint.y += next.getSelfBuilding().getY();
//		}
		int count = this.size();
		if (count != 0) {
			centerPoint.x /= count;
			centerPoint.y /= count;
			
			Point mapCenter = world.getMapCenterPoint();
			distanceToMapCenter = Distance.getDistance(centerPoint, mapCenter);
		}
	}
	
	private void createZonePolygon() {
		try {
			CompositeConvexHull convexHull = new CompositeConvexHull();
			for (SEUBuilding next : this) {
				int[] apexList = next.getSelfBuilding().getApexList();
				for (int i = 0; i < apexList.length; i = i + 2) {
					convexHull.addPoint(apexList[i], apexList[i + 1]);
				}
			}
//			for (SEUBuilding next : buildingsInRegion) {
//				int[] apexList = next.getSelfBuilding().getApexList();
//				for (int i = 0; i < apexList.length; i = i + 2) {
//					convexHull.addPoint(apexList[i], apexList[i + 1]);
//				}
//			}
			zonePolygon = convexHull.getConvexPolygon();
			
			radiusLength = Math.hypot(zonePolygon.getBounds().getWidth(), zonePolygon.getBounds().getHeight());
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		radiusLength /= 2.0;
	}

	@Override
	public int compareTo(SeuZone o) {
		if (this.zoneValue > o.getZoneValue())
			return -1;
		
		if (this.zoneValue < o.getZoneValue())
			return 1;
		
		return 0;
	}
	
	public boolean addBuilding(SEUBuilding building) {
		totalArea += building.getSelfBuilding().getTotalArea();
		totalGroundArea += building.getSelfBuilding().getGroundArea();
		totalInitialFuel += building.getInitialFuel();
		buildingCount ++;
		
		//return buildingsInRegion.add(building);
		return super.add(building);
	}
	
	public Polygon getZonePolygon() {
		return zonePolygon;
	}
	
	public Point getZoneCenter() {
		return centerPoint;
	}
	
	public double getTotalInitialFuel() {
		return totalInitialFuel;
	}
	
	public double getTotaoGroundArea() {
		return totalGroundArea;
	}
	
	public double getTotalArea() {
		return totalArea;
	}
	
	public int getZoneId() {
		return this.id;
	}
	
	public List<SeuZone> getNeighbourZones() {
		return neighbourZones;
	}
	
	public void addNeighbourZoneIds(Integer zoneId) {
		if (!neighbourZoneIds.contains(zoneId))
			neighbourZoneIds.add(zoneId);
	}
	
	public void addNeighbourZone(SeuZone zone) {
		if (zone == null)
			return;
		if (!neighbourZones.contains(zone))
			neighbourZones.add(zone);
	}
	
	public List<Integer> getNeighbourZoneIds() {
		return neighbourZoneIds;
	}
	
	public double getZoneValue() {
		return this.zoneValue;
	}
	
	public double getLocalValue() {
		return this.localValue;
	}
	
	public double getGlobalValue() {
		return this.globalValue;
	}
	
	public int getBuildingCount() {
		return this.buildingCount;
	}
	
	public void setRadiusLength(double radiusLength) {
		this.radiusLength = radiusLength;
	}
	
	public double getRadiusLength() {
		return this.radiusLength;
	}
	
	public void addSurroundingRoad(Road road) {
		surroundingRoad.add(road.getID());
	}
	
	public void addSurroundingRoad(EntityID road) {
		surroundingRoad.add(road);
	}
	
	public Set<EntityID> getSurroundingRoad() {
		return surroundingRoad;
	}
	
	public double getDistanceToMapCenter() {
		return this.distanceToMapCenter;
	}
	
	public Set<Road> getAllEntranceRoad() {
		if (allEntranceRoad == null) {
			allEntranceRoad = new FastSet<Road>();
			for (SEUBuilding next : this) {
				allEntranceRoad.addAll(world.getEntrance().getEntrance(next.getSelfBuilding()));
			}
		}
		
		return allEntranceRoad;
	}
	
	public Set<EntityID> getUnvisitBuildings() {
		return unvisitBuildings;
	}
	
	public void removeVisitedFromUnvisit(Collection<EntityID> visited) {
		this.unvisitBuildings.removeAll(visited);
	}
	
	public Set<EntityID> getBuriedHumans() {
		return this.buriedHumans;
	}
	
	public void addBuriedHuman(EntityID... humans) {
		for (EntityID next : humans) {
			this.buriedHumans.add(next);
		}
	}
	
	
	public boolean isAllEntranceCleared() {
		return this.unclearedEntrances.isEmpty();
	}
	
	public boolean containtBuilding(EntityID id) {
		return this.contains(world.getSeuBuilding(id));
	}
	
	public SEUWorldModel getWorld() {
		return world;
	}

	public void setWorld(SEUWorldModel world) {
		this.world = world;
	}
	
	public void setBelongPfClusterIndex(int pfClusterIndex) {
		this.belongPfClusterIndex = pfClusterIndex;
	}
	
	public int getBelongPfClusterIndex() {
		return this.belongPfClusterIndex;
	}
	public List<SEUBuilding> getBuldingsInRegion(){
		return this;
	}
	
	//yuanc14.10.10
		public boolean isNeighbourRegion(SeuZone region){
			return neighbourRegions.contains(region);
		}
}